/*
 * Copyright (c) 1995, 1996 Takahide Matsutsuka
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * REX-5535 (MB89352A (SPC)) SCSI Driver
 *    by Takahide Matsutsuka (matsu@cs.titech.ac.jp) February, 1996
 *
 * Acknowledgements: Many of the algorithms used in this driver are
 * inspired by the work of Tohru Kobayashi (koba@lsi.yokogawa.co.jp).
 * Thanks a lot!
 */
/*
 * This version work on:
 *    RATOC REX-5535/5535AC/5535AMC/5535X/5535XM
 *
 * $Id: spc.c,v 1.1.6.2 1998/09/03 15:59:24 shiba Exp $
 */

#include	<sys/param.h>
#include	<sys/systm.h>
#include	<sys/kernel.h>
#include	<sys/conf.h>
#include	<sys/ioctl.h>
#include	<sys/disklabel.h>
#include	<sys/buf.h>
#include 	<sys/malloc.h>
#include	<sys/syslog.h>

#include 	<scsi/scsi_all.h>
#include	<scsi/scsiconf.h>

#include	<i386/isa/isa.h>
#include	<i386/isa/isa_device.h>
#include	<i386/isa/icu.h>

#include	<machine/clock.h>

#include	"apm.h"
#if NAPM > 0
#include	<machine/apm_bios.h>
#endif /* NAPM > 0 */

/* pccard support */
#include	"card.h"
#if NCARD > 0
#include	<sys/select.h>
#include	<pccard/cardinfo.h>
#include	<pccard/slot.h>
#include	<pccard/driver.h>
#endif /* NCARD > 0 */

#include	<i386/isa/spc.h>

#include	"spc.h"
#ifndef NSPC
#define NSPC		1
#endif
#define	SPC_HOST_ID	7
#define	NTARGETS	8
#define	NLUNS		8
#define SPC_TIMEOUT	10000000

#define SPC_SET_TC(port, len)	do {\
					outb((port) + SPC_TCH, (len)>>16);\
					outb((port) + SPC_TCM, (len)>>8);\
					outb((port) + SPC_TCL, (len));\
				} while(0)
#define SPC_GET_TC(port)	(\
				 (inb((port) + SPC_TCH)<<16)|\
				 (inb((port) + SPC_TCM)<<8)|\
				 inb((port) + SPC_TCL)\
				 )
					
#define SQ_FLG_USEDISCON (1<<0)
#define SQ_FLG_RSNS	 (1<<1)		/* request sense */

static int	spc_unit = 0;

struct spc_queue {
    struct spc_queue	*cq_next;
    int			flags;
    struct scsi_xfer	*xs;
    u_char		*data;		/* data pointer */
    int			datalen;	/* data pointer */
    struct scsi_xfer	*alt_xs;	/* */
    u_char		*alt_data;	/* data pointer */
    int			alt_datalen;	/* data pointer */
    int			alt_statusbyte;
};

typedef enum {
    IN_BUSFREE = 0,
    IN_SELECTION_IN_PROGRESS,
    IN_CONNECTED_MO,
    IN_CONNECTED,
    IN_WAITDISCONNECT,
    IN_WAITBUSFREE
} SPC_STATE;

static struct spc_data { /* one of these per adapter */
    int			unit;
    int			baseport;	/* I/O base		*/
    SPC_STATE		state;
    struct scsi_link	sc_link;	/* each existing device	*/
    struct spc_queue	*cq_first;	/* command queue first	*/
    struct spc_queue	*cq_last;	/* command queue last	*/
    struct spc_queue	*fq_first;	/* free queue first 	*/
    struct spc_queue	*fq_last;	/* free queue last	*/
    struct spc_queue	*busy[NTARGETS][NLUNS];
    int			target;		/* current target	*/
    int			lun;		/* current lun		*/
    u_char		*data;		/* current data pointer	*/
    int			datalen;	/* current data pointer	*/
    int			statusbyte;
    int			slot;		/* information for reconfiguration */
    int			alive;		/* is driver alive?                */
} *spcdata[NSPC];

/* the below structure is so we have a default dev struct for out link struct */ static struct scsi_device	spc_dev = {
    NULL,			/* Use default error handler */
    NULL,			/* have a queue, served by this */
    NULL,			/* have no async handler */
    NULL,			/* Use default 'done' routine */
    "spc",
    0,
    { 0, 0 }
};

static inline void
spc_free_queue(struct spc_data *spc, struct spc_queue *sq)
{
    unsigned int	s = splbio();
    if( !spc->fq_first )
	spc->fq_first = sq;
    else
	spc->fq_last->cq_next = sq;
    spc->fq_last = sq;
    splx(s);
}

/* prototypes ... */
/* called by higher level drivers */
static int		spc_probe(struct isa_device *);
static int		spc_attach(struct isa_device *);
static int32_t		spc_scsi_cmd(struct scsi_xfer *);
static u_int32_t	spc_adapter_info(int);
static void		spc_minphys(struct buf *);
void			spcintr(int);

/* private routines */
static void	spc_done(struct spc_data *, int32_t);
static int	spc_command(struct spc_data *);
static void	spc_selection(struct spc_data *);
static int	spc_sel_sub(struct spc_data *, struct spc_queue *, int32_t *);
static int	spc_reselection(struct spc_data *, int);
static int	spc_poll(struct spc_data *, struct spc_queue *);
static int	spc_progtx(int, u_char *, int, int, int *);
static int	spc_progrx(int, u_char *, int, int, int *);
static int	spc_byte_in(int, int);
static int	spc_byte_out(int, int, int);
static int	spc_next_phase(int);
static int	spc_pollport(int, int, int);
static struct scsi_xfer	*spc_alloc_rsns(const struct scsi_xfer *);
static void	spc_free_rsns(struct scsi_xfer *);
static void	spc_complete(struct spc_data *);
static int	spc_init(struct spc_data *);
static void	spc_cleanup(struct spc_data *);
static void	spc_dumpreg(int, const char *);
#ifdef SCSI_DETACH
static void	spcdetach __P((struct isa_device *dev));
#endif

/* pccard support */
#if NCARD > 0
static int	spc_card_intr(struct pccard_devinfo *);
static void	spc_card_unload(struct pccard_devinfo *);
static int	spc_card_init(struct pccard_devinfo *);

static struct pccard_device spc_info =
{
    "spc",
    spc_card_init,
    spc_card_unload,
    spc_card_intr,
    0,				/* Attributes - presently unused */
    &bio_imask			/* Interrupt mask for device */
	/* This should also include net_imask?? */
};

DATA_SET(pccarddrv_set, spc_info);

static int probing_pccard = 0;

/*
 *	Initialize the device - called from Slot manager.
 *	if first is set, then initially check for
 *	the device's existence before initialising it.
 *	Once initialised, the device table may be set up.
 */
static int
spc_card_init(struct pccard_devinfo *devi)
{
    /*
     *      validate unit number.
     */
    static int already_spcinit;
    struct spc_data    *spc = spcdata[devi->pd_unit];

    if( devi->pd_unit >= NSPC )
	return(ENODEV);
    /*      Make sure it isn't already initialisd.  */
    if( already_spcinit == 1 ) {
	    if( spc_init(spc) == 0 )
		return(ENXIO);
	    if( spc_attach(&devi->isahd) == 0 )
		return(ENXIO);
	    return 0;
    }
    /*
     *      Probe the device. If a value is returned, the
     *      device was found at the location.
     */
#if 0
    printf("probe spc\n");
#endif
    probing_pccard = 1;
    if( spc_probe(&devi->isahd) == 0 ) {
	probing_pccard = 0;
	return(ENXIO);
    }
    probing_pccard = 0;
#if 0
    printf("attach spc\n");
#endif
    if( spc_attach(&devi->isahd) == 0 )
	return(ENXIO);
    already_spcinit = 1;
    return 0;
	
    /*
     *      XXX TODO:
     *      If it was already inited before, the device structure
     *      should be already initialised. Here we do
     *      reset spc,
     *      but I don't know whether this is the best way...
     */
    if( !spc_init(spc) )
	return (ENXIO);
    spc_dumpreg(spc->baseport, "resume");
    printf("spc%d: resumed\n", spc->unit);
    return 0;
}
/*
 *	spc_card_unload - unload the driver and clear the table.
 *	XXX TODO:
 *	This is called usually when the card is ejected, but
 *	can be caused by the modunload of a controller driver.
 *	The idea is reset the driver's view of the device
 *	and ensure that any driver entry points such as
 *	read and write do not hang.
 */
static void
spc_card_unload(struct pccard_devinfo *devi)
{
    int unit = devi->pd_unit;
    struct spc_data    *spc = spcdata[unit];

    /*
     *	Calling printf() in xxunload() makes hotswap operations
     *  unstable, so I removed all printf()'s in xxunload() except
     *  critical ones.
     *  Tatsumi Hosokawa <hosokawa@jp.FreeBSD.org>
     */

    spc->alive = 0;
#ifdef SCSI_DETACH
    spcdetach(&devi->isahd);
#endif
#if 0
    printf("spc%d: unload\n", unit);
#endif
}
/*
 *	spc_card_intr - Shared interrupt called from
 *	front end of PC-Card handler.
 */
static int
spc_card_intr(struct pccard_devinfo *devi)
{
    spcintr(devi->pd_unit);
    return 1;
}
#endif /* NCARD > 0 */

#define SPC_DISCONNECT 		/* undef if driver doesn't work well yet... */
#undef SPC_DEBUG		/* display trace data */

#ifdef SPC_DEBUG
#define SPC_TRACE(x)		printf x;
#else
#define	SPC_TRACE(x)
#endif

struct	isa_driver	spcdriver = {
    spc_probe, spc_attach, "spc",
};

struct scsi_adapter	spc_switch = {
    spc_scsi_cmd,
    spc_minphys,
    0,
    0,
    spc_adapter_info,
    "spc",
    { 0, 0 } 
};

/* 
 * INITIALIZATION ROUTINES (probe, attach ++)
 */

/*
 * spc_probe: probe for MB89352(A) SCSI controller
 * returns non-zero value if a controller is found.
 */
static int
spc_probe(struct isa_device *dvp)
{
    int			unit = spc_unit;
    struct spc_data	*spc;

    SPC_TRACE(("spc: probe\n"));

    if( (unit >= NSPC) && !dvp->id_reconfig ) {
	printf("spc%d: unit number too high\n", unit);
	return 0;
    }

    if ((dvp->id_flags & SPC_FLAGS_PCCARD_ONLY) && !probing_pccard)
	return 0;

    if( !dvp->id_reconfig ) {
        /* prepare structure for the unit */ 
        spc = malloc(sizeof(struct spc_data), M_TEMP, M_NOWAIT);
        bzero(spc, sizeof(struct spc_data));
        spc->baseport = dvp->id_iobase;
	spc->fq_first = spc->fq_last = 0;
        spcdata[unit] = spc;
        dvp->id_unit = unit;

	if( !spc_init(spc) ) {
	    /* init failed, maybe scsi bus is off-lined */
	    free(spc, M_TEMP);
	    return 0;
	}

	spc_unit++;	/* for the next adapter */
	printf("spc%d: REX5535 (MB89352A) PCMCIA SCSI adapter\n", unit);
    } else {
	/* reconfiguration time */
        spc = spcdata[dvp->id_unit];
        if( !spc_init(spc) ) 
            return 0;
    }

    return 16;
}

#ifdef SCSI_DETACH
static void
spcdetach(dev)
struct isa_device *dev;
{
      int unit = dev->id_unit;
      struct scsibus_data *scbus;

      scbus = (struct scsibus_data *)scsi_extend_get(spcdata[unit]->sc_link.scsibus);
      scsi_detachdev(scbus);
}
#endif


static int
spc_attach(struct isa_device *dvp)
{
    int     		unit = dvp->id_unit;
    struct spc_data	*spc = spcdata[unit];
    struct scsibus_data	*scbus;

    SPC_TRACE(("spc%d: attach\n",unit));

    if( dvp->id_reconfig )
	return 1;

    /*
     * fill in the prototype scsi_link.
     */
    spc->sc_link.adapter_unit = unit;
    spc->sc_link.adapter_targ = SPC_HOST_ID;
    spc->sc_link.adapter_softc = spc;
    spc->sc_link.adapter = &spc_switch;
    spc->sc_link.device = &spc_dev;

    /*
     * Prepare the scsibus_data area for the upperlevel
     * scsi code.
     */
    scbus = scsi_alloc_bus();
    if( !scbus ) return 0;
    scbus->adapter_link = &spc->sc_link;

    /*
     * ask the adapter what subunits are present
     */

    scsi_attachdevs(scbus);

    return 1;
}

static u_int32_t
spc_adapter_info(int unit)
{
    SPC_TRACE(("spc%d: adapter_info\n", unit));
    return 2; /* 2 outstanding requests at a time are acceptable */
}

static void 
spc_minphys(struct buf *bp)
{
    SPC_TRACE(("spc_minphys:%d\n", bp->b_bcount));
#if 0 /* probably not needed...can it be true? */
#define PAGESIZ		4096
#define SPC_NSEG	2
    if( bp->b_bcount > ((SPC_NSEG - 1) * PAGESIZ) )
	bp->b_bcount = ((SPC_NSEG - 1) * PAGESIZ);
#endif
}

/* 
 * DRIVER ROUTINES (scsi_cmd, selection, ++)
 */

/*
 * desirable sequence
 *   1.spc_scsi_cmd: receive scsi_xfer and queueing
 *   2.spcintr:      called by kernel interrupt handler
 *   3.proceed scsi sequence
 *    (1) spc_selection:    starts new scsi command from queue
 *    (2) spc_command:      proceed scsi phase
 *    (3) spc_reselection:  reselected by target
 *    (4) spc_complete:     if needed, do self request sense
 *    (5) spc_poll:         if booting time, don't use interrupt
 */
/*
 * spc_scsi_cmd: receive scsi command from higher level module
 */
static int32_t
spc_scsi_cmd(struct scsi_xfer *xs)
{
    int			unit = xs->sc_link->adapter_unit;
    struct spc_data	*spc = spcdata[unit];
    struct spc_queue	*sq;
    unsigned int	s;
    int32_t		erc;

    SPC_TRACE(("spc%d:spc_scsi_cmd\n", unit));
    if( !spc->alive ) {
	/* now card is off-lined */
	xs->error = XS_DRIVER_STUFFUP;
	return (COMPLETE);
    }
    /* queueing command */
    if( spc->fq_first ) {
	s = splbio();
	sq = spc->fq_first;
	spc->fq_first = spc->fq_first->cq_next;
	splx(s);
    } else {
	if( (sq = (struct spc_queue *) malloc(sizeof(struct spc_queue), 
					      M_TEMP, M_NOWAIT)) == NULL ) {
    	    /* can't malloc */
	    printf("spc%d: queue full\n", spc->unit);
	    return (TRY_AGAIN_LATER);
	}
    }

    sq->xs = xs;
    sq->data = xs->data;
    sq->datalen = xs->datalen;
    sq->cq_next = NULL;
    sq->alt_xs = NULL;
#ifndef SPC_DISCONNECT
    sq->flags = 0;
#else
    sq->flags = SQ_FLG_USEDISCON;
#endif

    if( xs->flags & SCSI_NOMASK ) {
	/* it's booting time. don't allow interrupts. */
        sq->flags &= ~SQ_FLG_USEDISCON;
	erc = spc_poll(spc, sq);
	spc_free_queue(spc, sq);
	return erc;
    }

    s = splbio();
    /* add to end of queue */
    if( spc->cq_first == NULL ) {
    	spc->cq_first = sq;
    } else {
        spc->cq_last->cq_next = sq;
    }
    spc->cq_last = sq;
    spc_selection(spc);					/* start selection */
    outb(spc->baseport + SPC_SCTL, SPC_SCTL_ENABLE);	/* enable interrupt */
    splx(s);
    return (SUCCESSFULLY_QUEUED);
}

/*
 * spcintr: called by kernel interrupt handler
 */
void
spcintr(int unit)
{
    struct spc_data	*spc = spcdata[unit];
    int			port = spc->baseport;
    int			intr, target = 0;

    SPC_TRACE(("spc%d: spcintr\n", unit));
    intr = inb(port + SPC_INTS);	/* read interrupt */
    if( intr & SPC_INTS_RESELECT ) {
        target = inb(port + SPC_TEMP);
    }
    outb(port + SPC_INTS, intr);	/* clear interrupt */

    /* HARDERR */
    if( intr & SPC_INTS_HARDERR ) {
	printf("spc%d: HARDERR SERR=0x%d\n", spc->unit, inb(port + SPC_SERR));
	goto err;
    }

    /* RESET */
    if( intr & SPC_INTS_RESET ) {
	printf("spc%d: reset condition\n", spc->unit);
	goto err;
    }

    /* SERVREQ */
    if( intr & SPC_INTS_SERVREQ ) {
	printf("spc%d: spurious interrupt(SERVREQ)\n", spc->unit);
	goto err;
    }

    /* DISCONNECT */
    if( intr & SPC_INTS_DISCONNECT ) {
        if( spc->state == IN_WAITDISCONNECT ) {
	    /* nothing to do */
	} else if( spc->state == IN_WAITBUSFREE ) {
	    spc_complete(spc);
	} else {
	    printf("spc%d: spurious interrupt(DISCONNECT) state=%d\n",
		   spc->unit, spc->state);
	    goto err;
	}
        spc->state = IN_BUSFREE;
    }

    /* RESELECT */
    if( intr & SPC_INTS_RESELECT ) {
	if( spc->state == IN_SELECTION_IN_PROGRESS ) {
	    outb(port + SPC_SCMD, SPC_SCMD_RESETATN); /* reset ATN */
	    if( (spc->busy[spc->target][spc->lun]->flags & SQ_FLG_RSNS) == 0 ) {
			spc->busy[spc->target][spc->lun] = NULL;
	    }
	}
	if( spc_reselection(spc, target) < 0 ) {
	    goto err;
	}
        if( spc_command(spc) < 0 ) {
	    goto err;
	}
	intr &= ~(SPC_INTS_COMPLETE);
    }

    /* COMPLETE */
    if( intr & SPC_INTS_COMPLETE ) {
    	if( spc->state == IN_SELECTION_IN_PROGRESS ) {
	    spc->state = IN_CONNECTED_MO;
	    spc->cq_first = spc->cq_first->cq_next;	/* delete from queue */
	    if( spc_command(spc) < 0 ) {
		goto err;
	    }
	} else {
	    printf("spc%d: spurious interrupt(COMPLETE) state=%d\n",
		   spc->unit, spc->state);
	    goto err;
	}
    }

    /* TIMEOUT */
    if( intr & SPC_INTS_TIMEOUT ) {
	outb(port + SPC_SCMD, SPC_SCMD_RESETATN);		/* reset ATN */
	if( spc->state == IN_SELECTION_IN_PROGRESS ) {
	    printf("spc%d: selection timeout id=%d\n", spc->unit, spc->target);
	    spc->statusbyte = -1;
	    spc->cq_first = spc->cq_first->cq_next;	/* delete from queue */
	    spc_complete(spc);
	} else {
	    printf("spc%d: spurious interrupt(TIMEOUT) state=%d\n",
		   spc->unit, spc->state);
	    goto err;
	}
        spc->state = IN_BUSFREE;
    }
    spc_selection(spc);
    return;

 err:
    (void) spc_init(spc);
    spc_cleanup(spc);
    spc->state = IN_BUSFREE;
    spc_selection(spc);
    return;
}

/*
 * (1) spc_selection: starts new scsi command
 */
static void
spc_selection(struct spc_data *spc)
{
    struct spc_queue	*sq;
    int32_t		erc;

    SPC_TRACE(("spc%d: spc_selection\n", spc->unit));

    if( spc->state != IN_BUSFREE ) {
        return;
    }

 again:
    if( (sq = spc->cq_first) != NULL ) {
	spc->state = IN_SELECTION_IN_PROGRESS;
	if( spc_sel_sub(spc, sq, &erc) < 0 ) {
	    sq->xs->flags |= ITSDONE;
	    sq->xs->resid = sq->xs->datalen;
	    sq->xs->error = erc;
	    scsi_done(sq->xs);
	    spc->cq_first = sq->cq_next; /* delete from queue */
	    spc_free_queue(spc, sq);
	    spc->state = IN_BUSFREE;
	    goto again;
	}
    }
}

static int
spc_sel_sub(struct spc_data *spc, struct spc_queue *sq, int32_t *erc)
{
    const struct scsi_xfer *xs;
    int			port = spc->baseport;
    int			target, lun, val;

    SPC_TRACE(("spc%d: spc_sel_sub\n", spc->unit));

    xs = sq->xs;
    target = xs->sc_link->target;
    lun = xs->sc_link->lun;

    if( sq->flags & SQ_FLG_RSNS ) {
    	/* request sense in progress */
        if( spc->busy[target][lun] != sq ) {
	    printf("spc%d: ??????\n", spc->unit);
	}
    } else {
        if( spc->busy[target][lun] ) {
	    /* target is busy */
	    printf("spc%d: target=%d, lun=%d is busy\n", spc->unit, target, lun);
	    *erc = XS_BUSY;
	    return -1;
        }
    }
    if( xs->flags & SCSI_RESET ) {
	/* command is empty, reset controller */
	printf("spc%d: reset controller by higher module\n", spc->unit);
	(void) spc_init(spc);
	*erc = XS_NOERROR;
	return -1;
    }

    spc->target = target;
    spc->lun = lun;
    spc->busy[target][lun] = sq;
    spc->data = xs->data;
    spc->datalen = xs->datalen;
    spc->statusbyte = 0;

    outb(port + SPC_PCTL, 0);			/* disable bus free interrupt */
	outb(port + SPC_SCMD, SPC_SCMD_SETATN);	/* set ATN for IDENTIFY MSG */
    val = inb(port + SPC_BDID);			/* get host id */
    val |= (1 << spc->target);			/* set target id */
    outb(port + SPC_TEMP, val);			/* set host id & target id */
    outb(port + SPC_TCH,  0x0f);		/* set replying time */
    outb(port + SPC_TCM,  0x46);
    outb(port + SPC_TCL,  0x04);		/* set twait */
    outb(port + SPC_SCMD, SPC_SCMD_SELECT);	/* exec selection phase! */
    *erc = XS_NOERROR;
    return 0;
}

/*
 * (2) spc_command: proceed scsi phase
 */
static int
spc_command(struct spc_data *spc)
{
    struct scsi_xfer	*xs;
    struct spc_queue	*sq;
    int			port = spc->baseport;
    int			phase, xfered, r, msg;

    SPC_TRACE(("spc%d: spc_command\n", spc->unit));

    sq = spc->busy[spc->target][spc->lun];
    xs = sq->xs;

 again0:
    if( (phase = spc_next_phase(port)) < 0 ) {
        return -1;
    }

    switch( phase ) {
    case SPC_PCTL_MSGOUT:
	outb(port + SPC_SCMD, SPC_SCMD_RESETATN); /* reset ATN */
    	if( spc->state == IN_CONNECTED_MO ) {
	    /* send identify message */
	    if( sq->flags & SQ_FLG_USEDISCON ) {
	        msg = 0xc0 | spc->lun;
	    } else {
	        msg = 0x80 | spc->lun;
	    }
	    if( spc_byte_out(port, msg, SPC_PCTL_MSGOUT) < 0 ) {
	        printf("spc%d: cannot send IDENTIFY MSG\n", spc->unit);
	        return -1;
	    }
	    spc->state = IN_CONNECTED;
	} else {
	    /* send No Operation message */
	    if( spc_byte_out(port, 0x08, SPC_PCTL_MSGOUT) < 0 ) {
	        printf("spc%d: cannot send No Operation MSG\n", spc->unit);
	        return -1;
	    }
	}
        break;
    case SPC_PCTL_MSGIN:
        if( (msg = spc_byte_in(port, SPC_PCTL_MSGIN)) < 0 ) {
	    return -1;
	}
	switch( msg ) {
	case SPC_MSG_SAVEDATAP:
	    sq->data = spc->data;	/* save data pointer */
	    sq->datalen = spc->datalen;	/* save data pointer */
	    break;
	case SPC_MSG_RESTOREP:
	    spc->data = sq->data;	/* restore pointers */
	    spc->datalen = sq->datalen;	/* restore pointers */
	    break;
	case SPC_MSG_DISCONNECT:
	    spc->state = IN_WAITDISCONNECT;
	    return 0;
	case SPC_MSG_CMDCOMPLETE:
	    spc->state = IN_WAITBUSFREE;
	    return 0;
	default:
	    printf("spc%d: invalid message 0x%x\n", spc->unit, msg);
	    return -1;
	}
        break;
    case SPC_PCTL_COMMAND:
        if( spc_progtx(port, (u_char *)xs->cmd, xs->cmdlen,
		       phase, &xfered) < 0 ) {
	    return -1;
	}
        break;
    case SPC_PCTL_DATAOUT:
        r = spc_progtx(port, spc->data, spc->datalen, phase, &xfered);
	spc->data += xfered;
	spc->datalen -= xfered;
        if( r < 0 ) {
	    return -1;
	}
	break;
    case SPC_PCTL_DATAIN:
        r = spc_progrx(port, spc->data, spc->datalen, phase, &xfered);
	spc->data += xfered;
	spc->datalen -= xfered;
        if( r < 0 ) {
	    return -1;
	}
	break;
    case SPC_PCTL_STATUS:
        if( (spc->statusbyte = spc_byte_in(port, SPC_PCTL_STATUS)) < 0 ) {
	    return -1;
	}
	spc->statusbyte &= 0x1e;
	break;
    default:
    	printf("spc%d: invalid phase 0x%x\n", spc->unit, phase);
	return -1;
    }
    goto again0;
}

/*
 * (3) spc_reselection: reselected by target
 */
static int
spc_reselection(struct spc_data *spc, int target)
{
    struct spc_queue	*sq;
    int			port = spc->baseport;
    int			id, phase, val;

    SPC_TRACE(("spc%d: spc_reselection\n", spc->unit));

    for(id = 0; id < 8; id++) {
        if( (target & (1<<id)) && (id != SPC_HOST_ID) ) {
	    break;
	}
    }
    if( id >= 8 ) {
        printf("spc%d: cannot get target ID 0x%02x\n", spc->unit, target);
	return -1;
    }
    if( (phase = spc_next_phase(port)) < 0 ) {
        return -1;
    }
    if( phase != SPC_PCTL_MSGIN ) {
	printf("spc%d: not MSGIN\n", spc->unit);
	return -1;
    }
    if( (val = spc_byte_in(port, SPC_PCTL_MSGIN)) < 0 ) {
	printf("spc%d: cannot recive IDENTIFY MSG\n", spc->unit);
	return -1;
    }
    if( (val & 0x80) == 0 ) {
        /* not identify message */
	printf("spc%d: bogus reselection\n", spc->unit);
	return -1;
    }
    spc->target = id;
    spc->lun = val & 0x07;	/* get LUN */
    sq = spc->busy[id][spc->lun];
    spc->state = IN_CONNECTED;
    spc->data = sq->data;	/* restore data pointer */
    spc->datalen = sq->datalen;	/* restore data pointer */
    return 0;
}

/*
 * (4) spc_complete: if needed, do self request sense
 */
static void
spc_complete(struct spc_data *spc)
{
    struct spc_queue *sq;

    SPC_TRACE(("spc%d: spc_complete\n", spc->unit));

    sq = spc->busy[spc->target][spc->lun];
    if( sq->flags & SQ_FLG_RSNS ) { /* request sense in progress */
        int statusbyte;
	struct scsi_xfer *sxs;
	
	statusbyte = spc->statusbyte;
	sxs = sq->xs;
	sq->xs = sq->alt_xs;
	sq->alt_xs = NULL;
	sq->flags &= ~SQ_FLG_RSNS;
	spc->data = sq->alt_data;
	spc->datalen = sq->alt_datalen;
	spc->statusbyte = sq->alt_statusbyte;
	spc_free_rsns(sxs);
	if( statusbyte == 0 ) {
            spc_done(spc, XS_SENSE);
	} else {
            printf("spc%d: request sense failed\n", spc->unit);
            spc_done(spc, XS_DRIVER_STUFFUP);
	}
    } else {
	switch( spc->statusbyte ) {
	case 0x00:
            spc_done(spc, XS_NOERROR);
	    break;
        case 0x02:
            if( sq->xs->flags & SCSI_ERR_OK ) {
                spc_done(spc, XS_NOERROR);
	    } else if( (sq->flags & SQ_FLG_RSNS) == 0 ) {
	        struct scsi_xfer *sxs;
	        sxs = spc_alloc_rsns(sq->xs);
	        sq->alt_xs = sq->xs;
	        sq->alt_data = spc->data;
	        sq->alt_datalen = spc->datalen;
	        sq->alt_statusbyte = spc->statusbyte;
	        sq->xs = sxs;
	        sq->flags |= SQ_FLG_RSNS;
		/* in queue */
		sq->cq_next = NULL;
		if( spc->cq_first == NULL ) {
		    spc->cq_first = sq;
		} else {
		    spc->cq_last->cq_next = sq;
		}
		spc->cq_last = sq;
	    }
	    break;
        case 0x08:
            spc_done(spc, XS_BUSY);
	    break;
        default:
            printf("spc%d: unknown statusbyte 0x%x\n",
		   spc->unit, spc->statusbyte);
            spc_done(spc, XS_DRIVER_STUFFUP);
        }
    }
}

/*
 * (5) spc_poll: use polling (at booting time)
 */
static int
spc_poll(struct spc_data *spc, struct spc_queue *sq)
{
    struct scsi_xfer	*xs;
    int			port = spc->baseport;
    int			phase, val, wait;
    int32_t		erc;

    SPC_TRACE(("spc%d: spc_poll\n", spc->unit));

    xs = sq->xs;
    outb(port + SPC_INTS, 0xff);		/* reset interrupt */
    spc->state = IN_SELECTION_IN_PROGRESS;
    if( spc_sel_sub(spc, sq, &erc) < 0 ) {
    	printf("spc%d: spc_poll: selection not start\n", spc->unit);
        (void) spc_init(spc);
	goto err;
    }

    /* need to check end of selection phase ... */
    for(wait = SPC_TIMEOUT;;) {
	if( (val = inb(port + SPC_INTS)) & SPC_INTS_COMPLETE ) {
            /* selection success */
            outb(port + SPC_INTS, 0xff);		/* reset interrupt */
	    break;
	}
	if( val & SPC_INTS_TIMEOUT ) {
	    /* selection timeout */
	    outb(port + SPC_SCMD, SPC_SCMD_RESETATN);	/* reset ATN */
	    goto err;
	}
	if( wait-- == 0 ) {
	    printf("spc%d: spc_poll: selecton not complete\n", spc->unit);
            (void) spc_init(spc);
	    goto err;
	}
    }

    spc->state = IN_CONNECTED_MO;
    if( spc_command(spc) < 0 ) {
        (void) spc_init(spc);
        goto err;
    }

    /* wait for bus free */
    if( spc_pollport(port+SPC_INTS, SPC_INTS_DISCONNECT,
		     SPC_INTS_DISCONNECT) < 0 ) {
	printf("spc%d: spc_poll: timeout at bus free\n", spc->unit);
        (void) spc_init(spc);
	goto err;
    }
    outb(port + SPC_INTS, 0xff);		/* reset interrupt */

    xs->status = spc->statusbyte;
    switch (xs->status) {
    case 0x00:
        xs->error = XS_NOERROR;
	break;
    case 0x02:
        if( xs->flags & SCSI_ERR_OK ) {
            xs->error = XS_NOERROR;
	} else if( (sq->flags & SQ_FLG_RSNS) == 0 ) {
	    struct scsi_xfer *sxs;
	    sxs = spc_alloc_rsns(xs);
	    sq->alt_xs = sq->xs;
	    sq->alt_data = spc->data;
	    sq->alt_datalen = spc->datalen;
	    sq->alt_statusbyte = spc->statusbyte;
	    sq->xs = sxs;
	    sq->flags |= SQ_FLG_RSNS;
	    if( spc_poll(spc, sq) != COMPLETE ) {
    		xs->error = XS_DRIVER_STUFFUP;
	        printf("spc%d: request sense failed\n", spc->unit);
	    } else {
	        xs->error = XS_SENSE;
	    }
	    sq->xs = sq->alt_xs;
	    sq->alt_xs = NULL;
	    sq->flags &= ~SQ_FLG_RSNS;
	    spc->data = sq->alt_data;
	    spc->datalen = sq->alt_datalen;
	    spc->statusbyte = sq->alt_statusbyte;
	    spc_free_rsns(sxs);
	} else {
    	    return (HAD_ERROR);
	}
	break;
    case 0x08:
        xs->error = XS_BUSY;
	break;
    default:
	printf("spc%d: unknown statusbyte 0x%x\n", spc->unit, xs->status);
    	xs->error = XS_DRIVER_STUFFUP;
    }
    xs->flags |= ITSDONE;
    xs->resid = spc->datalen;
    spc->state = IN_BUSFREE;
    if( (sq->flags & SQ_FLG_RSNS) == 0 ) {
        spc->busy[spc->target][spc->lun] = NULL;
    }
    return (COMPLETE);
 err:
    outb(port + SPC_INTS, 0xff);		/* reset interrupt */
    xs->error = XS_TIMEOUT;
    xs->flags |= ITSDONE;
    xs->resid = spc->datalen;
    spc->state = IN_BUSFREE;
    if( (sq->flags & SQ_FLG_RSNS) == 0 ) {
        spc->busy[spc->target][spc->lun] = NULL;
    }
    return (HAD_ERROR);
}

/* 
 * MISC FUNCTIONS
 */
static struct scsi_xfer
*spc_alloc_rsns(const struct scsi_xfer *xs)
{
    struct scsi_xfer	*sxs;
    struct scsi_sense	*scmd;

    if( (sxs = malloc(sizeof(struct scsi_xfer), M_TEMP, M_NOWAIT)) == NULL ) {
    	return NULL;
    }
    if( (scmd = malloc(sizeof(struct scsi_sense), M_TEMP, M_NOWAIT)) == NULL ) {
		free(sxs, M_TEMP);
    	return NULL;
    }
    bzero(sxs, sizeof(struct scsi_xfer));
    bzero(scmd, sizeof(struct scsi_sense));
    scmd->op_code = REQUEST_SENSE;
    scmd->byte2 = (xs->sc_link->lun << 5) & 0xe0;
    scmd->length = sizeof(struct scsi_sense_data);
    sxs->flags = SCSI_DATA_IN;
    sxs->cmd = (void *)scmd;
    sxs->cmdlen = sizeof(struct scsi_sense);
    sxs->data = (u_char *)&xs->sense;
    sxs->datalen = sizeof(struct scsi_sense_data);
    sxs->timeout = xs->timeout;
    sxs->sc_link = xs->sc_link;
    return sxs;
}

static void
spc_free_rsns(struct scsi_xfer *sxs)
{
    free(sxs->cmd, M_TEMP);
    free(sxs, M_TEMP);
}

static void
spc_dumpreg(int port, const char *p)
{
    printf("spc: %s SSTS=0x%x INTS=0x%x PSNS=0x%x PCTL=0x%x\n",
	   p,
	   inb(port + SPC_SSTS),
	   inb(port + SPC_INTS),
	   inb(port + SPC_PSNS),
	   inb(port + SPC_PCTL));
}

static int
spc_pollport(int port, int mask, int ptn)
{
    int	wait, val;

    for(wait = SPC_TIMEOUT; wait > 0; wait--) {
	if( ((val = inb(port)) & mask) == ptn ) {
	    return val;
	}
    }
    return -1;
}

static int
spc_next_phase(int port)
{
    int	psns;

    if( (psns = spc_pollport(port+SPC_PSNS, SPC_PSNS_REQ, SPC_PSNS_REQ)) < 0 ) {
		printf("spc: spc_next_phase: REQ not assert\n");
		return -1;
    }
    return psns & SPC_PSNS_PHASE;
}

static void
spc_done(struct spc_data *spc, int32_t xserror)
{
    struct spc_queue	*sq;

    sq = spc->busy[spc->target][spc->lun];
    spc->busy[spc->target][spc->lun] = NULL;
    sq->xs->flags |= ITSDONE;
    sq->xs->resid = spc->datalen;
    sq->xs->status = spc->statusbyte;
    sq->xs->error = xserror;
    if( sq->xs->resid ) { /* XXX */
        printf("spc_done: xs %d resid %d\n", sq->xs->datalen, sq->xs->resid);
    }
    scsi_done(sq->xs);
    spc_free_queue(spc, sq);
}

static int
spc_progtx(int port, u_char *data, int datalen, int phase, int *xfered)
{
    long	wait;
    int		n, ssts, datalenorg;

    datalenorg = datalen;
    outb(port + SPC_PCTL, phase);
    SPC_SET_TC(port, datalen);		/* transfer byte counter set */
    outb(port + SPC_SCMD, SPC_SCMD_PROGTX);	/* programed transfer */

    if( spc_pollport(port + SPC_SSTS, 0xf0, 0xb0) < 0 ) {
    	spc_dumpreg(port, "spc_progtx");
        *xfered = 0;
	return -1;
    }

    if( (inb(port + SPC_SSTS) & SPC_SSTS_DREMPTY) == 0 ) { /* XXX */
        spc_dumpreg(port, "AA spc_progtx: not empty");
    }

    for(;;) {
	for(wait = SPC_TIMEOUT;;) {
 	    if( ((ssts = inb(port + SPC_SSTS)) & SPC_SSTS_BUSY) == 0) {
	        goto doneWRITE;
	    }
 	    if( (ssts & 3) == SPC_SSTS_DREMPTY ) {
	        break;
	    }
 	    if( wait-- == 0 ) {
    		*xfered = datalenorg - SPC_GET_TC(port);
    		spc_dumpreg(port, "spc_progtx: timeout");
	        return -1;
	    }
 	}
 	if( datalen > 0 ) {
 	    if( datalen >= 8 ) {
	        n = 8;
 	    } else {
	        n = datalen;
	    }
 	    outsb(port + SPC_DREG, data, n);
 	    data += n;
 	    datalen -= n;
 	}
    }
 doneWRITE:
    *xfered = datalenorg - SPC_GET_TC(port);
    outb(port + SPC_INTS, SPC_INTS_COMPLETE|SPC_INTS_SERVREQ);
    return 0;
}

static int
spc_progrx(int port, u_char *data, int datalen, int phase, int *xfered)
{
    int		datalenorg, ssts, wait;

    datalenorg = datalen;
    outb(port + SPC_PCTL, phase);
    SPC_SET_TC(port, datalen);		/* transfer byte counter set */
    outb(port + SPC_SCMD, SPC_SCMD_PROGRX);	/* programed transfer */

    if( spc_pollport(port+SPC_SSTS, 0xd0, 0x90) < 0 ) {
    	spc_dumpreg(port, "spc_progrx");
        *xfered = 0;
        return -1;
    }

    while(1) {
        for(wait = SPC_TIMEOUT;;) {
            if( ((ssts = inb(port + SPC_SSTS)) & SPC_SSTS_DREMPTY) == 0 ) {
	        break;
	    }
	    if( (ssts & SPC_SSTS_BUSY) == 0) {
	        goto doneREAD;
	    }
	    if( wait-- == 0 ) {
    		spc_dumpreg(port, "spc_progrx: timeout");
    		*xfered = datalenorg - datalen;
		return -1;
	    }
	}
	if( datalen > 0 ) {
	    *data++ = inb(port + SPC_DREG);
	    datalen--;
	} else {
	    inb(port + SPC_DREG);	/* dummy read XXX */
	}
    }
 doneREAD:
    if( (inb(port + SPC_SSTS) & SPC_SSTS_DREMPTY) == 0 ) {	/* XXX */
        spc_dumpreg(port, "spc_progrx: not empty");
    }
    *xfered = datalenorg - datalen;
    outb(port + SPC_INTS, SPC_INTS_COMPLETE|SPC_INTS_SERVREQ);
    return 0;
}

static int
spc_byte_in(int port, int phase)
{
    int		retval;

    outb(port + SPC_PCTL, phase);
    outb(port + SPC_SCMD, SPC_SCMD_SETACK);
    if( spc_pollport(port+SPC_PSNS, SPC_PSNS_REQ, 0) < 0 ) {
        spc_dumpreg(port, "spc_byte_in");
        return -1;
    }
    retval = inb(port + SPC_TEMP); /* read */
    outb(port + SPC_SCMD, SPC_SCMD_RESETACK);
    return retval;
}

static int
spc_byte_out(int port, int data, int phase)
{
    outb(port + SPC_PCTL, phase);
    outb(port + SPC_TEMP, data); /* write */
    outb(port + SPC_SCMD, SPC_SCMD_SETACK);
    if( spc_pollport(port+SPC_PSNS, SPC_PSNS_REQ, 0) < 0 ) {
        spc_dumpreg(port, "spc_byte_out");
        return -1;
    }
    outb(port + SPC_SCMD, SPC_SCMD_RESETACK);
    return 0;
}

static int
spc_init(struct spc_data *spc)
{
    int		port = spc->baseport;
    long	count = SPC_TIMEOUT;

    outb(port + SPC_SCTL, 0x9a);		/* SPC software reset */
    outb(port + SPC_BDID, SPC_HOST_ID);		/* set initiator id */
    outb(port + SPC_SCMD, 0);
    outb(port + SPC_PCTL, 0);
    outb(port + SPC_TEMP, 0);
    outb(port + SPC_TCH , 0);
    outb(port + SPC_TCM , 0);
    outb(port + SPC_TCL , 0);
    outb(port + SPC_PSNS, 0);
    DELAY(100000);				/* wait 100ms */
    outb(port + SPC_SCTL, SPC_SCTL_DISABLE);	/* clear SPC reset */
    outb(port + SPC_SCMD, SPC_SCMD_RESET);	/* RESET SCSI bus */
    DELAY(100000);				/* wait 100ms */
    outb(port + SPC_SCMD, 0);			/* RESET SCSI bus */
    DELAY(5000000);				/* wait 5s */
    outb(port + SPC_SCMD, SPC_SCMD_BUSRELEASE);

    /* confirm bus release ... */
    while (inb(port + SPC_PSNS) && count) count--;
    if( !count ) {
        printf("spc%d: init failed, driver disabled\n", spc->unit);
	spc->alive = 0;
	return 0;
    }
    spc->alive = 1;
    return 1;
}

static void
spc_cleanup(struct spc_data *spc)
{
    int			id, lun;
    struct spc_queue	*sq;
    struct scsi_xfer	*sxs;

    printf("spc%d: cleanup current id=%d current lun=%d state=%d\n",
	   spc->unit, spc->target, spc->lun, spc->state);
    for(id = 0; id < NTARGETS; id++) {
	for(lun = 0; lun < NLUNS; lun++) {
	    if( (sq = spc->busy[id][lun]) == NULL ) {
		continue;
	    }
	    printf("spc%d: cleanup id=%d lun=%d\n", spc->unit, id, lun);
	    if( sq->flags & SQ_FLG_RSNS ) {
		sxs = sq->xs;
		if( sq->alt_xs == NULL ) {
		    continue;
		}
		sq->xs = sq->alt_xs;
		if( sxs ) {
		    spc_free_rsns(sxs);
		}
	    }
    	    sq->xs->flags |= ITSDONE;
    	    sq->xs->error = XS_TIMEOUT;
            scsi_done(sq->xs);
	    spc_free_queue(spc, sq);
	    spc->busy[id][lun] = NULL;
	}
    }
}
