/*	$NetBSD$	*/
/* $Id: ncr53c500.c,v 1.1.6.1 1998/07/27 10:50:20 kim Exp $ */
/*
 * [NetBSD for NEC PC98 series]
 *  Copyright (c) 1995, 1996 NetBSD/pc98 porting staff.
 *  Copyright (c) 1995, 1996 Naofumi HONDA.
 *  All rights reserved.
 * 
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef __NetBSD__
#include <dev/isa/ncr53c500if.h>
#endif /* __NetBSD__ */
#ifdef __FreeBSD__
#include <i386/scsi/ncr53c500/ncr53c500if.h>
#endif /* __FreeBSD__ */

/***************************************************
 * DEBUG
 ***************************************************/
#ifndef DDB
#define Debugger() panic("should call debugger here (ncr53c500.c)")
#endif /* ! DDB */

#define	NCV_DEBUG
#ifdef	NCV_DEBUG
int ncv_debug;

struct ncv_statics {
	int disconnect;
	int reselect;
} ncv_statics[NTARGETS];
#endif	/* NCV_DEBUG */

/* static */
static void ncv_pio_read __P((struct ncv_softc *, u_int8_t *, u_int));
static void ncv_pio_write __P((struct ncv_softc *, u_int8_t *, u_int));
static int ncv_synch __P((struct ncv_softc *, struct targ_info *, u_int, u_int));
static inline int ncv_reselected __P((struct ncv_softc *));
static inline int ncv_disconnected __P((struct ncv_softc *, struct ccb *cb));
static inline void ncv_pdma_end __P((struct ncv_softc *sc, struct targ_info *)); 
static inline void ncvhw_set_count __P((struct ncv_softc *sc, int));
static inline u_int ncvhw_get_count __P((struct ncv_softc *sc));
static inline void ncvhw_select_register_0 __P((struct ncv_softc *));
static inline void ncvhw_select_register_1 __P((struct ncv_softc *));
static inline void ncvhw_fpush __P((struct ncv_softc *, u_int8_t *, int));

static int ncv_world_start __P((struct ncv_softc *, int));
static void ncvhw_bus_reset __P((struct ncv_softc *));
static void ncvhw_reset __P((struct ncv_softc *));
static void ncvhw_init __P((struct ncv_softc *));
static void ncvhw_synch __P((struct ncv_softc *sc, struct targ_info *));
static int ncvhw_power __P((struct ncv_softc *, u_int));
static int ncvhw_start_selection __P((struct ncv_softc *sc, struct ccb *));
static void ncvhw_attention __P((struct ncv_softc *));

struct scsi_low_funcs ncv_funcs =
{
	SC_LOW_RESET_T ncv_world_start,
	SC_LOW_BUSRST_T ncvhw_bus_reset,

	SC_LOW_SELECT_T ncvhw_start_selection,

	SC_LOW_ATTEN_T ncvhw_attention,
	SC_LOW_SYNCH_T ncv_synch,

	SC_LOW_INTR_T ncv_sequencer,

	NULL,	/* SC_LOW_POWER_T ncvhw_power, */
};

/**************************************************************
 * hwfuncs
 **************************************************************/
static inline void
ncvhw_select_register_0(sc)
	struct ncv_softc *sc;
{
	DECLARE_IOPORT(sc);

	OUTB(cr0_cfg4, sc->sc_hw.cfg4);
}

static inline void
ncvhw_select_register_1(sc)
	struct ncv_softc *sc;
{
	DECLARE_IOPORT(sc);

	OUTB(cr1_cfg5, sc->sc_hw.cfg5);
}

static inline void
ncvhw_fpush(sc, buf, len)
	struct ncv_softc *sc;
	u_int8_t *buf;
	int len;
{
	int ptr;
	DECLARE_IOPORT(sc);

	for (ptr = 0; ptr < len; ptr ++)
		OUTB(cr0_sfifo, buf[ptr]);
}

int
ncvhw_check(sc)
	struct ncv_softc *sc;
{
	DECLARE_IOPORT(sc);

	ncvhw_select_register_0(sc);
	OUTB(cr0_cmd, CMD_NOP | CMD_DMA);
	if (INB(cr0_cmd) != (CMD_NOP | CMD_DMA))
		return ENODEV;

	OUTB(cr0_cmd, CMD_NOP);
	if (INB(cr0_cmd) != CMD_NOP)
		return ENODEV;

	return 0;
}

static void
ncvhw_reset(sc)
	struct ncv_softc *sc;
{
	DECLARE_IOPORT(sc);

	ncvhw_select_register_0(sc);

	/* dummy cmd twice */
	OUTB(cr0_cmd, CMD_NOP);
	OUTB(cr0_cmd, CMD_NOP);

	/* chip reset */
	OUTB(cr0_cmd, CMD_RSTCHIP);

	/* again dummy cmd twice */
	OUTB(cr0_cmd, CMD_NOP);
	OUTB(cr0_cmd, CMD_NOP);
}

static void
ncvhw_init(sc)
	struct ncv_softc *sc;
{
	DECLARE_IOPORT(sc);

	ncvhw_select_register_0(sc);
	OUTB(cr0_clk, sc->sc_hw.clk);
	OUTB(cr0_srtout, SEL_TOUT);
	OUTB(cr0_period, 0);
	OUTB(cr0_offs, 0);

	OUTB(cr0_cfg1, sc->sc_hw.cfg1);
	OUTB(cr0_cfg2, sc->sc_hw.cfg2);
	OUTB(cr0_cfg3, sc->sc_hw.cfg3);
	OUTB(cr0_tchsb, 0);

	ncvhw_select_register_1(sc);
	OUTB(cr1_fstat, 0x0);
	OUTB(cr1_pflag, 0x0);
	OUTB(cr1_atacmd, ATACMD_ENGAGE);

	ncvhw_select_register_0(sc);
}

static void
ncvhw_synch(sc, ti)
	struct ncv_softc *sc;
	struct targ_info *ti;
{
	DECLARE_IOPORT(sc);

	/* start selection */
	if (ti->ti_flags & SCSI_LOW_NOPARITY)
		OUTB(cr0_cfg1, sc->sc_hw.cfg1);
	else
		OUTB(cr0_cfg1, sc->sc_hw.cfg1 | C1_PARENB);
	OUTB(cr0_period, ti->ti_synch.period);
	OUTB(cr0_offs, ti->ti_synch.offset);
	OUTB(cr0_cfg3, ti->ti_synch.c3img);
}

static int
ncvhw_power(sc, flags)
	struct ncv_softc *sc;
	u_int flags;
{
	DECLARE_IOPORT(sc);

	if (flags == SCSI_LOW_POWDOWN)
	{
		printf("%s power down\n", sc->sc_dvname);
		ncvhw_select_register_1(sc);
		OUTB(cr1_atacmd, ATACMD_POWDOWN);
	}
	else
	{
		switch (sc->sc_rstep)
		{
		case 0:
			printf("%s resume step O\n", sc->sc_dvname);
			ncvhw_select_register_1(sc);
			OUTB(cr1_atacmd, ATACMD_ENGAGE);
			break;

		case 1:
			printf("%s resume step I\n", sc->sc_dvname);
			ncvhw_reset(sc);
			ncvhw_init(sc);
			break;
		}
	}

	return 0;
}

/**************************************************************
 * scsi low interface
 **************************************************************/
static void
ncvhw_attention(sc)
	struct ncv_softc *sc;
{
	DECLARE_IOPORT(sc);

	OUTB(cr0_cmd, CMD_SETATN);
	delay(200);
}

static void
ncvhw_bus_reset(sc)
	struct ncv_softc *sc;
{
	DECLARE_IOPORT(sc);

	ncvhw_select_register_0(sc);
	OUTB(cr0_cmd, CMD_FLUSH);
	OUTB(cr0_cmd, CMD_RSTSCSI);
	OUTB(cr0_cmd, CMD_NOP | CMD_DMA);
}

static int
ncvhw_start_selection(sc, cb)
	struct ncv_softc *sc;
	struct ccb *cb;
{
	struct targ_info *ti = cb->ti;
	int s;
	DECLARE_IOPORT(sc);

	ID_MSG_SETUP(ti, cb);
	ncvhw_select_register_0(sc);

	s = splhigh();

	if (sc->sc_disc > 0 && (INB(cr0_stat) & STAT_INT))
	{
		splx(s);
		return 1;
	}

	ncvhw_synch(sc, ti);
	OUTB(cr0_dstid, ti->ti_id);
	OUTB(cr0_cmd, CMD_FLUSH);

	OUTB(cr0_sfifo, ti->ti_msgout);

	if (cb->msgoutlen)
		OUTB(cr0_cmd, CMD_SELATNS);
	else
	{
		ncvhw_fpush(sc, cb->cmd, cb->cmdlen);
		OUTB(cr0_cmd, CMD_SELATN);
	}

	splx(s);

	SCSI_LOW_SETUP_PHASE(PH_SELSTART);
	return 0;
}

static int
ncv_world_start(sc, fdone)
	struct ncv_softc *sc;
	int fdone;
{
	struct targ_info *ti;
	u_int8_t stat;
	DECLARE_IOPORT(sc);

	if (sc->sc_flags & HW_INACTIVE)
		return EBUSY;

	/* reset current nexus */
	scsi_low_reset_nexus((struct scsi_low_softc *) sc, fdone);

	/* hardware info init */
	for (ti = sc->sc_titab.tqh_first; ti; ti = ti->ti_chain.tqe_next)
	{
		ti->ti_state = UNIT_RDY;
		ti->ti_synch.period = ti->ti_synch.offset = 0;
		ti->ti_synch.c3img = sc->sc_hw.cfg3;
		ti->ti_maxsynch.c3img = sc->sc_hw.cfg3;
		ti->ti_maxsynch.period = sc->sc_hw.mperiod;
		ti->ti_maxsynch.offset = sc->sc_hw.moffset;

		scsi_low_calcf(ti);
	}

	ncvhw_reset(sc);
	ncvhw_init(sc);
	scsi_low_bus_reset((struct scsi_low_softc *) sc);

	ncvhw_select_register_0(sc);
	INB(cr0_stat);
	stat = INB(cr0_istat);
	if ((stat & INTR_SBR) == 0)
		return ENODEV;
	delay(1000);
	stat = INB(cr0_istat);
	if (stat & INTR_SBR)
		return ENODEV;

#ifdef __NetBSD__
	softintr(sc->sc_mask);
#endif
	return 0;
}

static int
ncv_synch(sc, ti, period, offset)
	struct ncv_softc *sc;
	struct targ_info *ti;
	u_int period, offset;
{
	u_int hwperiod;

	hwperiod = 1000 / ((sc->sc_hw.clk == 0) ? 40 : (5 * sc->sc_hw.clk));

	if (period < 200 / 4 && period >= 100 / 4)
		ti->ti_synch.c3img |= C3_FSCSI;
	else
		ti->ti_synch.c3img &= ~C3_FSCSI;

	period = ((period * 40 / hwperiod) + 5) / 10;

	ti->ti_synch.period = period & 0x1f;
	ti->ti_synch.offset = offset;

	return 0;
}

/**************************************************************
 * PDMA
 **************************************************************/
static inline void
ncvhw_set_count(sc, count)
	struct ncv_softc *sc;
	int count;
{
	DECLARE_IOPORT(sc);

	OUTB(cr0_tclsb, (u_int8_t) count);
	OUTB(cr0_tcmsb, (u_int8_t) (count >> NBBY));
	OUTB(cr0_tchsb, (u_int8_t) (count >> (NBBY * 2)));
}

static inline u_int
ncvhw_get_count(sc)
	struct ncv_softc *sc;
{
	u_int count;
	DECLARE_IOPORT(sc);

	count = (u_int) INB(cr0_tclsb);
	count |= ((u_int) INB(cr0_tcmsb)) << NBBY;
	count |= ((u_int) INB(cr0_tchsb)) << (NBBY * 2);
	return count;
}

static inline void
ncv_pdma_end(sc, ti)
	struct ncv_softc *sc;
	struct targ_info *ti;
{
	int len;
	DECLARE_IOPORT(sc);

	sc->sc_flags &= ~HW_PDMASTART;
	if (ti->ti_phase == PH_DATA)
	{
		len = ncvhw_get_count(sc);
		if ((sc->sc_direction & SCSI_LOW_READ) == 0)
			len += (INB(cr0_sffl) & 0x1f);

		if ((u_int) len <= (u_int) sc->sc_scp.datalen)
		{
			sc->sc_scp.data += (sc->sc_scp.datalen - len);
			sc->sc_scp.datalen = len;
			if ((sc->sc_direction & SCSI_LOW_READ) &&
			    sc->sc_tdatalen != len)
				goto bad;
		}
		else
		{
bad:
			ti->ti_error |= PDMAERR;
			printf("%s strange count hw 0x%x soft 0x%x tlen 0x%x\n",
				sc->sc_dvname, len, sc->sc_scp.datalen,
				sc->sc_tdatalen);
		}
	}
	else
	{
		printf("%s data phase miss\n", sc->sc_dvname);
		ti->ti_error |= PDMAERR;
	}

	ncvhw_select_register_1(sc);
	OUTB(cr1_fstat, 0);
	ncvhw_select_register_0(sc);
}

static void
ncv_pio_read(sc, buf, reqlen)
	struct ncv_softc *sc;
	u_int8_t *buf;
	u_int reqlen;
{
	int tout = sc->sc_wc;
	register int len;
	register u_int8_t fstat;
	DECLARE_IOPORT(sc);

	ncvhw_select_register_1(sc);
	OUTB(cr1_pflag, 0);

	ncvhw_select_register_0(sc);
	ncvhw_set_count(sc, reqlen);
	OUTB(cr0_cmd, CMD_TRANS | CMD_DMA);

	ncvhw_select_register_1(sc);
	OUTB(cr1_fstat, FIFO_EN);
	sc->sc_flags |= HW_PDMASTART;

	while (reqlen >= FIFO_F_SZ && tout > 0)
	{
		fstat = INB(cr1_fstat);
		if (fstat & FIFO_F)
		{
			INSL(cr1_fdata, buf, FIFO_F_SZ / 4);
			buf += FIFO_F_SZ;
			reqlen -= FIFO_F_SZ;
			continue;
		}
		else if (fstat & FIFO_BRK)
			break;

		tout --;
	}

	if (reqlen >= FIFO_2_SZ)
	{
		fstat = INB(cr1_fstat);
		if (fstat & FIFO_2)
		{
			INSL(cr1_fdata, buf, FIFO_2_SZ / 4);
			buf += FIFO_2_SZ;
			reqlen -= FIFO_2_SZ;
		}
	}

	while (reqlen > 0 && tout > 0)
	{
		fstat = INB(cr1_fstat);
		if ((fstat & FIFO_E) == 0)
		{
			*buf++ = INB(cr1_fdata);
			reqlen --;
			continue;
		}
		else if (fstat & FIFO_BRK)
			break;
		tout --;
	}

	ncvhw_select_register_0(sc);
	sc->sc_tdatalen = reqlen;

	if (tout <= 0)
		printf("%s pio read timeout\n", sc->sc_dvname);
}

static void
ncv_pio_write(sc, buf, reqlen)
	struct ncv_softc *sc;
	u_int8_t *buf;
	u_int reqlen;
{
	int tout = sc->sc_wc;
	register u_int8_t fstat;
	DECLARE_IOPORT(sc);

	ncvhw_select_register_1(sc);
	OUTB(cr1_pflag, 0);

	ncvhw_select_register_0(sc);
	ncvhw_set_count(sc, reqlen);
	OUTB(cr0_cmd, CMD_TRANS | CMD_DMA);

	ncvhw_select_register_1(sc);
	OUTB(cr1_fstat, FIFO_EN);
	sc->sc_flags |= HW_PDMASTART;

	while (reqlen >= FIFO_F_SZ && tout > 0)
	{
		fstat = INB(cr1_fstat);
		if (fstat & FIFO_BRK)
			goto done;

		if (fstat & FIFO_E)
		{
			OUTSL(cr1_fdata, buf, FIFO_F_SZ / 4);
			buf += FIFO_F_SZ;
			reqlen -= FIFO_F_SZ;
		}
		else
			tout --;
	}

	while (reqlen > 0 && tout > 0)
	{
		fstat = INB(cr1_fstat);
		if (fstat & FIFO_BRK)
			break;

		if ((fstat & FIFO_F) == 0) /* fifo not full */
		{
			OUTB(cr1_fdata, *buf++);
			reqlen --;
		}
		else
			tout --;
	}

done:
	ncvhw_select_register_0(sc);

	if (tout <= 0)
		printf("%s pio write timeout\n", sc->sc_dvname);
}

/**************************************************************
 * disconnect & reselect (HW low)
 **************************************************************/
static inline int
ncv_reselected(sc)
	struct ncv_softc *sc;
{
	struct ccb *cb;
	u_int sid, imsg;
	DECLARE_IOPORT(sc);

	if ((INB(cr0_sffl) & 0x1f) != 2)
	{
		printf("%s illegal fifo bytes\n", sc->sc_dvname);
		return 0;
	}

	sid = (u_int) INB(cr0_sfifo);
	imsg = (u_int) INB(cr0_sfifo);
	if ((imsg & 0x80) == 0)
	{
		printf("%s lost msg identify\n", sc->sc_dvname);
		return 0;
	}

	sid = ffs(sid) - 1;
	cb = scsi_low_reselected((struct scsi_low_softc *) sc, sid, imsg & 0x7);
	if (cb == NULL)
		return 0;

#ifdef	NCV_DEBUG
	ncv_statics[sid].reselect ++;
#endif
	ncvhw_synch(sc, cb->ti);
	OUTB(cr0_dstid, sid);
	OUTB(cr0_cmd, CMD_MSGOK);

	return 1;
}

static inline int
ncv_disconnected(sc, cb)
	struct ncv_softc *sc;
	struct ccb *cb;
{
	DECLARE_IOPORT(sc);

	OUTB(cr0_cmd, CMD_FLUSH);
	OUTB(cr0_cfg1, sc->sc_hw.cfg1);
	OUTB(cr0_cmd, CMD_ENSEL);

#ifdef	NCV_DEBUG
	if (sc->sc_msgphase == DISCASSERT)
		ncv_statics[cb->ti->ti_id].disconnect ++;
#endif	/* NCV_DEBUG */

	scsi_low_disconnected((struct scsi_low_softc *) sc, cb);
	scsi_low_start((struct scsi_low_softc *) sc);

	return 1;
}

/**************************************************************
 * SEQUENCER
 **************************************************************/
int
ncv_sequencer(sc)
	struct ncv_softc *sc;
{
	struct targ_info *ti;
	struct ccb *cb;
	int len;
	u_int8_t status, ireason;
	DECLARE_IOPORT(sc);

	if (sc->sc_flags & HW_INACTIVE)
		return 0;

	/********************************************
	 * Status
	 ********************************************/
	ncvhw_select_register_0(sc);
	status = INB(cr0_stat);
	if ((status & STAT_INT) == 0)
		return 0;

	ireason = INB(cr0_istat);
	if (ireason & INTR_SBR)
	{
		u_int8_t val;

		/* avoid power off hangup */
		val = INB(cr0_cfg1);
		OUTB(cr0_cfg1, val | C1_SRR);

		/* status init */
		ti = sc->sc_titab.tqh_first;
		for ( ; ti; ti = ti->ti_chain.tqe_next)
			ti->ti_state = UNIT_RDY;

		printf("%s bus reset (power off ?)\n", sc->sc_dvname);
		return 1;
	}

	/********************************************
	 * Debug section
	 ********************************************/
#ifdef	NCV_DEBUG
	if (ncv_debug)
	{
		scsi_low_print((struct scsi_low_softc *) sc, NULL);
		printf("%s st %x ist %x\n\n", sc->sc_dvname,
			status, ireason);
		if (ncv_debug > 1)
			Debugger();
	}
#endif	/* NCV_DEBUG */

	/********************************************
	 * Reselect or Disconnect or Nexus check
	 ********************************************/
	/* (I) reselect */
	if (ireason == INTR_RESELECT)
	{
		if ((status & PHASE_MASK) == MESSAGE_IN_PHASE)
			return ncv_reselected(sc);

		return 1;
	}

	/* (II) nexus */
	if ((cb = sc->sc_nexus) == NULL)
		return 0;

	/********************************************
	 * Internal scsi phase
	 ********************************************/
	ti = cb->ti;
	switch (ti->ti_phase)
	{
	case PH_SELSTART:
		sc->sc_selid = NULL;

		if ((ireason & (INTR_SBR | INTR_DIS | INTR_ILL)) ||
		    (status & (STAT_PE | STAT_GE)))
			goto abort;

		if (cb->msgoutlen == 0)
		{
			if ((status & PHASE_MASK) != MESSAGE_OUT_PHASE)
				break;

			SCSI_LOW_SETUP_PHASE(PH_SELECTED);
			printf("%s msg identify failed\n", sc->sc_dvname);
		}
		else
		{
			if ((status & PHASE_MASK) != MESSAGE_OUT_PHASE)
			{
				ti->ti_error |= FATALIO;
				scsi_low_restart((struct scsi_low_softc *) sc,
					 "msgout error");
				return 1;
			}

			SCSI_LOW_SETUP_PHASE(PH_MSGOUTEXP);
		}
		break;

	default:
		if (sc->sc_selid)
		{
			scsi_low_print((struct scsi_low_softc *) sc, ti);
			panic("%s internal error\n", sc->sc_dvname);
		}

		if (sc->sc_flags & HW_PDMASTART)
			ncv_pdma_end(sc, ti);

abort:
		if (status & (STAT_PE | STAT_GE))
		{
			ti->ti_error |= PARITYERR;
			scsi_low_restart((struct scsi_low_softc *) sc,
					 "scsi bus error");
			return 1;
		}

		if (ireason & (INTR_DIS | INTR_ILL))
		{
			if (ireason & INTR_ILL)
			{
				ti->ti_error |= FATALIO;
				scsi_low_restart((struct scsi_low_softc *) sc,
					 "illegal cmd");
				return 1;
			}

			return ncv_disconnected(sc, cb);
		}
	}

	/********************************************
	 * Scsi phase sequencer
	 ********************************************/
	switch (status & PHASE_MASK)
	{
	case DATA_OUT_PHASE: /* data out */
		SCSI_LOW_SETUP_PHASE(PH_DATA);
		sc->sc_direction = SCSI_LOW_WRITE;
		ncv_pio_write(sc, sc->sc_scp.data, sc->sc_scp.datalen);
		break;

	case DATA_IN_PHASE: /* data in */
		SCSI_LOW_SETUP_PHASE(PH_DATA);
		sc->sc_direction = SCSI_LOW_READ;
		ncv_pio_read(sc, sc->sc_scp.data, sc->sc_scp.datalen);
		break;

	case COMMAND_PHASE: /* cmd out */
		SCSI_LOW_SETUP_PHASE(PH_CMD);
		OUTB(cr0_cmd, CMD_FLUSH);
		ncvhw_fpush(sc, cb->cmd, cb->cmdlen);
		OUTB(cr0_cmd, CMD_TRANS);
		break;

	case STATUS_PHASE: /* status in */
		SCSI_LOW_SETUP_PHASE(PH_COMPSEQ);
		OUTB(cr0_cmd, CMD_FLUSH);
		OUTB(cr0_cmd, CMD_ICCS);
		break;

	default:
		break;

	case MESSAGE_OUT_PHASE: /* msg out */
		OUTB(cr0_cmd, CMD_FLUSH);

		len = scsi_low_msgout((struct scsi_low_softc *) sc, ti, cb);
		if (len == 0)
		{
			OUTB(cr0_sfifo, ti->ti_msgout);
			OUTB(cr0_cmd, CMD_TRANS);
			if (cb->msgoutlen == 0)
				OUTB(cr0_cmd, CMD_RSTATN);
		}
		else if (len > 0)
		{
			ncvhw_fpush(sc, cb->msgout, len);
			OUTB(cr0_cmd, CMD_TRANS);
			OUTB(cr0_cmd, CMD_RSTATN);
		}
		break;

	case MESSAGE_IN_PHASE: /* msg in */
		len = INB(cr0_sffl) & 0x1f;
		if (ti->ti_phase == PH_COMPSEQ)
		{
			if ((ireason & INTR_FC) && len == 2)
			{
				ti->ti_status =
					INB(cr0_sfifo);
				len --;
			}
			else
			{
				scsi_low_restart((struct scsi_low_softc *) sc,
						 "compseq error");

				break;
			}
		}
		else if (ireason & INTR_BS)
		{
			OUTB(cr0_cmd, CMD_FLUSH);
			OUTB(cr0_cmd, CMD_TRANS);
			break;
		}

		SCSI_LOW_SETUP_PHASE(PH_MSGIN);

		if ((ireason & INTR_FC) && len == 1)
		{
			ti->ti_msgin[ti->ti_msginptr++] =
				INB(cr0_sfifo);
			scsi_low_msgin((struct scsi_low_softc *) sc, cb);
			OUTB(cr0_cmd, CMD_MSGOK);
		}
		else
		{
			ti->ti_error |= MSGERR;
			printf("%s st %x ist %x\n\n", sc->sc_dvname,
				status, ireason);
			scsi_low_restart((struct scsi_low_softc *) sc,
						 "hw msgin error");
		}
		break;
	}

	return 1;
}
