/* $Id: ncr53c500if.c,v 1.1.6.2 1998/09/03 15:59:28 shiba Exp $ */
/*
 * [NetBSD for NEC PC98 series]
 *  Copyright (c) 1995, 1996 NetBSD/pc98 porting staff.
 *  Copyright (c) 1995, 1996 Naofumi HONDA.
 *  All rights reserved.
 * 
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef __NetBSD__
#include <dev/isa/ncr53c500if.h>
#endif /* __NetBSD__ */
#ifdef __FreeBSD__
#include <i386/scsi/ncr53c500/ncr53c500if.h>
#include "ncv.h"
#endif /* __FreeBSD__ */

/***************************************************
 * PARAMS
 ***************************************************/
#define	NCV_MAX_CCB	(2 * (NTARGETS - 1))

/***************************************************
 * ISA DEVICE STRUCTURE
 ***************************************************/
#ifdef __NetBSD__
int ncvintr __P((void *));

struct cfdriver ncv_cd = {
	NULL, "ncv", DV_DULL
};

struct scsi_device ncv_dev = {
	NULL,	/* Use default error handler */
	NULL,	/* have a queue, served by this */
	NULL,	/* have no async handler */
	NULL,	/* Use default 'done' routine */
};

struct scsi_adapter ncv = {
	scsi_low_scsi_cmd,
	scsi_low_scsi_minphys,
	scsi_low_target_open,
	0,
};
#endif /* __NetBSD__ */

#ifdef __FreeBSD__
void ncvintr __P((int));

/* pccard support */
#if NCARD > 0
static int	ncv_card_intr(struct pccard_devinfo *);
static void	ncv_card_unload(struct pccard_devinfo *);
static int	ncv_card_resume(struct ncv_softc *);
static int	ncv_card_init(struct pccard_devinfo *);

static struct pccard_device ncv_info = {
	"ncv",
	ncv_card_init,
	ncv_card_unload,
	ncv_card_intr,
	0,				/* Attributes - presently unused */
	&bio_imask			/* Interrupt mask for device */
		/* This should also include net_imask?? */
};

DATA_SET(pccarddrv_set, ncv_info);

#ifdef  SCSI_DETACH
static void ncvdetach(struct isa_device *dev);
#endif
#endif /* NCARD */

struct isa_driver ncvdriver = {
	ncvprobe,
	ncvattach,
	"ncv"
};

struct scsi_device ncv_dev = {
	NULL,	/* Use default error handler */
	NULL,	/* have a queue, served by this */
	NULL,	/* have no async handler */
	NULL,	/* Use default 'done' routine */
	"ncv",
	0,
	{0, 0}
};

static struct scsi_adapter ncv = {
	scsi_low_scsi_cmd,
	scsi_low_scsi_minphys,
	scsi_low_target_open,
	0,
	ncv_adapter_info,
	"ncv",
	{0, 0},
};
u_int32_t ncv_adapter_info (int unit)
{
	return 1;
}

static struct ncv_softc *ncvdata[NNCV];
#endif /* __FreeBSD__ */

/**************************************************************
 * DECLARE
 **************************************************************/
extern int delaycount;
extern struct scsi_low_funcs ncv_funcs;

/**************************************************************
 * General probe attach
 **************************************************************/
#ifdef __FreeBSD__
#if NCARD > 0
static int
ncv_card_resume(struct ncv_softc *sc)
{
#if 0 /* XXX ncv_init */
	if (!ncv_init(sc))
		return (ENXIO);
#endif
	printf("%s: resumed\n", sc->sc_dvname);
	return (0);
}

static int
ncv_card_init(struct pccard_devinfo *devi)
{
	int unit = devi->pd_unit;
	struct ncv_softc *sc = ncvdata[unit];

	if (NNCV <= unit)
		return (ENODEV);
		/* XXX -- ncv_init */
	printf("probe ncv\n");
	if (ncvprobe(&devi->isahd) == 0)
		return (ENXIO);
	printf("attach ncv\n");
	if (ncvattach(&devi->isahd) == 0)
		return (ENXIO);
	return (0);
}

static void
ncv_card_unload(struct pccard_devinfo *devi)
{
	struct ncv_softc *sc = ncvdata[devi->pd_unit];

	printf("%s: unload\n", sc->sc_dvname);
/*	scsi_low_deactivate((struct scsi_low_softc *)sc); */
#ifdef  SCSI_DETACH
        ncvdetach(&devi->isahd);
#endif 
}

static int
ncv_card_intr(struct pccard_devinfo *devi)
{

	(void) ncv_sequencer(ncvdata[devi->pd_unit]);
	return 1;
}
#endif /* NCARD > 0 */
#endif /* FreeBSD */

static inline int
ncvhw_probe (struct ncv_softc *sc, int flags)
{

	if (NCV_CLKFACTOR(flags) > CLK_35M_F)
	{
		printf("%s invalid dvcfg flags\n", sc->sc_dvname);
		return (-1);
	}

	sc->sc_hw = ncv_template;
	if (NCV_C5IMG(flags) != 0)
	{
		sc->sc_hw.cfg5 = NCV_C5IMG(flags);
		sc->sc_hw.clk = NCV_CLKFACTOR(flags);

		if (NCV_SPECIAL(flags) & NCVHWCFG_MAX10M)
			sc->sc_hw.mperiod = 100 / 4;

		/* XXX:	
		 * RATOC scsi cards have fatal fifo asic bug.
		 * To avoid it, currently make sync offset 0 (async)!
		 */
		if (NCV_SPECIAL(flags) & NCVHWCFG_FIFOBUG)
		{
			sc->sc_hw.mperiod = 0;
			sc->sc_hw.moffset = 0;
		}

		if (NCV_SPECIAL(flags) & NCVHWCFG_SCSI1)
			sc->sc_hw.cfg2 &= ~C2_SCSI2;

		if (NCV_SPECIAL(flags) & NCVHWCFG_SLOW)
			sc->sc_hw.cfg1 |= C1_SLOW;
	}

	/* setup configuration image 3 */
	if (sc->sc_hw.clk != CLK_40M_F && sc->sc_hw.clk <= CLK_25M_F)
		sc->sc_hw.cfg3 &= ~C3_FCLK;

#ifdef	__FreeBSD__
	sc->sc_cfgflags = (flags & 0xffff);
#else
	sc->sc_cfgflags = DVCFG_MINOR(flags);
#endif
	sc->sc_wc = delaycount * 2000;	/* 2 sec */
	sc->sc_funcs = &ncv_funcs;

	if (ncvhw_check(sc))
		return (-1);

	return (0);
}

#ifdef __NetBSD__
int
ncvprobe(parent, self, aux)
	struct device *parent;
	struct device *self;
	void *aux;
{
	struct ncv_softc *sc = (void *)self;
	struct isa_attach_args *ia = aux;
	struct ncv_hw *hw;
	DECLARE_IOPORT(sc);

	strcpy(sc->sc_dvname, sc->sc_dev.dv_xname);
	if (ia->ia_iobase == IOBASEUNK || ia->ia_irq == IRQUNK)
		return 0;

	bc = ia->ia_bc;
	if (bus_io_map(bc, ia->ia_iobase, NCVIOSZ, &ioh))
		return 0;
	sc->sc_bc = bc;
	sc->sc_ioh = ioh;
	sc->sc_dealyioh = ia->ia_delayioh;

	if (ncvhw_probe (sc, ia->ia_cfgflags) < 0)
	{
		bus_io_unmap(bc, ioh, NCVIOSZ);
		return 0;
	}
	else if (scsi_low_attach((struct scsi_low_softc *) sc))
		return 0;
	else
	{
		ia->ia_iosize = NCVIOSZ;
		return 1;
	}
}
#endif /* __NetBSD__ */

#ifdef __FreeBSD__
int ncvprobe(struct isa_device *dev)
{
	struct ncv_softc *sc;
	int unit = dev->id_unit;
	struct ncv_hw *hw;
	char dvname[SCSI_LOW_DVNAME_LEN];
	DECLARE_IOPORT(sc);

	sprintf(dvname, "ncv%d", unit);

	if (unit >= NNCV && !dev->id_reconfig)
	{
		printf("%s: unit number too high\n", dvname);
		return (0);
	}

	if (dev->id_iobase == 0)
	{
		printf("%s: no ioaddr is given\n", dvname);
		return (0);
	}
	if (dev->id_reconfig || ncvdata[unit] != NULL)
	{
		sc = ncvdata[unit];
		sc->sc_ioport = dev->id_iobase;
		if (ncvhw_probe(sc, dev->id_flags) < 0)
			return (0);
	}
	else
	{
		sc = malloc(sizeof(struct ncv_softc), M_TEMP, M_NOWAIT);
		if (sc == NULL)
		{
			printf("%s: cannot malloc!\n", sc->sc_dvname);
			return (0);
		}
		bzero(sc, sizeof(struct ncv_softc));
		strcpy (sc->sc_dvname, dvname);
		sc->unit = unit;
		ncvdata[unit] = sc;
		sc->sc_ioport = dev->id_iobase;

		if (ncvhw_probe(sc, dev->id_flags) < 0)
		{
			free(sc, M_TEMP);
			ncvdata[unit] = NULL;
			return (0);
		}
	}
	return (NCVIOSZ);
}
#endif

int
ncvprint(aux, name)
	void *aux;
	char *name;
{

	if (name != NULL)
		printf("%s: scsibus ", name);
	return (UNCONF);
}

#ifdef __FreeBSD__
#define adapter_target adapter_targ
#define openings opennings
#endif

#ifdef __FreeBSD__
static inline void 
ncv_setup_sc(register struct ncv_softc *sc)
#else
inline void ncv_setup_sc(sc)
	register struct ncv_softc *sc;
#endif
{

	scsi_low_init_ccbque(NCV_MAX_CCB);
	TAILQ_INIT(&sc->sc_start);

	sc->sc_link.adapter_softc = sc;
	sc->sc_link.adapter_target = 7;
	sc->sc_link.openings = 2;
	sc->sc_link.adapter = &ncv;
	sc->sc_link.device = &ncv_dev;
}

#undef adapter_target
#undef openings

#ifdef __NetBSD__
void
ncvattach(parent, self, aux)
	struct device *parent;
	struct device *self;
	void *aux;
{
	struct ncv_softc *sc = (void *)self;
	struct isa_attach_args *ia = aux;

	printf("\n");
	sc->sc_bc = ia->ia_bc;
	sc->sc_dealyioh = ia->ia_delayioh;
	if (bus_io_map(sc->sc_bc, ia->ia_iobase, NCVIOSZ, &sc->sc_ioh))
		panic("%s: couldn't map io\n", sc->sc_dvname);

	ncv_setup_sc(sc);
	sc->sc_mask = (1 << ia->ia_irq);
	sc->sc_ih = isa_intr_establish(ia->ia_ic, ia->ia_irq, IST_EDGE,
	     IPL_BIO, ncvintr, sc);

	config_found(self, &sc->sc_link, ncvprint);
	timeout(scsi_low_timeout, sc, SCSI_LOW_TIMEOUT_CHECK_INTERVAL * hz);
}
#endif /* __NetBSD__ */

#ifdef __FreeBSD__

#ifdef  SCSI_DETACH
static void
ncvdetach(struct isa_device *dev)
{
        int unit = dev->id_unit;
        struct scsibus_data *scbus;
 
        scbus = (struct scsibus_data *)scsi_extend_get(ncvdata[unit]->sc_link.scsibus);
        scsi_detachdev(scbus);
}       
#endif  /* SCSI_DETACH */

int
ncvattach(dev)
	struct isa_device *dev;
{
	int unit = dev->id_unit;
	struct ncv_softc *sc = ncvdata[unit];
	struct scsibus_data *scbus;

	if (scsi_low_attach((struct scsi_low_softc *) sc))
	{
		printf("%s: scsi low attach failed\n", sc->sc_dvname);
		return (0);
	}

	ncv_setup_sc(sc);
	sc->sc_link.adapter_unit = unit;

	/*
	 * Prepare the scsibus_data area for the upperlevel
	 * scsi code.
	 */
	scbus = scsi_alloc_bus();
	if (!scbus)
		return (0);
	scbus->adapter_link = &sc->sc_link;
	/*
	 * ask the adapter what subunits are present
	 */
	scsi_attachdevs(scbus);
	timeout(scsi_low_timeout, sc, SCSI_LOW_TIMEOUT_CHECK_INTERVAL * hz);
	return 1;
}
#endif /* __FreeBSD__ */

#ifdef __NetBSD__
int ncvintr (arg)
	void *arg;
{
	return (ncv_sequencer((struct ncv_softc *)arg));
}
#endif

#ifdef __FreeBSD__
void
ncvintr (unit)
	int unit;
{
	(void) ncv_sequencer(ncvdata[unit]);
}
#endif
