/*	$NetBSD$	*/
/* $Id: scsi_low.c,v 1.1.6.1 1998/07/27 10:47:47 kim Exp $ */
/*
 * [NetBSD for NEC PC98 series]
 *  Copyright (c) 1995, 1996 NetBSD/pc98 porting staff.
 *  Copyright (c) 1995, 1996 Naofumi HONDA.
 *  All rights reserved.
 * 
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef __NetBSD__
#include <dev/isa/scsi_low_if.h>
#endif /* __NetBSD__ */
#ifdef __FreeBSD__
#include <i386/scsi/scsi_low/scsi_low_if.h>
#endif /* __FreeBSD__ */

/**************************************************************
 * power control
 **************************************************************/
#ifdef	SCSI_LOW_POWFUNC
static void scsi_low_recover __P((void *));

static void
scsi_low_recover(arg)
	void *arg;
{
	struct scsi_low_softc *sc = arg;
	int s = splbio();

	switch (sc->sc_rstep)
	{
	case 0:
		sc->sc_rstep = 1;
		(*sc->sc_funcs->scsi_low_power) (sc, SCSI_LOW_ENGAGE);
		timeout(scsi_low_recover, sc, 1);
		break;

	case 1:
		sc->sc_flags &= ~HW_RESUME;
		scsi_low_start(sc);
		break;
	}

	splx(s);
}
#endif	/* SCSI_LOW_POWFUNC */

int
scsi_low_reset(sc, flags)
	struct scsi_low_softc *sc;
	u_int flags;
{

#ifdef	SCSI_LOW_POWFUNC
	untimeout(scsi_low_recover, sc);
	sc->sc_flags &= ~(HW_POWDOWN | HW_RESUME);

	sc->sc_oc = 1;
	sc->sc_powc = SCSI_LOW_POWDOWN_TC;
#endif	/* SCSI_LOW_POWFUNC */

	return ((*sc->sc_funcs->scsi_low_reset)(sc, flags));
}

/**************************************************************
 * allocate targ_info
 **************************************************************/
struct targ_info *
scsi_low_alloc_ti(sc, targ)
	struct scsi_low_softc *sc;
	int targ;
{
	struct targ_info *ti;

	if (sc->sc_titab.tqh_first == NULL)
		TAILQ_INIT(&sc->sc_titab);

	ti = malloc(sizeof(struct targ_info), M_DEVBUF, M_NOWAIT);
	if (ti == NULL)
		panic("%s short of memory\n", sc->sc_dvname);

	bzero(ti, sizeof(struct targ_info));
	ti->ti_id = targ;
	ti->ti_sc = sc;
	ti->ti_quirks = 0;
#ifdef SDEV_NOPARITY
	ti->ti_quirks |= SDEV_NOPARITY;
#endif
#ifdef SDEV_NODISC
	ti->ti_quirks |= SDEV_NODISC;
#endif
	ti->ti_cfgflags = 0xffff0000 | SCSI_LOW_SYNC;

	sc->sc_ti[targ] = ti;
	TAILQ_INSERT_TAIL(&sc->sc_titab, ti, ti_chain);

	return ti;
}

void
scsi_low_free_ti(sc)
	struct scsi_low_softc *sc;
{
	struct targ_info *ti, *tib;

	for (ti = sc->sc_titab.tqh_first; ti; ti = tib)
	{
		tib = ti->ti_chain.tqe_next;
		free(ti, M_DEVBUF);
	}
}

int
scsi_low_attach(sc)
	struct scsi_low_softc *sc;
{
	int i;

	for (i = 0; i < NTARGETS; i ++)
		scsi_low_alloc_ti((struct scsi_low_softc *) sc, i);

	if (scsi_low_reset(sc, 0))
	{
		scsi_low_free_ti((struct scsi_low_softc *) sc);
		return EINVAL;
	}

	return 0;
}

/**************************************************************
 * timeout
 **************************************************************/
void
scsi_low_timeout(arg)
	void *arg;
{
	struct scsi_low_softc *sc = arg;
	struct ccb *cb;
	int reset, s = splbio();

	/* check */
	reset = 0;
	if (cb = sc->sc_nexus)
	{
		if ((cb->tc -= SCSI_LOW_TIMEOUT_CHECK_INTERVAL) < 0)
		{
			cb->ti->ti_error |= TIMEOUTIO;
			reset = 1;
		}
	}
	else if (sc->sc_disc > 0)
	{
		struct targ_info *ti;

		for (ti = sc->sc_titab.tqh_first; ti; ti = ti->ti_chain.tqe_next)
		{
			if ((cb = ti->ti_disccb) == NULL)
				continue;

			if ((cb->tc -= SCSI_LOW_TIMEOUT_CHECK_INTERVAL) < 0)
			{
				ti->ti_error |= TIMEOUTIO;
				reset = 1;
			}
		}
	}
	else
	{
		cb = sc->sc_start.tqh_first;
		if (cb)
		{
			if ((cb->tc -= SCSI_LOW_TIMEOUT_CHECK_INTERVAL) < 0)
			{
				cb->ti->ti_error |= TIMEOUTIO;
				reset = 1;
			}
		}
#ifdef	SCSI_LOW_POWFUNC
		else
		{
			if (sc->sc_oc == 0)
			{
				sc->sc_powc --;
				if (sc->sc_powc < 0)
				{
					sc->sc_powc = SCSI_LOW_POWDOWN_TC;

					if ((sc->sc_flags & (HW_POWDOWN | HW_RESUME)) == 0)
					{
						sc->sc_flags |= HW_POWDOWN;
						if (sc->sc_funcs->scsi_low_power)
							(*sc->sc_funcs->scsi_low_power) (sc, SCSI_LOW_POWDOWN);
					}
				}
			}
			else
			{
				sc->sc_powc = SCSI_LOW_POWDOWN_TC;
				sc->sc_oc = 0;
			}
		}
#endif	/* SCSI_LOW_POWFUNC */
	}

	/* try to recover */
	if (reset)
	{
		struct targ_info *ti;

		printf("%s: scsi bus hangup\n", sc->sc_dvname);
		for (ti = sc->sc_titab.tqh_first; ti; ti = ti->ti_chain.tqe_next)
			scsi_low_print(sc, ti);

		scsi_low_reset (sc, 1);
		scsi_low_start(sc);
	}

	timeout(scsi_low_timeout, sc, SCSI_LOW_TIMEOUT_CHECK_INTERVAL * hz);
	splx(s);
}

/**************************************************************
 * CCB
 **************************************************************/
GENERIC_CCB_STATIC_ALLOC(scsi_low, ccb)
GENERIC_CCB(scsi_low, ccb, ccb_chain)

/**************************************************************
 * SCSI INTERFACE
 **************************************************************/
#define	SCSI_LOW_MINPHYS	0x10000

void
scsi_low_scsi_minphys(bp)
	struct buf *bp;
{

	if (bp->b_bcount > SCSI_LOW_MINPHYS)
		bp->b_bcount = SCSI_LOW_MINPHYS;
	minphys(bp);
}

int
scsi_low_scsi_cmd(xs)
	struct scsi_xfer *xs;
{
	struct scsi_low_softc *sc = xs->sc_link->adapter_softc;
	struct targ_info *ti;
	struct ccb *cb;
	int s;

	if (sc->sc_cfgflags & CFG_NOATTEN)
	{
		if (xs->sc_link->lun > 0)
		{
			xs->error = XS_DRIVER_STUFFUP;
			return COMPLETE;
		}
	}

	if ((cb = scsi_low_get_ccb(xs->flags & SCSI_NOSLEEP)) == NULL)
		return TRY_AGAIN_LATER;

	cb->ti = ti = sc->sc_ti[xs->sc_link->target];

	cb->xs = xs;
	cb->lun = xs->sc_link->lun;
	cb->rcnt = 0;

	s = splbio();

	TAILQ_INSERT_TAIL(&sc->sc_start, cb, ccb_chain);
	if (sc->sc_nexus == NULL)
		scsi_low_start(sc);

	if (xs->flags & SCSI_POLL)
	{
		int timeo = xs->timeout;

		while ((xs->flags & ITSDONE) == 0 && timeo -- > 0)
		{
			delay(1000);	/* 1 ms */
			(*sc->sc_funcs->scsi_low_intr) (sc);
		}

		if (timeo <= 0 && (xs->flags & ITSDONE) == 0)
		{
			scsi_low_reset(sc, 0);
			ti->ti_error |= (TIMEOUTIO | ABORTIO);
			scsi_low_done(sc, cb);
			SCSI_LOW_SETUP_PHASE(PH_NULL);
			sc->sc_nexus = NULL;
		}

		splx(s);
		return COMPLETE;
	}

	splx(s);
	return SUCCESSFULLY_QUEUED;
}

/**************************************************************
 * Start & Done
 **************************************************************/
static struct scsi_test_unit_ready unit_ready_cmd;
static struct scsi_start_stop ss_cmd = { START_STOP, 0, {0,0,}, SSS_START, };

#define	IO_BLOCK_CHECK(sc)				\
{							\
	if (sc->sc_flags & HW_INACTIVE)			\
		return;					\
}

void
scsi_low_start(sc)
	struct scsi_low_softc *sc;
{
	struct scsi_xfer *xs;
	struct targ_info *ti;
	struct ccb *cb;
	int i, wc;

	IO_BLOCK_CHECK(sc)

#ifdef	SCSI_LOW_POWFUNC
	sc->sc_oc ++;

	if (sc->sc_flags & (HW_POWDOWN | HW_RESUME))
	{
		if (sc->sc_flags & HW_RESUME)
			return;

		sc->sc_flags &= ~HW_POWDOWN;
		if (sc->sc_funcs->scsi_low_power)
		{
			sc->sc_flags |= HW_RESUME;
			sc->sc_rstep = 0;
			(*sc->sc_funcs->scsi_low_power) (sc, SCSI_LOW_ENGAGE);

			timeout(scsi_low_recover, sc, 1);
			return;
		}
	}
#endif	/* SCSI_LOW_POWFUNC */

	/* setup nexus */
	if ((cb = sc->sc_nexus) == NULL)
	{
		for (cb = sc->sc_start.tqh_first; cb; cb = cb->ccb_chain.tqe_next)
			if (cb->ti->ti_phase == PH_NULL)
				break;

		if (cb == NULL)
			return;

		if (cb != sc->sc_start.tqh_first)
		{
			TAILQ_REMOVE(&sc->sc_start, cb, ccb_chain);
			TAILQ_INSERT_HEAD(&sc->sc_start, cb, ccb_chain);
		}

		sc->sc_nexus = cb;
	}

	xs = cb->xs;
	ti = cb->ti;
	cb->msgoutlen = 0;
	ti->ti_msginptr = 0;
	ti->ti_error = 0;
	SCSI_LOW_SETUP_MSGPHASE(NOMSG);

	if (ti->ti_tflags & TARG_CA)
	{
		/* XXX:
		 * ahc and ncr(PCI) drivers forget to init the sense_data
		 * buffer! send-pr!
		 */
		bzero(&sc->sc_sense, sizeof(sc->sc_sense));
#ifdef __NetBSD__
		sc->sc_sense_cmd.opcode = REQUEST_SENSE;
#else
		sc->sc_sense_cmd.op_code = REQUEST_SENSE;
#endif
		sc->sc_sense_cmd.byte2 = (cb->lun << 5);
		sc->sc_sense_cmd.length = sizeof(sc->sc_sense);
		cb->cmd = (u_int8_t *) &sc->sc_sense_cmd;
		cb->cmdlen = sizeof(sc->sc_sense_cmd);
		cb->data = (u_int8_t *) &sc->sc_sense;
		cb->datalen = sizeof(sc->sc_sense);
		cb->tcmax = 15;
	}
	else switch(ti->ti_state)
	{
	case UNIT_RDY:
		cb->cmd = (u_int8_t *) &unit_ready_cmd;
		cb->cmdlen = sizeof(unit_ready_cmd);
		cb->datalen = 0;
		cb->tcmax = 15;
		break;

	case UNIT_START:
		cb->cmd = (u_int8_t *) &ss_cmd;
		cb->cmdlen = sizeof(ss_cmd);
		cb->datalen = 0;
		cb->tcmax = 30;
		break;

	case UNIT_SYNCH:
		ti->ti_state = UNIT_OK;
		ti->ti_tflags |= TARG_SYNCH_PEND;

	case UNIT_OK:
		if ((ti->ti_tflags & TARG_SYNCH_PEND) &&
		    (xs->flags & SCSI_POLL) == 0)
		{
			ti->ti_tflags &= ~TARG_SYNCH_PEND;
			if (ti->ti_maxsynch.offset > 0)
			{
				ti->ti_state = UNIT_SYNCH;
				cb->msgout[0] = MSG_EXTEND;
				cb->msgout[1] = MSG_EXTEND_SYNCHLEN;
				cb->msgout[2] = MSG_EXTEND_SYNCHCODE;
				cb->msgout[3] = ti->ti_maxsynch.period;
				cb->msgout[4] = ti->ti_maxsynch.offset;
				cb->msgoutlen = MSG_EXTEND_SYNCHLEN + 2;
				cb->cmd = (u_int8_t *) &unit_ready_cmd;
				cb->cmdlen = sizeof(unit_ready_cmd);
				cb->datalen = 0;
				cb->tcmax = 15;
				break;
			}
		}

		cb->cmd = (u_int8_t *) xs->cmd;
		cb->cmdlen = xs->cmdlen;
		cb->data = xs->data;
		cb->datalen = xs->datalen;
		cb->tcmax = (xs->timeout >> 10);
		break;
	}

	/* timeout */
	if (cb->tcmax < SCSI_LOW_MIN_TOUT)
		cb->tcmax = SCSI_LOW_MIN_TOUT;
	cb->tc = cb->tcmax;

	/* setup data pointer */
	sc->sc_scp.data = ti->ti_sscp.data = cb->data;
	sc->sc_scp.datalen = ti->ti_sscp.datalen = cb->datalen;

	if (((*sc->sc_funcs->scsi_low_start_bus)(sc, cb)) == 0)
	{
		sc->sc_selid = ti;
	}
	else
	{
		sc->sc_selid = NULL;
		sc->sc_nexus = NULL;
		SCSI_LOW_SETUP_PHASE(PH_NULL);
	}
}

int
scsi_low_done(arg, cb)
	void *arg;
	struct ccb *cb;
{
	struct scsi_low_softc *sc = arg;
	struct scsi_xfer *xs = cb->xs;
	struct targ_info *ti = cb->ti;

	if (ti->ti_error == 0)
	{
		if (ti->ti_tflags & TARG_CA)
		{
			xs->sense = sc->sc_sense;
			xs->error = XS_SENSE;
			cb->rcnt = MAXRETRY;
		}
		else switch (ti->ti_status)
		{
		case ST_GOOD:
			if (sc->sc_scp.datalen != 0 && xs->bp)
			{
				ti->ti_error |= PDMAERR;
				xs->error = XS_DRIVER_STUFFUP;
			}
			else
				xs->error = XS_NOERROR;
			break;

		case ST_CHKCOND:
		case ST_MET:
			ti->ti_tflags |= TARG_CA;
			return 1;

		case ST_BUSY:
			ti->ti_error |= ERRBUSY;
			xs->error = XS_BUSY;
			break;

		default:
			ti->ti_error |= FATALIO;
			xs->error = XS_DRIVER_STUFFUP;
			break;
		}

		if (ti->ti_state < UNIT_OK)
		{
check:
			ti->ti_tflags &= ~TARG_CA;
			ti->ti_state ++;
			cb->rcnt = 0;
			return 1;
		}
	}
	else
	{
		if (ti->ti_phase == PH_SELSTART)
		{
			xs->error = XS_TIMEOUT;
			if (ti->ti_state == UNIT_RDY)
				ti->ti_error |= ABORTIO;
		}
		else if (ti->ti_state < UNIT_OK)
		{
			goto check;
		}
		else
			xs->error = XS_DRIVER_STUFFUP;

		if (ti->ti_error & ABORTIO)
			cb->rcnt = MAXRETRY;
	}

	/* internal retry check */
	if (xs->error == XS_NOERROR)
	{
		xs->resid = 0;
		ti->ti_tflags &= ~TARG_CA;
	}
	else
	{
		if (xs->bp && xs->error != XS_SENSE)
		{
			printf("%s: buffered io error\n", sc->sc_dvname);
			scsi_low_print(sc, ti);
		}

		ti->ti_tflags &= ~TARG_CA;
		if (cb->rcnt ++ < MAXRETRY)
			return 1;
	}

	xs->flags |= ITSDONE;
	scsi_done(xs);

	/* free our target */
	TAILQ_REMOVE(&sc->sc_start, cb, ccb_chain);
	scsi_low_free_ccb(cb);
	return 0;
}

/**************************************************************
 * Reset
 **************************************************************/
void
scsi_low_reset_nexus(sc, fdone)
	struct scsi_low_softc *sc;
	int fdone;
{
	struct targ_info *ti;
	struct ccb *cb;

	/* current nexus */
	if (fdone && (cb = sc->sc_nexus) && cb->rcnt ++ >= MAXRETRY)
	{
		cb->ti->ti_error |= FATALIO;
		scsi_low_done(sc, cb);
	}

	/* disconnected nexus */
	for (ti = sc->sc_titab.tqh_first; ti; ti = ti->ti_chain.tqe_next)
	{
		if (ti->ti_phase == PH_DISC)
		{
			cb = ti->ti_disccb;
			ti->ti_disccb = NULL;
			TAILQ_INSERT_HEAD(&sc->sc_start, cb, ccb_chain);
			if (fdone && cb->rcnt ++ >= MAXRETRY)
			{
				ti->ti_error |= FATALIO;
				scsi_low_done(sc, cb);
			}
		}

		SCSI_LOW_SETUP_PHASE(PH_NULL);
		ti->ti_tflags &= ~TARG_CA;
	}

	sc->sc_flags &= ~HW_PDMASTART;
	SCSI_LOW_SETUP_MSGPHASE(NOMSG);
	sc->sc_disc = 0;
	sc->sc_selid = NULL;
	sc->sc_nexus = NULL;
}

/* misc */
static int tw_pos;
static char tw_chars[] = "|/-\\";

void
scsi_low_twiddle_wait(void)
{

#ifdef __NetBSD__
	cnputc('\b');
	cnputc(tw_chars[tw_pos++]);
#endif
	tw_pos %= (sizeof(tw_chars) - 1);
	delay(TWIDDLEWAIT);
}

void
scsi_low_bus_reset(sc)
	struct scsi_low_softc *sc;
{
	int i;

	(*sc->sc_funcs->scsi_low_bus_reset) (sc);

	printf("%s: try to reset scsi bus  ", sc->sc_dvname);
	for (i = 0; i <= SCSI2_RESET_DELAY / TWIDDLEWAIT ; i++)
		scsi_low_twiddle_wait();
#ifdef __NetBSD__
	cnputc('\b');
#endif
	printf("\n");
}

int
scsi_low_restart(sc, s)
	struct scsi_low_softc *sc;
	u_char *s;
{
	int error;

	if (s)
		printf("%s: scsi bus restart. reason: %s\n",
			sc->sc_dvname, s);

	if (error = scsi_low_reset(sc, 0))
		return error;

	scsi_low_start(sc);
	return 0;
}

/**************************************************************
 * disconnect and reselect
 **************************************************************/
struct ccb *
scsi_low_reselected(sc, targ, lun)
	struct scsi_low_softc *sc;
	u_int targ;
	u_int lun;
{
	struct targ_info *ti;
	struct ccb *cb;
	u_char *s;

	/* check collision */
	if ((ti = sc->sc_selid) != NULL)
	{
		SCSI_LOW_SETUP_PHASE(PH_NULL);
		sc->sc_selid = NULL;
	}

	if (targ >= NTARGETS || targ == HOST_SCSI_ID)
	{
		s = "scsi id illegal";
		goto bad;
	}

	ti = sc->sc_ti[targ];
	if (ti->ti_phase != PH_DISC)
	{
		s = "phase mismatch";
		goto bad;
	}

	cb = ti->ti_disccb;
	if (lun != SCSI_LOW_LUNUNK && cb->lun != lun)
	{
		s = "lun mismatch";
		goto bad;
	}

	/* clear the disconnect state */
	ti->ti_disccb = NULL;
	SCSI_LOW_SETUP_PHASE(PH_RESEL);
	sc->sc_disc --;

	/* setup the nexus */
	TAILQ_INSERT_HEAD(&sc->sc_start, cb, ccb_chain);
	sc->sc_nexus = cb;

	/* recover a saved data pointer */
	sc->sc_scp = ti->ti_sscp;
	return cb;

bad:
	printf("%s: reselect %s (targ:lun) = (%x:%x)\n",
		sc->sc_dvname, s, targ, lun);
	return 0;
}

int
scsi_low_disconnected(sc, cb)
	struct scsi_low_softc *sc;
	struct ccb *cb;
{
	struct targ_info *ti = cb->ti;

	/* check phase completion */
	switch (sc->sc_msgphase)
	{
	case DISCASSERT:
		TAILQ_REMOVE(&sc->sc_start, cb, ccb_chain);
		ti->ti_disccb = cb;
		SCSI_LOW_SETUP_PHASE(PH_DISC);
		sc->sc_disc ++;
		break;

	case NOMSG:
		ti->ti_error |= FATALIO;

	case CMDCOMPLETE:
		scsi_low_done(sc, cb);
		SCSI_LOW_SETUP_PHASE(PH_NULL);
		break;
	}

	SCSI_LOW_SETUP_MSGPHASE(NOMSG);
	sc->sc_nexus = NULL;
	return 1;
}

/**************************************************************
 * msgout
 **************************************************************/
int
scsi_low_msgout(sc, ti, cb)
	struct scsi_low_softc *sc;
	struct targ_info *ti;
	struct ccb *cb;
{
	int len;

	if (ti->ti_phase == PH_MSGOUT)
	{
		if (cb->rcnt ++ < MAXRETRY)
			cb->msgoutlen = ti->ti_omsgoutlen;
	}
	else
		SCSI_LOW_SETUP_PHASE(PH_MSGOUT);

	if (ti->ti_ophase == PH_SELECTED)
	{
		if (cb->msgoutlen)
		{
			(*sc->sc_funcs->scsi_low_attention)(sc);
			SCSI_LOW_SETUP_PHASE(PH_MSGOUTEXP)
		}

		ID_MSG_SETUP(ti, cb);
		return 0;
	}

	if (cb->msgoutlen == 0)
	{
		cb->msgout[0] = MSG_REJECT;
		len = 1;
	}
	else
	{
		ti->ti_msgout = cb->msgout[0];
		ti->ti_emsgout = cb->msgout[2];
		len = cb->msgoutlen;
	}

	ti->ti_omsgoutlen = cb->msgoutlen;
	cb->msgoutlen = 0;

	return len;
}

/**************************************************************
 * msgin
 **************************************************************/
static void scsi_low_msgin_ext __P((struct targ_info *, struct ccb *));
static void scsi_low_msgin_error __P((struct targ_info *, u_int, u_int8_t));
static int scsi_low_synch __P((struct targ_info *));
static void scsi_low_msg_reject __P((struct targ_info *));

#define	PERIOD	(ti->ti_msgin[3])
#define	OFFSET	(ti->ti_msgin[4])

static int
scsi_low_synch(ti)
	struct targ_info *ti;
{
	struct scsi_low_softc *sc = ti->ti_sc;
	u_int period = 0, offset = 0, speed;
	u_char *s;

	if (PERIOD >= ti->ti_maxsynch.period &&
	    OFFSET <= ti->ti_maxsynch.offset)
	{
		if (offset = OFFSET)
			period = PERIOD;
		s = offset ? "synchronous" : "async";
	}
	else
		s = "illegal";

	(*sc->sc_funcs->scsi_low_synch) (sc, ti, period, offset);

	printf("%s(%d:0): <%s> offset %d period 0x%x chip(0x%x) ",
		sc->sc_dvname, ti->ti_id, s, offset, period,
		(u_int) ti->ti_synch.period);

	if (period)
	{
		speed = 2500 / period;
		printf("%d.%d M/s", speed / 10, speed % 10);
	}

	printf("\n");
	return 0;
}

#define	MSGWAIT(cnt)	{if(ti->ti_msginptr < (cnt)) return;}

static void
scsi_low_msgin_ext(ti, cb)
	struct targ_info *ti;
	struct ccb *cb;
{
	struct scsi_low_softc *sc = ti->ti_sc;
	int count;
	u_int reqlen;
	u_int32_t *ptr;

	MSGWAIT(2);

	reqlen = ti->ti_msgin[1];
	if (reqlen == 0)
		reqlen = 256;

	if (ti->ti_msginptr >= MAXMSGLEN)
		ti->ti_msginptr = 3;	/* XXX */

	MSGWAIT(reqlen + 2);

	switch (MKMSG_EXTEND(ti->ti_msgin[1], ti->ti_msgin[2]))
	{
	case MKMSG_EXTEND(MSG_EXTEND_MDPLEN, MSG_EXTEND_MDPCODE):
		ptr = (u_int32_t *)(&ti->ti_msgin[3]);
		count = (int) htonl((long) (*ptr));

		if(sc->sc_scp.datalen - count >= 0 &&
		   sc->sc_scp.datalen - count <= cb->datalen)
		{
			sc->sc_scp.datalen -= count;
			sc->sc_scp.data += count;
		}
		else
			scsi_low_msgin_error(ti, 7, MSG_REJECT);
		break;

	case MKMSG_EXTEND(MSG_EXTEND_SYNCHLEN, MSG_EXTEND_SYNCHCODE):
		if (cb->ti->ti_state == UNIT_SYNCH)
			scsi_low_synch(ti);
		else
		{
			(*sc->sc_funcs->scsi_low_attention)(sc);

			cb->msgout[0] = MSG_EXTEND;
			cb->msgout[1] = MSG_EXTEND_SYNCHLEN;
			cb->msgout[2] = MSG_EXTEND_SYNCHCODE;
			cb->msgout[3] = 0;
			cb->msgout[4] = 0;
			cb->msgoutlen = MSG_EXTEND_SYNCHLEN + 2;
		}
		break;

	case MKMSG_EXTEND(MSG_EXTEND_WIDELEN, MSG_EXTEND_WIDECODE):
		(*sc->sc_funcs->scsi_low_attention)(sc);

		cb->msgout[0] = MSG_EXTEND;
		cb->msgout[1] = MSG_EXTEND_WIDELEN;
		cb->msgout[2] = MSG_EXTEND_WIDECODE;
		cb->msgout[3] = 0;
		cb->msgoutlen = MSG_EXTEND_WIDELEN + 2;
		break;

	default:
		scsi_low_msgin_error(ti, 0, MSG_REJECT);
		break;
	}

	ti->ti_msginptr = 0;
	return;
}

static void
scsi_low_msg_reject(ti)
	struct targ_info *ti;
{
	struct scsi_low_softc *sc = ti->ti_sc;
	struct ccb *cb;
	char *s = "unexpected msg reject";

	if ((cb = sc->sc_nexus) == NULL)
		return;

	switch (ti->ti_ophase)
	{
	case PH_CMD:
		s = "cmd rejected";
		break;

	case PH_MSGOUT:
		if (ti->ti_msgout & 0x80)
		{
			s = "identify msg rejected";
		}
		else if (ti->ti_msgout == MSG_EXTEND)
		{
			switch (ti->ti_emsgout)
			{
			case MSG_EXTEND_SYNCHCODE:
				s = "synchrnous msg rejected";
				break;

			default:
				break;
			}
		}
		break;

	default:
		break;
	}

	printf("%s: <msgin> %s\n", sc->sc_dvname, s);
	ti->ti_error |= MSGERR;
}

static void
scsi_low_msgin_error(ti, count, msg)
	struct targ_info *ti;
	u_int count;
	u_int8_t msg;
{
	struct scsi_low_softc *sc = ti->ti_sc;
	struct ccb *cb;
	int n;

	MSGWAIT(count);

	printf("%s: illegal msg", sc->sc_dvname);
	for (n = 0; n < ti->ti_msginptr; n ++)
		printf("[0x%x] ", (u_int) ti->ti_msgin[n]);
	printf("\n");

	ti->ti_msginptr = 0;
	if ((cb = sc->sc_nexus) == NULL)
		return;

	(*sc->sc_funcs->scsi_low_attention) (sc);

	cb->msgoutlen = 1;
	cb->msgout[0] = msg;
}

void
scsi_low_msgin(sc, cb)
	struct scsi_low_softc *sc;
	struct ccb *cb;
{
	struct targ_info *ti = cb->ti;

	switch (ti->ti_msgin[0])
	{
	case MSG_SAVESP:
		ti->ti_sscp = sc->sc_scp;
		break;

	case MSG_RESTORESP:
		sc->sc_scp = ti->ti_sscp;
		break;

	case MSG_REJECT:
		scsi_low_msg_reject(ti);
		break;

	case MSG_NOOP:
		break;

	case MSG_I_ERROR:/* all I -> T : nothing to do*/
	case MSG_ABORT:
	case MSG_PARITY:
	case MSG_RESET:
	case 0xe:
	case 0xf:
		scsi_low_msgin_error(ti, 1, MSG_REJECT);
		goto resume;

	case MSG_EXTEND:
		scsi_low_msgin_ext(ti, cb);
		goto resume;

	case 0xd:
		scsi_low_msgin_error(ti, 2, MSG_REJECT);
		goto resume;

	case MSG_DISCON:
		SCSI_LOW_SETUP_MSGPHASE(DISCASSERT);
		break;

	case MSG_COMP:
		SCSI_LOW_SETUP_MSGPHASE(CMDCOMPLETE);
		break;

	case MSG_LCOMP:
	case MSG_LCOMP_F:
		scsi_low_msgin_error(ti, 1, MSG_REJECT);
		goto resume;

	default:
		if (ti->ti_msgin[0] & 0x80)
		{
			if ((ti->ti_msgin[0] & 0x07) != cb->lun)
			{
				printf("%s: lun error\n", sc->sc_dvname);
				scsi_low_msgin_error(ti, 1, MSG_ABORT);
			}
			break;
		}
		else if (ti->ti_msgin[0] < 0x20)
			scsi_low_msgin_error(ti, 1, MSG_REJECT);
		else if (ti->ti_msgin[0] < 0x30)
			scsi_low_msgin_error(ti, 2, MSG_REJECT);
		else
			scsi_low_msgin_error(ti, 1, MSG_REJECT);
		goto resume;
	}

	ti->ti_msginptr = 0;

resume:
	return;
}

#define	MAXOFFSET	0x10

void
scsi_low_calcf(ti)
	struct targ_info *ti;
{
	u_int period;
	u_int8_t offset;
	struct scsi_low_softc *sc = ti->ti_sc;

	ti->ti_flags &= ~SCSI_LOW_DISC;
#ifdef SDEV_NODISC
	if ((sc->sc_cfgflags & CFG_NODISC) == 0 &&
	    (ti->ti_quirks & SDEV_NODISC) == 0 &&
	    (ti->ti_cfgflags & SCSI_LOW_DISC) != 0)
		ti->ti_flags |= SCSI_LOW_DISC;
#endif

	ti->ti_flags |= SCSI_LOW_NOPARITY;
#ifdef SDEV_NOPARITY
	if ((sc->sc_cfgflags & CFG_NOPARITY) == 0 &&
	    (ti->ti_quirks & SDEV_NOPARITY) == 0 &&
	    (ti->ti_cfgflags & SCSI_LOW_NOPARITY) == 0)
		ti->ti_flags &= ~SCSI_LOW_NOPARITY;
#endif

	ti->ti_flags &= ~SCSI_LOW_SYNC;
	if ((ti->ti_cfgflags & SCSI_LOW_SYNC) &&
	    (sc->sc_cfgflags & CFG_ASYNC) == 0)
	{
		offset = SCSI_LOW_OFFSET(ti->ti_cfgflags);
		if (offset > ti->ti_maxsynch.offset)
			offset = ti->ti_maxsynch.offset;
		ti->ti_flags |= SCSI_LOW_SYNC;
	}
	else
		offset = 0;
	
	if (offset > 0)
	{
		period = SCSI_LOW_PERIOD(ti->ti_cfgflags);
		if (period > 100)
			period = 100;	/* MAX 10M FAST SCSI2 */
		if (period)
			period = 2500 / period;
		if (period < ti->ti_maxsynch.period)
			period = ti->ti_maxsynch.period;
	}
	else
		period = 0;

	ti->ti_maxsynch.offset = offset;
	ti->ti_maxsynch.period = period;
}

int
scsi_low_target_open(link, cf)
	struct scsi_link *link;
	struct cfdata *cf;
{
	u_int target = link->target;
	u_int lun = link->lun;
	struct scsi_low_softc *sc;
	struct targ_info *ti;
	u_int quirks, flags;

	sc = (struct scsi_low_softc *) link->adapter_softc;
	ti = sc->sc_ti[target];
	ti->ti_quirks = (u_int) link->quirks;
	ti->ti_cfgflags = cf->cf_flags;
	if (ti->ti_state > UNIT_SYNCH)
		ti->ti_state = UNIT_SYNCH;

	scsi_low_calcf(ti);

	printf("%s(%d:%d): max period(0x%x) max offset(%d) flags 0x%b\n",
		sc->sc_dvname, target, lun,
		ti->ti_maxsynch.period,
		ti->ti_maxsynch.offset,
		ti->ti_flags, SCSI_LOW_BITS);
	return 0;
}

/**********************************************************
 * DEBUG SECTION
 **********************************************************/
static u_char *phase[] =
{
	"FREE", "ARBSTART", "SELSTART", "SELECTED",
	"CMDOUT", "DATA", "MSGIN", "MSGOUT", "STATIN", "DISC",
	"COMPSEQ", "RESEL", "EXPMSGOUT"
};

void
scsi_low_print(sc, ti)
	struct scsi_low_softc *sc;
	struct targ_info *ti;
{
	struct ccb *cb = sc->sc_nexus;

	if (ti == NULL)
		ti = cb ? cb->ti : NULL;

	printf("%s <DEBUG INFO> targ 0x%x nexus 0x%x discs %d\n",
	       sc->sc_dvname, ti, cb, sc->sc_disc);

	/* target stat */
	if (ti)
	{
		struct sc_p *sp = &sc->sc_scp;
		u_int lun = cb ? cb->lun : 0;

		printf("%s(%d:%d) ph<%s> ", sc->sc_dvname,
		       ti->ti_id, lun, phase[(int) ti->ti_phase]);

		printf("msgptr %x msg[0] %x status %x tflags %x\n",
		       (u_int) (ti->ti_msginptr), (u_int) (ti->ti_msgin[0]),
		       ti->ti_status, ti->ti_tflags);

		printf("msgout %x emsgout %x ti_omsglen %x flags %b\n",
		        (u_int) ti->ti_msgout, (u_int) ti->ti_emsgout,
			ti->ti_omsgoutlen, ti->ti_flags, SCSI_LOW_BITS);

		printf("datalen %x dataaddr %x ", sp->datalen, sp->data);

		if (cb)
			printf("cmdlen %x cmdaddr %x cmd[0] %x odatalen %x",
			       cb->cmdlen, cb->cmd, (u_int) cb->cmd[0],
			       cb->datalen);

		printf("\n");
		printf("error flags %b\n", ti->ti_error, SCSI_LOW_ERRORBITS);
	}
}
