/*	$NetBSD$	*/
/* $Id: scsi_low.h,v 1.1.6.1 1998/07/27 10:47:48 kim Exp $ */
/*
 * [NetBSD for NEC PC98 series]
 *  Copyright (c) 1995, 1996 NetBSD/pc98 porting staff.
 *  Copyright (c) 1995, 1996 Naofumi HONDA.
 *  All rights reserved.
 * 
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef	_SCSI_LOW_H_
#define	_SCSI_LOW_H_

#ifdef __NetBSD__
#include <dev/isa/scsi_dvcfg.h>
#endif
#ifdef __FreeBSD__
#include <i386/scsi/scsi_low/scsi_dvcfg.h>
#endif

#define	SCSI_LOW_SYNC		DVF_SCSI_SYNC
#define	SCSI_LOW_DISC		DVF_SCSI_DISC
#define	SCSI_LOW_WAIT		DVF_SCSI_WAIT
#define	SCSI_LOW_LINK		DVF_SCSI_LINK
#define	SCSI_LOW_QTAG		DVF_SCSI_QTAG
#define	SCSI_LOW_NOPARITY	DVF_SCSI_NOPARITY
#define	SCSI_LOW_SAVESP		DVF_SCSI_SAVESP
#define	SCSI_LOW_DEFCFG		DVF_SCSI_DEFCFG
#define	SCSI_LOW_BITS		DVF_SCSI_BITS

#define	SCSI_LOW_PERIOD(XXX)	DVF_SCSI_PERIOD(XXX)
#define	SCSI_LOW_OFFSET(XXX)	DVF_SCSI_OFFSET(XXX)

#define	HOST_SCSI_ID	7

#define SCSI_LOW_MIN_TOUT	24
#define SCSI_LOW_TIMEOUT_CHECK_INTERVAL 4
#define	SCSI_LOW_POWDOWN_TC	15

#define	NTARGETS	8
#define	MAXMSGLEN	16
#define	MAXRETRY	3

struct targ_info;

struct ccb {
	TAILQ_ENTRY(ccb) ccb_chain;

	struct scsi_xfer *xs;		/* scsi upper */

	struct targ_info *ti;		/* targ_info */

	u_int lun;			/* lun */

	int rcnt;			/* retry counter */

	/*****************************************
	 * scsi cmd & data
	 *****************************************/
	u_int8_t *cmd;			/* scsi cmd */
	int cmdlen;

	u_int8_t *data;			/* scsi data */
	u_int datalen;

	u_int8_t msgout[MAXMSGLEN];	/* scsi msgout */
	u_int msgoutlen;

	/*****************************************
	 * timeout counter
	 *****************************************/
	int tc;
	int tcmax;
};

/* ccb assert */
#ifdef __NetBSD__
#include <dev/isa/ccbque.h>
#endif
#ifdef __FreeBSD__
#include <i386/scsi/scsi_low/ccbque.h>
#endif

GENERIC_CCB_ASSERT(scsi_low, ccb)

/* scsi pointer */
struct sc_p {
	u_int8_t *data;
	int datalen;
};

#define	SCSI_LOW_SETUP_PHASE(PHASE)			\
{							\
	ti->ti_ophase = ti->ti_phase;			\
	ti->ti_phase = (PHASE);				\
}

#define	SCSI_LOW_SETUP_MSGPHASE(PHASE)			\
{							\
	sc->sc_msgphase = (PHASE);			\
}

struct scsi_low_softc;

struct targ_info {
	TAILQ_ENTRY(targ_info) ti_chain;	/* targ_info link */

	struct scsi_low_softc *ti_sc;		/* our softc */
	u_int ti_id;				/* scsi id */

#define	TARG_CA		0x01
#define	TARG_SYNCH_PEND	0x02
	u_int ti_tflags;			/* target state I */

#define	UNIT_RDY	0x00
#define	UNIT_START	0x01
#define	UNIT_SYNCH	0x02
#define	UNIT_OK		0x03
	u_int ti_state;				/* target state II */

	u_int ti_flags;
	u_int ti_cfgflags;
	u_int ti_quirks;

#define	FATALIO		0x01
#define	ABORTIO		0x02
#define	TIMEOUTIO	0x04
#define	PDMAERR		0x08
#define	MSGERR		0x10
#define	PARITYERR	0x20
#define	ERRBUSY		0x40
#define	SCSI_LOW_ERRORBITS "\020\007busy\006parity\005msgerr\004pdmaerr\003timeout\002abort\001fatal"
	u_int ti_error;				/* error flags */

	/*****************************************
	 * scsi phase data
	 *****************************************/
	struct sc_p ti_sscp;			/* saved scsi data pointer */

#define	PH_NULL		0x00
#define	PH_ARBSTART	0x01
#define	PH_SELSTART	0x02
#define	PH_SELECTED	0x03
#define	PH_CMD		0x04
#define	PH_DATA		0x05
#define	PH_MSGIN	0x06
#define	PH_MSGOUT	0x07
#define	PH_STAT		0x08
#define	PH_DISC		0x09
#define	PH_COMPSEQ	0x0a
#define	PH_RESEL	0x0b
#define	PH_MSGOUTEXP	0x0c
	u_int ti_phase;				/* scsi phase */
	u_int ti_ophase;			/* old scsi phase */

	u_int8_t ti_status;			/* status in */

	u_int ti_msginptr;			/* msgin ptr */
	u_int8_t ti_msgin[MAXMSGLEN];		/* msgin buffer */

	u_int8_t ti_msgout;			/* last msgout byte */
	u_int8_t ti_emsgout;			/* last msgout byte */
	u_int ti_omsgoutlen;			/* for retry msgout */

	struct synch {
		u_int8_t offset;
		u_int8_t period;
		u_int8_t c3img;
	} ti_synch, ti_maxsynch;		/* synch data */

	struct ccb *ti_disccb;			/* disconnected */
};

/*************************************************
 * COMMON HEADER STRUCTURE
 *************************************************/
struct scsi_low_softc;
typedef struct scsi_low_softc *sc_low_t;

#define	SC_LOW_RESET_T (int (*) __P((sc_low_t, int)))
#define	SC_LOW_BUSRST_T (void (*) __P((sc_low_t)))
#define	SC_LOW_SELECT_T (int (*) __P((sc_low_t, struct ccb *)))
#define	SC_LOW_ATTEN_T (void (*) __P((sc_low_t)))
#define	SC_LOW_SYNCH_T (int (*) __P((sc_low_t, struct targ_info *, u_int, u_int)))
#define	SC_LOW_INTR_T (int (*) __P((void *)))
#define	SC_LOW_POWER_T (int (*) __P((sc_low_t, u_int)))

struct scsi_low_funcs {
	int (*scsi_low_reset) __P((sc_low_t, int));
	void (*scsi_low_bus_reset) __P((sc_low_t));

	int (*scsi_low_start_bus) __P((sc_low_t, struct ccb *));

	void (*scsi_low_attention) __P((sc_low_t));
	int (*scsi_low_synch) __P((sc_low_t, struct targ_info *, u_int, u_int)); 
	int (*scsi_low_intr) __P((void *));

#define	SCSI_LOW_POWDOWN	1
#define	SCSI_LOW_ENGAGE		2
	int (*scsi_low_power) __P((sc_low_t, u_int));
};

/* sc_flags */
#define	HW_POWDOWN	0x01
#define	HW_RESUME	0x02
#define	HW_PDMASTART	0x04
#define	HW_INACTIVE	0x08

/* msg phase */
#define	NOMSG		0x00
#define	DISCASSERT	0x01
#define	CMDCOMPLETE	0x02

/* read write */
#define	SCSI_LOW_WRITE	0
#define	SCSI_LOW_READ	1

#define	CFG_NODISC	0x01
#define	CFG_NOPARITY	0x02
#define	CFG_NOATTEN	0x04
#define	CFG_ASYNC	0x08

TAILQ_HEAD(targ_info_tab, targ_info);

#define	SCSI_LOW_DVNAME_LEN	16

#define	SCSI_LOW_SOFTC						\
	OS_DEPEND_DEVICE_HEADER					\
	OS_DEPEND_MISC_HEADER					\
	OS_DEPEND_SCSI_HEADER					\
								\
	struct targ_info *sc_ti[NTARGETS];			\
								\
	struct targ_info_tab sc_titab;				\
								\
	struct ccbtab sc_start;					\
	struct ccb *sc_nexus;					\
								\
	int sc_disc;						\
	struct targ_info *sc_selid;				\
								\
	u_int sc_flags;						\
								\
	u_int sc_msgphase;					\
								\
	struct sc_p sc_scp;					\
								\
	u_int sc_direction;					\
	int sc_tdatalen;					\
								\
	u_int sc_mask;						\
	int sc_wc;						\
								\
	u_int sc_oc;						\
	int sc_powc;						\
	u_int sc_rstep;						\
								\
	u_int sc_cfgflags;					\
								\
	struct scsi_low_funcs *sc_funcs;			\
								\
	struct scsi_sense sc_sense_cmd;				\
	struct scsi_sense_data sc_sense;			\
								\
	u_char sc_dvname[SCSI_LOW_DVNAME_LEN];

/*************************************************
 * SCSI LOW SOFTC
 *************************************************/
struct scsi_low_softc {
	SCSI_LOW_SOFTC
};

/*************************************************
 * PROTOTYPE
 *************************************************/
int scsi_low_attach __P((struct scsi_low_softc *));
void scsi_low_timeout __P((void *));
int scsi_low_scsi_cmd __P((struct scsi_xfer *));
void scsi_low_scsi_minphys __P((struct buf *));
void scsi_low_start __P((struct scsi_low_softc *));
int scsi_low_done __P((void *, struct ccb *));
void scsi_low_reset_nexus __P((struct scsi_low_softc *, int));
int scsi_low_msgout __P((struct scsi_low_softc *, struct targ_info *, struct ccb *));
void scsi_low_msgin __P((struct scsi_low_softc *, struct ccb *));
int scsi_low_restart __P((struct scsi_low_softc *, u_char *));
void scsi_low_free_ti __P((struct scsi_low_softc *));
struct targ_info *scsi_low_alloc_ti __P((struct scsi_low_softc *, int));

#define	SCSI_LOW_LUNUNK	((u_int) -1)
struct ccb *scsi_low_reselected __P((struct scsi_low_softc *, u_int, u_int));
int scsi_low_disconnected __P((struct scsi_low_softc *, struct ccb *));

/* reset */
#define	SCSI2_RESET_DELAY	5000000
#define	TWIDDLEWAIT	10000
void scsi_low_twiddle_wait __P((void));
void scsi_low_bus_reset __P((struct scsi_low_softc *));

void scsi_low_print __P((struct scsi_low_softc *, struct targ_info *));

int scsi_low_target_open __P((struct scsi_link *, struct cfdata *));
void scsi_low_calcf __P((struct targ_info *));
int scsi_low_reset __P((struct scsi_low_softc *, u_int));

#ifdef __NetBSD__
int scsi_low_activate __P((pisa_device_args_t));
int scsi_low_deactivate __P((pisa_device_args_t));
#endif

#ifdef __FreeBSD__
int scsi_low_activate __P((struct scsi_low_softc *, struct isa_device *));
int scsi_low_deactivate __P((struct scsi_low_softc *));
#endif

/* misc */
static inline u_int8_t scsi_low_identify __P((struct targ_info *ti));

static inline u_int8_t
scsi_low_identify(ti)
	struct targ_info *ti;
{

	return (ti->ti_flags & SCSI_LOW_DISC) ? 0xc0 : 0x80;
}

#define	ID_MSG_SETUP(TI, CB)					\
{								\
	(TI)->ti_msgout = scsi_low_identify(TI) | (CB)->lun;	\
	(TI)->ti_omsgoutlen = 0;				\
}

/* XXX: use scsi_message.h */
#define	ST_GOOD		0x00
#define	ST_CHKCOND	0x02
#define	ST_MET		0x04
#define	ST_BUSY		0x08
#define	ST_INTERGOOD	0x10
#define	ST_INTERMET	0x14
#define	ST_CONFLICT	0x18
#define	ST_QUEFULL	0x28

#define	MSG_COMP	0x00
#define	MSG_EXTEND	0x01

#define	MKMSG_EXTEND(XLEN, XCODE) ((((u_int)(XLEN)) << NBBY) | ((u_int)(XCODE)))
#define	MSG_EXTEND_MDPCODE	0x00
#define	MSG_EXTEND_MDPLEN	0x05
#define	MSG_EXTEND_SYNCHCODE	0x01
#define	MSG_EXTEND_SYNCHLEN	0x03
#define	MSG_EXTEND_WIDECODE	0x03
#define	MSG_EXTEND_WIDELEN	0x02

#define	MSG_SAVESP	0x02
#define	MSG_RESTORESP	0x03
#define	MSG_DISCON	0x04
#define	MSG_I_ERROR	0x05
#define	MSG_ABORT	0x06
#define	MSG_REJECT	0x07
#define	MSG_NOOP	0x08
#define	MSG_PARITY	0x09
#define	MSG_LCOMP	0x0a
#define	MSG_LCOMP_F	0x0b
#define	MSG_RESET	0x0c
#endif	/* !_SCSI_LOW_H_ */
