/*	$NetBSD$	*/
/* $Id: scsi_low_pisa.c,v 1.1.6.1 1998/07/27 10:47:50 kim Exp $ */
/*
 * [NetBSD for NEC PC98 series]
 *  Copyright (c) 1995, 1996 NetBSD/pc98 porting staff.
 *  Copyright (c) 1995, 1996 Naofumi HONDA.
 *  All rights reserved.
 * 
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef __NetBSD__
#include <dev/isa/scsi_low_if.h>
#endif /* __NetBSD__ */
#ifdef __FreeBSD__
#include <i386/scsi/scsi_low/scsi_low_if.h>
#endif /* __FreeBSD__ */

#ifdef __NetBSD__
#define	SCSIBUS_RESCAN
#endif

#ifdef __FreeBSD__
static inline int
scsi_low_deactivate_1(struct scsi_low_softc *sc)
#else
inline int
scsi_low_deactivate_1(sc)
	struct scsi_low_softc *sc;
#endif
{

	sc->sc_flags |= HW_INACTIVE;

#ifdef	SCSI_LOW_POWFUNC
	untimeout(scsi_low_recover, sc);
#endif	/* SCSI_LOW_POWFUNC */
	untimeout(scsi_low_timeout, sc);

	return 0;
}

#ifdef __NetBSD__
int
scsi_low_deactivate(arg)
	pisa_device_args_t arg;
{
	return (scsi_low_deactivate_1(arg->id));
}
#endif /* __NetBSD__ */

#ifdef __FreeBSD__
int
scsi_low_deactivate(struct scsi_low_softc *sc)
{

	return (scsi_low_deactivate_1(sc));
}
#endif /* __FreeBSD__ */

#ifdef __FreeBSD__
static inline int
scsi_low_activate_1(struct scsi_low_softc *sc, int flags)
#else
inline int
scsi_low_activate_1(sc, flags)
	struct scsi_low_softc *sc;
	int flags;
#endif
{
	int i, error;

	sc->sc_flags &= ~HW_INACTIVE;
#ifdef	__FreeBSD__
	sc->sc_cfgflags = ((sc->sc_cfgflags << 16) | (flags & 0xffff));
#else	/* __NetBSD__ */
	sc->sc_cfgflags = DVCFG_MKCFG(DVCFG_MAJOR(sc->sc_cfgflags),
				      DVCFG_MINOR(flags));
#endif

#ifdef __NetBSD__
	sc->sc_mask = (1 << ia->ia_irq);
#endif

	if (error = scsi_low_reset(sc, 0))
	{
		sc->sc_flags |= HW_INACTIVE;
		return error;
	}

	timeout(scsi_low_timeout, sc, SCSI_LOW_TIMEOUT_CHECK_INTERVAL * hz);

	/* rescan the scsi bus */
#ifdef	SCSIBUS_RESCAN
	if (arg->event == PISA_EVENT_INSERT && sc->sc_start.tqh_first == NULL)
		scsi_probe_busses((int) sc->sc_link.scsibus, -1, -1);
#endif
	/* start again */
	scsi_low_start(sc);
	return 0;
}

#ifdef __NetBSD__
int
scsi_low_activate(arg)
	pisa_device_args_t arg;
{
	struct isa_attach_args *ia = arg->ia;

	return (scsi_low_activate_1(arg->id, ia->ia_cfgflags));
}
#endif /* __NetBSD__ */

#ifdef __FreeBSD__
int
scsi_low_activate(sc, dev)
	struct scsi_low_softc *sc;
	struct isa_device *dev;
{
	return (scsi_low_activate_1(sc, dev->id_flags));
}
#endif /* __FreeBSD__ */
