/*	$NetBSD$	*/
/* $Id: tmc18c30.c,v 1.1.6.1 1998/07/27 10:49:08 kim Exp $ */
/*
 * [NetBSD for NEC PC98 series]
 *  Copyright (c) 1996 NetBSD/pc98 porting staff.
 *  Copyright (c) 1996 Naofumi HONDA.
 *  Copyright (c) 1996 Kouichi Matsuda.
 *  All rights reserved.
 * 
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef __NetBSD__
#include <dev/isa/tmc18c30if.h>
#endif
#ifdef __FreeBSD__
#include <i386/scsi/tmc18c30/tmc18c30if.h>
#endif

/***************************************************
 * USER SETTINGS
 ***************************************************/
/* DEVICE CONFIGURATION FLAGS (MINOR)
 *
 * 0x01   DISCONECT OFF
 * 0x02   PARITY LINE OFF
 * 0x04   IDENTIFY MSG OFF ( = single lun)
 * 0x08   SYNC TRANSFER OFF
 */

/* #define	STG_SYNC_SUPPORT */	/* NOT YET but easy */

/* For the 512 fifo type: change below */
#define	TMC18C30_FIFOSZ	0x800
#define	TMC18C30_FCB	1

#define	TMC18C50_FIFOSZ	0x2000
#define	TMC18C50_FCB	2

/***************************************************
 * DEBUG
 ***************************************************/
#ifndef DDB
#define Debugger() panic("should call debugger here (tmc18c30.c)")
#endif /* ! DDB */

#define	STG_DEBUG
#ifdef	STG_DEBUG
int stg_debug;

struct stg_statics {
	int disconnect;
	int reselect;
} stg_statics[NTARGETS];
#endif	/* STG_DEBUG */

/**************************************************************
 * DECLARE
 **************************************************************/
/* static */
static void stg_pio_read __P((struct stg_softc *, struct targ_info *));
static void stg_pio_write __P((struct stg_softc *, struct targ_info *));
static int stg_xfer __P((struct stg_softc *, u_int8_t *, int, int));
static int stg_synch __P((struct stg_softc *, struct targ_info *, u_int, u_int));
static int stg_reselected __P((struct stg_softc *));
static inline int stg_disconnected __P((struct stg_softc *, struct ccb *cb));
static inline void stg_pdma_end __P((struct stg_softc *, struct ccb *, struct targ_info *));
static int stghw_select_targ_wait __P((struct stg_softc *, int));
static void stghw_init __P((struct stg_softc *));
static void stghw_synch __P((struct stg_softc *, struct targ_info *));

static int stg_world_start __P((struct stg_softc *, int));
static void stghw_reset __P((struct stg_softc *));
static int stghw_start_selection __P((struct stg_softc *sc, struct ccb *));
static void stghw_bus_reset __P((struct stg_softc *));
static void stghw_attention __P((struct stg_softc *));

struct scsi_low_funcs stgfuncs =
{
	SC_LOW_RESET_T stg_world_start,
	SC_LOW_BUSRST_T stghw_bus_reset,

	SC_LOW_SELECT_T stghw_start_selection,

	SC_LOW_ATTEN_T stghw_attention,
	SC_LOW_SYNCH_T stg_synch,

	SC_LOW_INTR_T stg_sequencer,

	NULL,
};

/****************************************************
 * hwfuncs
 ****************************************************/
int
stghw_check(sc)
	struct stg_softc *sc;
{
	u_int16_t lsb, msb;
	DECLARE_IOPORT(sc);

	sc->sc_chip = TMCCHIP_UNK;
	sc->sc_fsz = TMC18C50_FIFOSZ;
	sc->sc_fcb = TMC18C50_FCB;
	sc->sc_fcsp = 0;

	sc->sc_fcRinit = FCTL_INTEN;
	sc->sc_fcWinit = FCTL_PARENB | FCTL_INTEN;

	if (sc->sc_cfgflags & CFG_NOATTEN)
		sc->sc_busc = 0;
	else
		sc->sc_busc = BCTL_ATN;

	lsb = INB(tmc_idlsb);
	msb = INB(tmc_idmsb);
	switch (msb << 8 | lsb)
	{
		case 0x6127:
			/* TMCCHIP_1800 not supported. (it's my policy) */
			sc->sc_chip = TMCCHIP_1800;
			return EINVAL;

		case 0x60e9:
			sc->sc_chip = TMCCHIP_18C50;
			sc->sc_fcsp |= FCTL_CLRINT;
			if (INB(tmc_cfg2) & 0x02)
			{
				sc->sc_chip = TMCCHIP_18C30;
				sc->sc_fsz = TMC18C30_FIFOSZ;
				sc->sc_fcb = TMC18C30_FCB;
			}
			break;

		default:
			return ENODEV;
	}

	sc->sc_icinit = ICTL_ALLINT | sc->sc_fcb;
	return 0;
}

static void
stghw_reset(sc)
	struct stg_softc *sc;
{
	DECLARE_IOPORT(sc);

	OUTB(tmc_ictl, 0);
	OUTB(tmc_bctl, BCTL_BUSFREE);
	OUTB(tmc_fctl, sc->sc_fcsp | sc->sc_fcRinit | FCTL_CLRFIFO);
	OUTB(tmc_fctl, sc->sc_fcRinit);
	OUTB(tmc_ictl, sc->sc_icinit);

	OUTB(tmc_ssctl, 0);
}

static void
stghw_init(sc)
	struct stg_softc *sc;
{
	DECLARE_IOPORT(sc);

	OUTB(tmc_ictl, 0);
	OUTB(tmc_bctl, BCTL_BUSFREE);
	OUTB(tmc_fctl, sc->sc_fcsp | sc->sc_fcRinit | FCTL_CLRFIFO);
	OUTB(tmc_fctl, sc->sc_fcRinit);
	OUTB(tmc_ictl, sc->sc_icinit);

	OUTB(tmc_ssctl, 0);
}

static void
stghw_synch(sc, ti)
	struct stg_softc *sc;
	struct targ_info *ti;
{
#ifdef	STG_SYNCH
	DECLARE_IOPORT(sc);

	OUTB(tmc_ssctl, ti->ti_synch.c3img);
#endif
}

/****************************************************
 * scsi low interface
 ****************************************************/
static void
stghw_attention(sc)
	struct stg_softc *sc;
{
	DECLARE_IOPORT(sc);

	OUTB(tmc_bctl, BCTL_BUSEN | BCTL_ATN);
}

static void
stghw_bus_reset(sc)
	struct stg_softc *sc;
{
	DECLARE_IOPORT(sc);

	OUTB(tmc_ictl, 0);
	OUTB(tmc_fctl, 0);
	OUTB(tmc_bctl, BCTL_RST);
	delay(100000);
	OUTB(tmc_bctl, BCTL_BUSFREE);
}

static int
stghw_start_selection(sc, cb)
	struct stg_softc *sc;
	struct ccb *cb;
{
	struct targ_info *ti = cb->ti;
	register u_int8_t stat;
	int s, tc = sc->sc_wc;
	DECLARE_IOPORT(sc);

	if (ti->ti_flags & SCSI_LOW_NOPARITY)
		sc->sc_fcRinit &= ~FCTL_PARENB;
	else
		sc->sc_fcRinit |= FCTL_PARENB;

	s = splhigh();
	if (sc->sc_disc > 0)
	{
		stat = INB(tmc_bstat);
		if ((stat & RESEL_PHASE_MASK)== PHASE_RESELECTED)
		{
			splx(s);
			return 1;
		}

		if (stat & (BSTAT_BSY | BSTAT_SEL))
		{
			delay(1);		/* XXX */

			stat = INB(tmc_bstat);
			if (stat & (BSTAT_BSY | BSTAT_SEL))
			{
				splx(s);
				return 1;
			}
		}

		OUTB(tmc_bctl, BCTL_BUSFREE);
		OUTB(tmc_scsiid, (1 << HOST_SCSI_ID));
		OUTB(tmc_fctl, sc->sc_fcRinit | FCTL_ARBIT);
		splx(s);
	}
	else
	{
		splx(s);

		do 
		{
		  	stat = INB(tmc_bstat);
			if ((stat & BSTAT_BSY) == 0)
				break;

			if (stat & (BSTAT_SEL | BSTAT_REQ))
			{
				printf("%s hardware failure(0x%x)\n", 
					sc->sc_dev.dv_xname, (u_int) stat);
				return 1;
			}
			else if (tc -- <= 0)
			{
				printf("%s bus free timeout\n", 
					sc->sc_dev.dv_xname);
				return 1;
			}
		}
		while (1);

		OUTB(tmc_bctl, BCTL_BUSFREE);
		OUTB(tmc_scsiid, (1 << HOST_SCSI_ID));
		OUTB(tmc_fctl, sc->sc_fcRinit | FCTL_ARBIT);
	}

	SCSI_LOW_SETUP_PHASE(PH_ARBSTART);
	return 0;
}

static int
stg_world_start(sc, fdone)
	struct stg_softc *sc;
	int fdone;
{
	struct targ_info *ti;
	u_int stat;
	int error;

	if (sc->sc_flags & HW_INACTIVE)
		return EBUSY;

	/* reset current nexus */
	scsi_low_reset_nexus((struct scsi_low_softc *) sc, fdone);

	sc->sc_cfgflags |= CFG_ASYNC;		/* XXX */

	/* hardware info init */
	for (ti = sc->sc_titab.tqh_first; ti; ti = ti->ti_chain.tqe_next)
	{
		ti->ti_state = UNIT_RDY;
		ti->ti_synch.period = ti->ti_synch.offset = 0;
		ti->ti_maxsynch.offset = 8;

		scsi_low_calcf(ti);
	}

	if (error = stghw_check(sc))
		return error;

	stghw_reset(sc);
	scsi_low_bus_reset((struct scsi_low_softc *) sc);
	stghw_init(sc);

#ifdef __NetBSD__
	softintr(sc->sc_mask);
#endif
	return 0;
}

static int
stg_synch(sc, ti, period, offset)
	struct stg_softc *sc;
	struct targ_info *ti;
	u_int period, offset;
{
	u_int res;

	period = period << 2;

	if (period >= 200)
	{
		ti->ti_synch.c3img = (period - 200) / 50;
		if (period % 50)
			ti->ti_synch.c3img ++;
		ti->ti_synch.c3img |= SSCTL_SYNCHEN;
	}
	else if (period >= 100)
	{
		ti->ti_synch.c3img = (period - 100) / 50;
		if (period % 50)
			ti->ti_synch.c3img ++;
		ti->ti_synch.c3img |= SSCTL_SYNCHEN | SSCTL_FSYNCHEN;
	}

	ti->ti_synch.period = period;
	ti->ti_synch.offset = offset;
	return 0;
}

/**************************************************************
 * PDMA functions
 **************************************************************/
static inline void
stg_pdma_end(sc, cb, ti)
	struct stg_softc *sc;
	struct ccb *cb;
	struct targ_info *ti;
{
	u_int len, tres;
	DECLARE_IOPORT(sc);

	sc->sc_flags &= ~HW_PDMASTART;

	if (ti->ti_phase == PH_DATA)
	{
		if ((sc->sc_direction & SCSI_LOW_READ) == 0)
		{
			len = INW(tmc_fdcnt);
			if (len)
			{
				tres = len + sc->sc_scp.datalen;
				if (tres <= (u_int) cb->datalen)
				{
					sc->sc_scp.data -= len;
					sc->sc_scp.datalen = tres;
				}
				else
				{
					ti->ti_error |= PDMAERR;
					printf("%s len %x >= datalen %x\n",
						sc->sc_dev.dv_xname,
						len, sc->sc_scp.datalen);
				}
			}
		}
	}
	else
	{

		printf("%s data phase miss\n", sc->sc_dev.dv_xname);
		ti->ti_error |= PDMAERR;
	}

#ifdef	STG_SYNCH
	OUTB(tmc_ssctl, 0);
#endif
	OUTB(tmc_fctl, sc->sc_fcRinit);
}

static void
stg_pio_read(sc, ti)
	struct stg_softc *sc;
	struct targ_info *ti;
{
	int tout = sc->sc_wc;
	u_int res;
	u_int8_t stat;
	DECLARE_IOPORT(sc);

	OUTB(tmc_fctl, sc->sc_fcRinit | FCTL_FIFOEN);

#ifdef	STG_SYNCH
	stghw_synch(sc, ti);
#endif

	sc->sc_flags |= HW_PDMASTART;
	while (sc->sc_scp.datalen > 0 && tout -- > 0)
	{
		res = INW(tmc_fdcnt);
		if (res == 0)
		{
			stat = INB(tmc_bstat);
			if ((stat & BSTAT_PHMASK) == BSTAT_IO)
				continue;
			break; /* phase mismatch */
		}

		/* XXX */
		if (res > sc->sc_scp.datalen)
			break;

		sc->sc_scp.datalen -= res;
		if (res & 1)
		{
			sc->sc_scp.data[0] = INB(tmc_rfifo);
			sc->sc_scp.data ++;
			res --;
		}

		INSW(tmc_rfifo, sc->sc_scp.data, res >> 1);
		sc->sc_scp.data += res;
	}

	if (tout <= 0)
		printf("%s pio read timeout\n", sc->sc_dev.dv_xname);
}

#define	WFIFO_CRIT	0x100

static void
stg_pio_write(sc, ti)
	struct stg_softc *sc;
	struct targ_info *ti;
{
	u_int res;
	int tout = sc->sc_wc;
	register u_int8_t stat;
	DECLARE_IOPORT(sc);

	stat = sc->sc_fcWinit | FCTL_FIFOEN | FCTL_FIFOW;
	OUTB(tmc_fctl, stat | FCTL_CLRFIFO);
	OUTB(tmc_fctl, stat);

#ifdef	STG_SYNCH
	stghw_synch(sc, ti);
#endif

	sc->sc_flags |= HW_PDMASTART;
	while (sc->sc_scp.datalen > 0 && tout -- > 0)
	{
		stat = INB(tmc_bstat);
		if ((stat & BSTAT_PHMASK) != 0)
			break;

		if (INW(tmc_fdcnt) >= WFIFO_CRIT)
			continue;

		res = (sc->sc_scp.datalen > WFIFO_CRIT) ? WFIFO_CRIT :
							  sc->sc_scp.datalen;
		sc->sc_scp.datalen -= res;
		if ((res & 0x01) != 0)
		{
			OUTB(tmc_wfifo, *sc->sc_scp.data++);
			res --;
		}

		OUTSW(tmc_wfifo, sc->sc_scp.data, res >> 1);
		sc->sc_scp.data += res;
	}

	if (tout <= 0)
		printf("%s pio write timeout\n", sc->sc_dev.dv_xname);
}

static int
stg_xfer(sc, buf, len, phase)
	struct stg_softc *sc;
	u_int8_t *buf;
	int len;
	int phase;
{
	int wc, ptr;
	u_int8_t stat;
	DECLARE_IOPORT(sc);

	if (phase & BSTAT_IO)
		OUTB(tmc_fctl, sc->sc_fcRinit);
	else
		OUTB(tmc_fctl, sc->sc_fcWinit);

	for (ptr = 0; len > 0; len --)
	{
		wc = sc->sc_wc;

		do
		{
			stat = INB(tmc_bstat);
			if ((stat & BSTAT_PHMASK) != (phase & BSTAT_PHMASK))
				goto bad;
			if (stat & BSTAT_REQ)
				break;
			if (wc -- <= 0)
			{
				printf("%s xfer tout\n", sc->sc_dev.dv_xname);
				goto bad;
			}
		}
		while (1);

		if (phase & BSTAT_IO)
			buf[ptr ++] = INB(tmc_rdata);
		else
			OUTB(tmc_wdata, buf[ptr ++]);
	}

bad:
	OUTB(tmc_fctl, sc->sc_fcRinit);
	return len;
}

/**************************************************************
 * disconnect & reselect (HW low)
 **************************************************************/
static int
stg_reselected(sc)
	struct stg_softc *sc;
{
	struct ccb *cb;
	u_int sid;
	DECLARE_IOPORT(sc);

	sid = (u_int) INB(tmc_scsiid);
	sid &= ~(1 << HOST_SCSI_ID);
	sid = ffs(sid) - 1;
	cb = scsi_low_reselected((struct scsi_low_softc *) sc, sid, SCSI_LOW_LUNUNK);
	if (cb == NULL)
		return 0;

#ifdef	STG_DEBUG
	stg_statics[sid].reselect ++;
#endif

	/* assert a busy line and ack the selection */
	OUTB(tmc_scsiid, (1 << HOST_SCSI_ID) || (1 << sid));
	OUTB(tmc_fctl, sc->sc_fcsp |
					  sc->sc_fcRinit | FCTL_CLRFIFO);
	OUTB(tmc_fctl, sc->sc_fcRinit);
	OUTB(tmc_bctl, BCTL_BSY | BCTL_BUSEN);
	return 1;
}

static inline int
stg_disconnected(sc, cb)
	struct stg_softc *sc;
	struct ccb *cb;
{
	DECLARE_IOPORT(sc);

	/* clear bus status & fifo */
	OUTB(tmc_fctl, sc->sc_fcRinit | FCTL_CLRFIFO);
	OUTB(tmc_fctl, sc->sc_fcRinit);
	OUTB(tmc_bctl, BCTL_BUSFREE);

#ifdef	STG_DEBUG
	if (sc->sc_msgphase == DISCASSERT)
		stg_statics[cb->ti->ti_id].disconnect ++;
#endif
	scsi_low_disconnected((struct scsi_low_softc *) sc, cb);
	scsi_low_start((struct scsi_low_softc *) sc);
	return 1;
}

/**************************************************************
 * SEQUENCER
 **************************************************************/
static int
stghw_select_targ_wait(sc, targ)
	struct stg_softc *sc;
	int targ;
{
	int wc, error = EIO;
	DECLARE_IOPORT(sc);

	OUTB(tmc_scsiid, (1 << HOST_SCSI_ID) | (1 << targ));
	OUTB(tmc_fctl, sc->sc_fcWinit & (~FCTL_INTEN));
	OUTB(tmc_bctl, BCTL_BUSEN | BCTL_SEL | sc->sc_busc);

	for (wc = 50000; wc; wc--)
	{
		if (INB(tmc_bstat) & BSTAT_BSY)
		{
			error = 0;
			break;
		}

		delay(1);
	}

	OUTB(tmc_fctl, sc->sc_fcRinit | FCTL_CLRFIFO);
	OUTB(tmc_fctl, sc->sc_fcRinit);
	return error;
}

int
stg_sequencer(sc)
	struct stg_softc *sc;
{
	struct targ_info *ti;
	struct ccb *cb;
	int len;
	u_int8_t status, astatus;
	DECLARE_IOPORT(sc);

	/*******************************************
	 * interrupt check
	 *******************************************/
	if (sc->sc_flags & HW_INACTIVE)
		return 0;

	astatus = INB(tmc_astat);
	status = INB(tmc_bstat);

	if ((astatus & ASTAT_STATMASK) == 0)
		return 0;

	if (astatus & ASTAT_SCSIRST)
	{
		OUTB(tmc_fctl,
			sc->sc_fcRinit | FCTL_CLRFIFO);
		OUTB(tmc_fctl, sc->sc_fcRinit);
		return 1;
	}

	/*******************************************
	 * debug section
	 *******************************************/
#ifdef	STG_DEBUG
	if (stg_debug)
	{
		scsi_low_print((struct scsi_low_softc *) sc, NULL);
		printf("%s st %x ist %x\n\n", sc->sc_dev.dv_xname,
		       status, astatus);
		if (stg_debug > 1)
			Debugger();
	}
#endif	/* STG_DEBUG */

	/*******************************************
	 * reselection & nexus
	 *******************************************/
	if ((status & RESEL_PHASE_MASK)== PHASE_RESELECTED)
		return stg_reselected(sc);

	if ((cb = sc->sc_nexus) == NULL)
		return 0;
	ti = cb->ti;

	if ((astatus & ASTAT_PARERR) &&
	     ti->ti_phase != PH_ARBSTART &&
	    (ti->ti_flags & SCSI_LOW_NOPARITY) == 0)
	{
		ti->ti_error |= PARITYERR;
		scsi_low_restart((struct scsi_low_softc *) sc,
				"scsi bus parity error");
		return 1;
	}

	/*******************************************
	 * aribitration & selection
	 *******************************************/
	switch (ti->ti_phase)
	{
	case PH_ARBSTART:
		OUTB(tmc_fctl,
			       sc->sc_fcRinit | FCTL_CLRFIFO);

		if ((astatus & ASTAT_ARBIT) == 0)
		{
			OUTB(tmc_fctl, sc->sc_fcRinit);
			SCSI_LOW_SETUP_PHASE(PH_NULL);
			sc->sc_selid = NULL;
			sc->sc_nexus = NULL;
			return 1;
		}
		else
			OUTB(tmc_fctl, sc->sc_fcWinit);

		/* assert select line */
		OUTB(tmc_scsiid,
				(1 << HOST_SCSI_ID) | (1 << ti->ti_id));
		OUTB(tmc_bctl,
				BCTL_SEL | BCTL_BUSEN | sc->sc_busc);
		SCSI_LOW_SETUP_PHASE(PH_SELSTART);
		return 1;

	case PH_SELSTART:
		sc->sc_selid = NULL;
		if ((status & BSTAT_BSY) == 0)
			if (stghw_select_targ_wait(sc, ti->ti_id))
				return stg_disconnected(sc, cb);

		/* attention assert */
		OUTB(tmc_fctl, sc->sc_fcRinit);
		OUTB(tmc_bctl, BCTL_BUSEN | sc->sc_busc);
		SCSI_LOW_SETUP_PHASE(PH_SELECTED);
		return 1;

	case PH_RESEL:
		/* clear a busy line */
		OUTB(tmc_fctl, sc->sc_fcRinit);
		OUTB(tmc_bctl, BCTL_BUSEN);
		break;
	}

	/*******************************************
	 * scsi seq
	 *******************************************/
	if (sc->sc_flags & HW_PDMASTART)
		stg_pdma_end(sc, cb, ti);

	switch (status & PHASE_MASK)
	{
	case COMMAND_PHASE:
		SCSI_LOW_SETUP_PHASE(PH_CMD);
		stg_xfer(sc, cb->cmd, cb->cmdlen, COMMAND_PHASE);
		break;

	case DATA_OUT_PHASE:
		SCSI_LOW_SETUP_PHASE(PH_DATA);
		sc->sc_direction = SCSI_LOW_WRITE;
		stg_pio_write(sc, ti);
		break;

	case DATA_IN_PHASE:
		SCSI_LOW_SETUP_PHASE(PH_DATA);
		sc->sc_direction = SCSI_LOW_READ;
		stg_pio_read(sc, ti);
		break;

	case STATUS_PHASE:
		SCSI_LOW_SETUP_PHASE(PH_STAT);
		ti->ti_status = INB(tmc_rdata);
		break;

	case MESSAGE_OUT_PHASE:
		OUTB(tmc_fctl, sc->sc_fcWinit);
		len = scsi_low_msgout((struct scsi_low_softc *) sc, ti, cb);
		if (len == 0)
		{
			if (cb->msgoutlen == 0)
				OUTB(tmc_bctl, BCTL_BUSEN);
			OUTB(tmc_wdata, ti->ti_msgout);
		}
		else if (len > 0)
		{
			stg_xfer(sc, cb->msgout, len, MESSAGE_OUT_PHASE);
			OUTB(tmc_bctl, BCTL_BUSEN);
		}
		OUTB(tmc_fctl, sc->sc_fcRinit);
		break;

	case MESSAGE_IN_PHASE:
		SCSI_LOW_SETUP_PHASE(PH_MSGIN);
		ti->ti_msgin[ti->ti_msginptr++] = INB(tmc_rdata);
		scsi_low_msgin((struct scsi_low_softc *)sc, cb);
		if (sc->sc_msgphase)
			return stg_disconnected(sc, cb);
		break;

	case BUSFREE_PHASE:
		printf("%s unexpected disconnection\n", sc->sc_dev.dv_xname);
		return stg_disconnected(sc, cb);

	default:
		printf("%s unknown phase bus %x intr %x\n",
			sc->sc_dev.dv_xname, status, astatus);
		break;
	}

	return 1;
}
