/*	$NetBSD$	*/
/* $Id: tmc18c30if.c,v 1.1.6.2 1998/09/03 15:59:31 shiba Exp $ */
/*
 * [NetBSD for NEC PC98 series]
 *  Copyright (c) 1996 NetBSD/pc98 porting staff.
 *  Copyright (c) 1996 Naofumi HONDA.
 *  Copyright (c) 1996 Kouichi Matsuda.
 *  All rights reserved.
 * 
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef __NetBSD__
#include <dev/isa/tmc18c30if.h>
#endif
#ifdef __FreeBSD__
#include <i386/scsi/tmc18c30/tmc18c30if.h>
#include "stg.h"
#endif

/***************************************************
 * PARAMS
 ***************************************************/
#define STG_MAX_CCB (2 * (NTARGETS - 1))

/***************************************************
 * ISA DEVICE STRUCTURE
 ***************************************************/
#ifdef __NetBSD__
int stgintr __P((void *));

struct cfdriver stg_cd = {
	NULL, "stg", DV_DULL
};

struct scsi_device stg_dev = {
	NULL,	/* Use default error handler */
	NULL,	/* have a queue, served by this */
	NULL,	/* have no async handler */
	NULL,	/* Use default 'done' routine */
};

struct scsi_adapter stg = {
	scsi_low_scsi_cmd,
	scsi_low_scsi_minphys,
	scsi_low_target_open,
	0,
};
#endif /* __NetBSD__ */

#ifdef __FreeBSD__
void stgintr (int unit);

/* pccard support */
#if NCARD > 0
static int	stg_card_intr(struct pccard_devinfo *);
static void	stg_card_unload(struct pccard_devinfo *);
static int	stg_card_resume(struct stg_softc *);
static int	stg_card_init(struct pccard_devinfo *);

static struct pccard_device stg_info = {
	"stg",
	stg_card_init,
	stg_card_unload,
	stg_card_intr,
	0,				/* Attributes - presently unused */
	&bio_imask			/* Interrupt mask for device */
		/* This should also include net_imask?? */
};

DATA_SET(pccarddrv_set, stg_info);

#ifdef SCSI_DETACH
static void stgdetach(struct isa_device *dev);
#endif
#endif /* NCARD */

struct isa_driver stgdriver = {
	stgprobe,
	stgattach,
	"stg"
};

struct scsi_device stg_dev = {
	NULL,	/* Use default error handler */
	NULL,	/* have a queue, served by this */
	NULL,	/* have no async handler */
	NULL,	/* Use default 'done' routine */
	"stg",
	0,
	{0, 0}
};

static struct scsi_adapter stg = {
	scsi_low_scsi_cmd,
	scsi_low_scsi_minphys,
	scsi_low_target_open,
	0,
	stg_adapter_info,
	"stg",
	{0, 0},
};
u_int32_t stg_adapter_info (int unit)
{
	return 1;
}

static struct stg_softc *stgdata[NSTG];
#endif /* __FreeBSD__ */

/**************************************************************
 * DECLARE
 **************************************************************/
extern int delaycount;
extern struct scsi_low_funcs stgfuncs;

/**************************************************************
 * General probe attach
 **************************************************************/
#ifdef __FreeBSD__
#if NCARD > 0
static int
stg_card_resume(struct stg_softc *sc)
{
#if 0 /* XXX stg_init */
	if (!stg_init(sc))
		return (ENXIO);
#endif
	printf("%s: resumed\n", sc->sc_dvname);
	return (0);
}

static int
stg_card_init(struct pccard_devinfo *devi)
{
	int unit = devi->pd_unit;
	struct stg_softc *sc = stgdata[unit];

	if (NSTG <= unit)
		return (ENODEV);
	/* XXX -- ncv_init */
	printf("probe stg\n");
	if (stgprobe(&devi->isahd) == 0)
		return (ENXIO);
	printf("attach stg\n");
	if (stgattach(&devi->isahd) == 0)
		return (ENXIO);
	return (0);
}

static void
stg_card_unload(struct pccard_devinfo *devi)
{
	struct stg_softc *sc = stgdata[devi->pd_unit];

	printf("%s: unload\n", sc->sc_dvname);
/*	scsi_low_deactivate((struct scsi_low_softc *)sc); */
#ifdef SCSI_DETACH
	stgdetach(&devi->isahd);
#endif
}

static int
stg_card_intr(struct pccard_devinfo *devi)
{

	(void) stg_sequencer(stgdata[devi->pd_unit]);
	return 1;
}
#endif /* NCARD > 0 */
#endif /* FreeBSD */

#ifdef __NetBSD__
int
stgprobe(parent, self, aux)
	struct device *parent;
	struct device *self;
	void *aux;
{
	struct stg_softc *sc = (void *)self;
	struct isa_attach_args *ia = aux;
	struct stg_hw *hw;
	int rv = 0;
	DECLARE_IOPORT(sc);

	if (ia->ia_iobase == IOBASEUNK || ia->ia_irq == IRQUNK)
		return 0;

	bc = ia->ia_bc;
	if (bus_io_map(bc, ia->ia_iobase, STGIOSZ, &ioh))
		return 0;

	sc->sc_bc = bc;
	sc->sc_ioh = ioh;
	sc->sc_dealyioh = ia->ia_delayioh;
#ifdef	__FreeBSD__
	sc->sc_cfgflags = (ia->ia_cfgflags & 0xffff);
#else	/* __NetBSD__ */
	sc->sc_cfgflags = DVCFG_MINOR(ia->ia_cfgflags);
#endif
	sc->sc_wc = delaycount * 2000;	/* 2 sec */
	sc->sc_funcs = &stgfuncs;

	if (stghw_check(sc))
		goto bad;

	if (scsi_low_attach((struct scsi_low_softc *) sc))
		goto bad;

	ia->ia_iosize = STGIOSZ;
	rv = 1;

bad:
	bus_io_unmap(bc, ioh, STGIOSZ);
	return rv;
}
#endif /* __NetBSD__ */

#ifdef __FreeBSD__
static int stghw_probe (struct stg_softc *sc, int flags)
{
	sc->sc_cfgflags = flags;
	sc->sc_wc = delaycount * 2000;	/* 2 sec */
	sc->sc_funcs = &stgfuncs;
	return (stghw_check(sc));
}

int stgprobe(struct isa_device *dev)
{
	struct stg_softc *sc;
	int unit = dev->id_unit;
	struct stg_hw *hw;
	char dvname[SCSI_LOW_DVNAME_LEN];
	DECLARE_IOPORT(sc);

	sprintf(dvname, "stg%d", unit);

	if (unit >= NSTG && !dev->id_reconfig)
	{
		printf("%s: unit number too high\n", dvname);
		return (0);
	}

	if (dev->id_iobase == 0)
	{
		printf("%s: no ioaddr is given\n", dvname);
		return (0);
	}

	if (dev->id_reconfig || stgdata[unit] != NULL)
	{
		sc = stgdata[unit];
		sc->sc_ioport = dev->id_iobase;

#ifdef	__FreeBSD__
		if (stghw_probe(sc, dev->id_flags & 0xffff))
			return (0);
#else	/* __NetBSD__ */
		if (stghw_probe(sc, DVCFG_MINOR(dev->id_flags)))
			return (0);
#endif
	}
	else
	{
		sc = malloc(sizeof(struct stg_softc), M_TEMP, M_NOWAIT);
		if (sc == NULL)
		{
			printf("%s: cannot malloc!\n", sc->sc_dvname);
			return (0);
		}
		bzero(sc, sizeof(struct stg_softc));
		strcpy (sc->sc_dvname, dvname);
		sc->unit = unit;
		stgdata[unit] = sc;
		sc->sc_ioport = dev->id_iobase;

		if (stghw_probe(sc, dev->id_flags))
		{
			free(sc, M_TEMP);
			stgdata[unit] = NULL;
			return (0);
		}
	}
	return (STGIOSZ);
}
#endif

int
stgprint(aux, name)
	void *aux;
	char *name;
{

	if (name != NULL)
		printf("%s: scsibus ", name);
	return UNCONF;
}

#ifdef __FreeBSD__
#define adapter_target adapter_targ
#define openings opennings
#endif

#ifdef __FreeBSD__
static inline void 
stg_setup_sc(struct stg_softc *sc)
#else
inline void stg_setup_sc(sc)
	struct stg_softc *sc;
#endif
{

	scsi_low_init_ccbque(STG_MAX_CCB);
	TAILQ_INIT(&sc->sc_start);

	sc->sc_link.adapter_softc = sc;
	sc->sc_link.adapter_target = 7;
	sc->sc_link.adapter = &stg;
	sc->sc_link.device = &stg_dev;
	sc->sc_link.openings = 2;
}

#undef adapter_target
#undef openings

#ifdef __NetBSD__
void
stgattach(parent, self, aux)
	struct device *parent;
	struct device *self;
	void *aux;
{
	struct stg_softc *sc = (void *)self;
	struct isa_attach_args *ia = aux;

	printf("\n");

	sc->sc_bc = ia->ia_bc;
	sc->sc_dealyioh = ia->ia_delayioh;
	if (bus_io_map(sc->sc_bc, ia->ia_iobase, STGIOSZ, &sc->sc_ioh))
		panic("%s: couldn't map io\n", sc->sc_dev.dv_xname);

	stg_setup_sc(sc);
	sc->sc_mask = (1 << ia->ia_irq);
	sc->sc_ih = isa_intr_establish(ia->ia_ic, ia->ia_irq, IST_EDGE,
	     IPL_BIO, stgintr, sc);

	timeout(scsi_low_timeout, sc, SCSI_LOW_TIMEOUT_CHECK_INTERVAL * hz);
}
#endif /* __NetBSD__ */

#ifdef __FreeBSD__

#ifdef SCSI_DETACH
static void
stgdetach(struct isa_device *dev)
{
	int unit = dev->id_unit;
	struct scsibus_data *scbus;

	scbus = (struct scsibus_data *)scsi_extend_get(stgdata[unit]->sc_link.scsibus);
}
#endif

int
stgattach(dev)
	struct isa_device *dev;
{
	int unit = dev->id_unit;
	struct stg_softc *sc = stgdata[unit];
	struct scsibus_data *scbus;

	if (scsi_low_attach((struct scsi_low_softc *) sc))
	{
		printf("%s: scsi low attach failed\n", sc->sc_dvname);
		return (0);
	}
	stg_setup_sc(sc);
	sc->sc_link.adapter_unit = unit;

	/*
	 * Prepare the scsibus_data area for the upperlevel
	 * scsi code.
	 */
	scbus = scsi_alloc_bus();
	if (!scbus)
		return (0);
	scbus->adapter_link = &sc->sc_link;
	/*
	 * ask the adapter what subunits are present
	 */
	scsi_attachdevs(scbus);
	timeout(scsi_low_timeout, sc, SCSI_LOW_TIMEOUT_CHECK_INTERVAL * hz);
	return 1;
}
#endif /* __FreeBSD__ */

#ifdef __NetBSD__
int stgintr (arg)
	void *arg;
{
	return (stg_sequencer((struct stg_softc *)arg));
}
#endif

#ifdef __FreeBSD__
void
stgintr (int unit)
{
	(void) stg_sequencer(stgdata[unit]);
}
#endif
