/*-
 * pccard noise interface.
 * Nate Williams, October 1997.
 * This file is in the public domain.
 */

#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/systm.h>

#include <machine/clock.h>

#include <pccard/driver.h>

#undef PCCARD_BEEP
#ifndef	PCCARD_BEEP
#define	PCCARD_BEEP	2
#endif	/* !PCCARD_BEEP */

#define	PCCARD_BEEP_PITCH0	1600
#define	PCCARD_BEEP_DURATION0	20
#define	PCCARD_BEEP_PITCH1	1200
#define	PCCARD_BEEP_DURATION1	40
#define	PCCARD_BEEP_PITCH2	3200
#define	PCCARD_BEEP_DURATION2	40

static void pccard_insert_beep_type0(void)
{
	/* dummy */
}

static void pccard_remove_beep_type0(void)
{
	/* dummy */
}

static void pccard_success_beep_type0(void)
{
	/* dummy */
}

static void pccard_failure_beep_type0(void)
{
	/* dummy */
}


static void pccard_insert_beep_type1(void)
{
	sysbeep(PCCARD_BEEP_PITCH0, PCCARD_BEEP_DURATION0);
}

static void pccard_remove_beep_type1(void)
{
	sysbeep(PCCARD_BEEP_PITCH0, PCCARD_BEEP_DURATION0);
}

static void pccard_success_beep_type1(void)
{
	sysbeep(PCCARD_BEEP_PITCH1, PCCARD_BEEP_DURATION1);
}

static void pccard_failure_beep_type1(void)
{
	sysbeep(PCCARD_BEEP_PITCH2, PCCARD_BEEP_DURATION2);
}

static void pccard_insert_beep0_type2(void *dummy)
{
	sysbeep_cancel();
	sysbeep(1200, 5);
}

static void pccard_insert_beep_type2(void)
{
	sysbeep(1600, 20);
	timeout(pccard_insert_beep0_type2, NULL, hz / 10);
}

static void pccard_remove_beep0_type2(void *dummy)
{
	sysbeep_cancel();
	sysbeep(1600, 5);
}

static void pccard_remove_beep_type2(void)
{
	sysbeep(1200, 20);
	timeout(pccard_remove_beep0_type2, NULL, hz / 10);
}

static void pccard_success_beep1_type2(void *dummy)
{
	sysbeep_cancel();
	sysbeep(800, 15);
}

static void pccard_success_beep0_type2(void *dummy)
{
	sysbeep_cancel();
	sysbeep(1000, 20);
	timeout(pccard_success_beep1_type2, NULL, hz / 15);
}

static void pccard_success_beep_type2(void)
{
	sysbeep(1200, 20);
	timeout(pccard_success_beep0_type2, NULL, hz / 15);
}

static void pccard_failure_beep1_type2(void *dummy)
{
	sysbeep_cancel();
	sysbeep(2800, 15);
}

static void pccard_failure_beep0_type2(void *dummy)
{
	sysbeep_cancel();
	sysbeep(2400, 20);
	timeout(pccard_failure_beep1_type2, NULL, hz / 15);
}
static void pccard_failure_beep_type2(void)
{
	sysbeep(2000, 20);
	timeout(pccard_failure_beep0_type2, NULL, hz / 15);
}

static void (*insert)(void) = pccard_insert_beep_type0;
static void (*remove)(void) = pccard_remove_beep_type0;
static void (*success)(void) = pccard_success_beep_type0;
static void (*failure)(void) = pccard_failure_beep_type0;

static enum beepstate allow_beep = BEEP_OFF;

/*
 * timeout function to keep lots of noise from
 * happening with insertion/removals.
 */
static void enable_beep(void *dummy)
{
	/* Should never be needed */
	untimeout(enable_beep, (void *)NULL);

	allow_beep = 1;
}

#ifdef PCCARD_BEEP
int pccard_beep_select(int type)
{
	int	errcode = 0;

	switch (type) {
	case 0:
		insert = pccard_insert_beep_type0;
		remove = pccard_remove_beep_type0;
		success = pccard_success_beep_type0;
		failure = pccard_failure_beep_type0;
		break;
	case 1:
		insert = pccard_insert_beep_type1;
		remove = pccard_remove_beep_type1;
		success = pccard_success_beep_type1;
		failure = pccard_failure_beep_type1;
		break;
	case 2:
		insert = pccard_insert_beep_type2;
		remove = pccard_remove_beep_type2;
		success = pccard_success_beep_type2;
		failure = pccard_failure_beep_type2;
		break;
	default:
		errcode = 1;
		break;
	}
	return errcode;
}

void pccard_insert_beep(void)
{
	insert();
}

void pccard_remove_beep(void)
{
	remove();
}

void pccard_success_beep(void)
{
	success();
}

void pccard_failure_beep(void)
{
	failure();
}

#else
void pccard_insert_beep(void)
{
	if (allow_beep == BEEP_ON) {
		sysbeep(PCCARD_BEEP_PITCH0, PCCARD_BEEP_DURATION0);
		allow_beep = 0;
		timeout(enable_beep, (void *)NULL, hz / 5);
	}
}

void pccard_remove_beep(void)
{
	if (allow_beep == BEEP_ON) {
		sysbeep(PCCARD_BEEP_PITCH0, PCCARD_BEEP_DURATION0);
		allow_beep = 0;
		timeout(enable_beep, (void *)NULL, hz / 5);
	}
}

void pccard_success_beep(void)
{
	sysbeep(PCCARD_BEEP_PITCH1, PCCARD_BEEP_DURATION1);
}

void pccard_failure_beep(void)
{
	sysbeep(PCCARD_BEEP_PITCH2, PCCARD_BEEP_DURATION2);
}

void pccard_beep_select(enum beepstate state)
{
	allow_beep = state;
}
#endif /* PCCARD_BEEP */
