/*
 * Copyright (c) 1998 NAKAGAWA, Yoshihisa <nakagawa@jp.freebsd.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: pcicvar.h,v 1.1.4.2 1998/09/02 17:57:51 shiba Exp $
 */

#ifndef	_PCCARD_PCICVAR_H_
#define	_PCCARD_PCICVAR_H_

#define	PCIC_IO_WIN	2
#define	PCIC_MEM_WIN	5

#ifndef PCIC_MAX_SLOTS
#define	PCIC_MAX_SLOTS	4
#endif

#define	PCIC_I82365	0		/* Intel chip */
#define	PCIC_IBM	1		/* IBM clone */
#define	PCIC_VLSI	2		/* VLSI chip */
#define	PCIC_PD672X	3		/* Cirrus logic 672x */
#define	PCIC_PD6710	4		/* Cirrus logic 6710 */
#define	PCIC_PD6729	5		/* Cirrus logic 6729 */
#define	PCIC_VG468	6		/* Vadem 468 */
#define	PCIC_VG469	7		/* Vadem 469 */
#define	PCIC_RF5C396	8		/* Ricoh RF5C396 */
#define	PCIC_IBM_KING	9		/* IBM KING PCMCIA Controller */
#define	PCIC_PC98	10		/* NEC PC98 PCMCIA Controller */
#define	PCIC_TI1130	11		/* TI PCI1130 CardBus bridge */
#define PCIC_TI1131	12		/* TI PCI1131 CardBus bridge */
#define PCIC_PD6832	13		/* Cirrus logic PD6832 CardBus bridge */
#define PCIC_OZ6730	14		/* O2Micro OZ6730 PC-card bridge */
#define PCIC_TOSHIBA	15		/* Toshiba CardBus bridge */
#define PCIC_RL5C465	16		/* Ricoh RL5C465 CardBus bridge */
#define PCIC_TI1250	17		/* TI PCI1250 CardBus bridge */
#define	PCIC_I82365DF	18		/* Intel chip */
#define	PCIC_RL5C475	19		/* Ricoh RL5C475/476 CardBus bridge */
#define PCIC_TI1220	20		/* TI PCI1220 CardBus bridge */
#define PCIC_TI1221	21		/* TI PCI1221 CardBus bridge */
#define	PCIC_RL5C478	22		/* Ricoh RL5C478 CardBus bridge */

/*
 *	Address of the controllers. Each controller can manage
 *	two PCMCIA slots. Up to 8 slots are supported in total.
 *	The PCIC controller is accessed via an index port and a
 *	data port. The index port has the 8 bit address of the
 *	register accessed via the data port. How I long for
 *	real memory mapped I/O!
 *	The top two bits of the index address are used to
 *	identify the port number, and the lower 6 bits
 *	select one of the 64 possible data registers.
 */

#ifndef IO_PCIC1
#define IO_PCIC1	0x3e0
#endif
#ifndef IO_PCIC2
#define IO_PCIC2	0x3e2
#endif
#ifndef IO_PCIC3
#define IO_PCIC3	0x3e4
#endif

#ifndef PCIC_IRQ
#define PCIC_IRQ	5
#endif

/*
 *	Per-slot data table.
 */
struct pcic_slot {
	int slotnum;			/* My slot number */
	int index;			/* Index register */
	int data;			/* Data register */
	int offset;			/* Offset value for index */
	int irq;			/* controller irq */
	char controller;		/* Device type */
	char revision;			/* Device Revision */
	struct slot *slt;		/* Back ptr to slot */
	u_char (*getb)(struct pcic_slot *sp, int reg);
	void   (*putb)(struct pcic_slot *sp, int reg, u_char val);
	u_char	*regs;			/* Pointer to regs in mem */
};

int	pcic_probe	__P((struct slot_ctrl *cinfo, int iobase, int irq));
int	pcic_attach	__P((struct slot_ctrl *cinfo, int flags));
void	pcicintr	__P((int unit)); /* XXXX */
void	pcic_intr_sc	__P((void *arg));

#endif	/* !_PCCARD_PCICVAR_H_ */
