/*
 * Copyright (c) 1997 Ted Faber
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice immediately at the beginning of the file, without modification,
 *    this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Absolutely no warranty of function or purpose is made by the author
 *    Ted Faber.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: pcic_p.c,v 1.4.2.2 1998/02/07 20:42:29 nate Exp $
 */

#include "pci.h"
#if NPCI > 0

/* share the PCI code with 2.2-RELENG and 3.0-CURRENT */
#ifndef	FREEBSD22
#define	FREEBSD22
#endif

#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/systm.h>
#include <sys/malloc.h>

#include <pci/pcireg.h>
#include <pci/pcivar.h>
#include <pci/pcic_p.h>

/* for slot.h */
#include <i386/isa/isa_device.h>
#include <pccard/cardinfo.h>
#include <sys/select.h>

#include <pccard/slot.h>
#include <pccard/pcicvar.h>
#include <pccard/cardinfo.h>

#ifndef FREEBSD22
#include <sys/interrupt.h>
#endif	/* FREEBSD22 */
#include <vm/vm.h>
#include <vm/pmap.h>
#ifdef	FREEBSD22
/* enable 3.0-compatible routine for 2.2.x */
#define PCIFBSD22
#endif

/*
 * XXX
 * Do our own re-probe protection until a configuration
 * manager can do it for us.  This ensures that we don't
 * reprobe a card already found by the PCI probes.
 */
static struct pcic_found
{
	u_long	port;
	char	probed;
} found[] =
{
	{ IO_PCIC1, 0 },
	{ IO_PCIC2, 0 },
	{ IO_PCIC3, 0 }
};

static u_long
get_pci_pcic_addr(void)
{
	u_long	iobase = 0;
	int	i;

	for (i=0; i < sizeof(found)/sizeof(struct pcic_found); i++) {
		if (!found[i].probed) {
			found[i].probed = 1;
			iobase = found[i].port;
			break;
		}
	}
	return iobase;
}

struct pci_pcic_info
{
	u_long	port;
	int	slots;
	pcici_t	tag;
#ifdef CB_TEST
	struct cb_slot *cb_info;
#endif /* CB_TEST */
	struct pci_pcic_info *next;
};

static struct pci_pcic_info *info_list = NULL;
#ifdef CB_TEST
struct cb_slot cb_slots[MAXSLOT];
#endif /* CB_TEST */

static void
set_pci_pcic_info(u_long iobase, pcici_t tag, int unit)
{
	struct pci_pcic_info *info;

	for (info = info_list; info; info = info->next) {
		if (info->port == iobase) {
			info->slots++;
			return;
		}
	}

	info = malloc(sizeof *info, M_DEVBUF, M_NOWAIT);
	if (info) {
		info->port = iobase;
		info->slots = 1;
		info->tag = tag;
		info->next = info_list;
#ifdef CB_TEST
		info->cb_info = &cb_slots[unit];
#endif /* CB_TEST */
		info_list = info;
	}
}

int
get_pci_pcic_info(u_long iobase, pcici_t *tag)
{
	struct pci_pcic_info *info;

	for (info = info_list; info; info = info->next) {
		if (info->port == iobase) {
			*tag = info->tag;
			return 1;
		}
	}
	return 0;
}

int
get_pci_pcic_slots(u_long iobase)
{
	struct pci_pcic_info *info;

	for (info = info_list; info; info = info->next) {
		if (info->port == iobase) {
			return info->slots;
		}
	}
	return 0;
}

#ifdef CB_TEST
struct cb_slot *
get_cb_info(u_long iobase)
{
	struct pci_pcic_info *info;

	for (info = info_list; info; info = info->next) {
		if (info->port == iobase) {
			return info->cb_info;
		}
	}
	return NULL;
}
#endif /* CB_TEST */

#ifdef	PCIFBSD22	/* compatibility with FreeBSD 3.0 PCI code */
u_long pci_cfgread(pcici_t tag, u_long reg, int bytes);
void pci_cfgwrite(pcici_t tag, u_long reg, u_long data, int bytes);

u_long 
pci_cfgread(pcici_t tag, u_long reg, int bytes)
{
	int off;
	u_long x;

	off = (reg % 4);
	x = pci_conf_read(tag, reg - off);
	x >>= (8 * off);
	switch (bytes) {
	case 1:
		x &= 0xffu;
		break;
	case 2:
		x &= 0xffffu;
		break;
	}
	return x;
}

void   
pci_cfgwrite(pcici_t tag, u_long reg, u_long data, int bytes)
{
	int off;
	u_long x, mask;

	off = (reg % 4);
	x = pci_conf_read(tag, reg - off);
	switch (bytes) {
	case 1:
		data &= 0xffu;
		mask = 0xffu;
		break;
	case 2:
		data &= 0xffffu;
		mask = 0xffffu;
		break;
	}
	data <<= (off * 8);
	mask <<= (off * 8);
	x &= (~mask);
	x |= data;
	pci_conf_write(tag, reg - off, x);
}
#endif	/* PCIFBSD22 */

static u_long pcic_pci_count = 0;
static char *pcic_pci_probe(pcici_t, pcidi_t);
static void  pcic_pci_attach(pcici_t, int);

#ifdef LKM
struct pci_device pcic_pci_driver = {
#else
static struct pci_device pcic_pci_driver = {
#endif
	"pcic",
	pcic_pci_probe,
	pcic_pci_attach,
	&pcic_pci_count,
	NULL
};

#ifndef LKM
DATA_SET(pcidevice_set, pcic_pci_driver);
#endif

/*
 * Return the ID string for the controller if the vendor/product id
 * matches, NULL otherwise.
 */
static char *
pcic_pci_probe(pcici_t tag, pcidi_t type)
{
#ifdef CB_TEST
	u_long  data, class, subclass;
#endif /* CB_TEST */

	switch (type) {
	/* 32bit CardBus bridges */
	case PCI_DEVICE_ID_PCIC_CLPD6832_CARDBUS:
		return "Cirrus Logic PD6832 PCI/CardBus Bridge";
	case PCI_DEVICE_ID_PCIC_TI1130_CARDBUS:
		return ("TI PCI-1130 PCI-CardBus Bridge");
	case PCI_DEVICE_ID_PCIC_TI1131_CARDBUS:
		return ("TI PCI-1131 PCI-CardBus Bridge");
	case PCI_DEVICE_ID_PCIC_TI1220_CARDBUS:
		return ("TI PCI-1220 PCI-CardBus Bridge");
	case PCI_DEVICE_ID_PCIC_TI1221_CARDBUS:
		return ("TI PCI-1221 PCI-CardBus Bridge");
	case PCI_DEVICE_ID_PCIC_TI1250_CARDBUS:
		return ("TI PCI-1250 PCI-CardBus Bridge");
	case PCI_DEVICE_ID_TOSHIBA_TOPIC95_CARDBUS:
		return ("Toshiba ToPIC95 PCI-CardBus Bridge");
	case PCI_DEVICE_ID_TOSHIBA_TOPIC97_CARDBUS:
		return ("Toshiba ToPIC97 PCI-CardBus Bridge");
	case PCI_DEVICE_ID_RICOH_RL5C465_CARDBUS:
		return("Ricoh RL5C465 PCI-CardBus Brige");
	case PCI_DEVICE_ID_RICOH_RL5C475_CARDBUS:
		return("Ricoh RL5C475 PCI-CardBus Brige");
	case PCI_DEVICE_ID_RICOH_RL5C476_CARDBUS:
		return("Ricoh RL5C476 PCI-CardBus Brige");
	case PCI_DEVICE_ID_RICOH_RL5C478_CARDBUS:
		return("Ricoh RL5C478 PCI-CardBus Brige");
	/* 16bit PC-card bridges */
	case PCI_DEVICE_ID_PCIC_CLPD6729:
		return "Cirrus Logic PD6729/6730 PC-card Bridge";
	case PCI_DEVICE_ID_PCIC_OZ6729:
		return "O2micro OZ6729 PC-Card Bridge";
	case PCI_DEVICE_ID_PCIC_OZ6730:
		return "O2micro OZ6730 PC-Card Bridge";

	default:
#ifdef CB_TEST
		data = pci_conf_read(tag, PCI_CLASS_REG);
		class = data & PCI_CLASS_MASK;
		subclass = data & PCI_SUBCLASS_MASK;

		if (class == PCI_CLASS_BRIDGE
		    && subclass == PCI_SUBCLASS_BRIDGE_CARDBUS) {
			return("generic PCI/CardBus Brige");
		}
#endif /* CB_TEST */
		break;
	}
	return (NULL);
}

static void pci_legacy_attach(pcici_t tag, pcidi_t id, u_long legacy16);
static void generic_cardbus_attach(pcici_t tag, pcidi_t id, int unit);
static void print_cardbus_registers(pcici_t tag);

/*
 * TI1XXX PCI-CardBus Host Adapter specific function code.
 * This function is separated from pcic_pci_attach().
 * Support Device: TI1130,TI1131,TI1250,TI1220.
 * Test Device: TI1221.
 * Takeshi Shibagaki(shiba@jp.freebsd.org).
 */
static void
ti1xxx_pci_init(pcici_t tag, u_long pcic_pci_id, u_int8_t func)
{
	u_long	syscntl,devcntl,cardcntl;
	char	buf[128];
	int 	ti113x = (pcic_pci_id == PCI_DEVICE_ID_PCIC_TI1130_CARDBUS)
			 || (pcic_pci_id == PCI_DEVICE_ID_PCIC_TI1131_CARDBUS);

	syscntl  = pci_cfgread(tag, TI113X_PCI_SYSTEM_CONTROL, 4);
	devcntl  = pci_cfgread(tag, TI113X_PCI_DEVICE_CONTROL, 1);
	cardcntl = pci_cfgread(tag, TI113X_PCI_CARD_CONTROL,   1);

	switch(ti113x){
		case 0 :
			strcpy(buf, "TI12XX PCI Config Reg: ");
			break;
		case 1 :
			strcpy(buf, "TI113X PCI Config Reg: ");
			/*
			 * Defalut card control register setting is PCI interrupt.
			 * The method of this code switches PCI INT and ISA IRQ
			 * by bit 7 of Bridge Control Register(Offset:0x3e,0x13e).
			 * Takeshi Shibagaki(shiba@jp.freebsd.org)
			 */
			if(!func) cardcntl |=  TI113X_CARDCNTL_PCI_IRQ_ENA;
			cardcntl |= TI113X_CARDCNTL_PCI_IREQ;
			cardcntl |= TI113X_CARDCNTL_PCI_CSC;
			if(syscntl & TI113X_SYSCNTL_CLKRUN_ENA){
				if(syscntl & TI113X_SYSCNTL_CLKRUN_SEL)
					strcat(buf, "[clkrun irq 12]");
				else
					strcat(buf, "[clkrun irq 10]");
			}
			break;
	}
	if(cardcntl & TI113X_CARDCNTL_RING_ENA)
		strcat(buf, "[ring enable]");
	if(cardcntl & TI113X_CARDCNTL_SPKR_ENA)
		strcat(buf, "[speaker enable]");
	if(syscntl & TI113X_SYSCNTL_PWRSAVINGS)
		strcat(buf, "[pwr save]");
	switch(devcntl & TI113X_DEVCNTL_INTR_MASK){
		case TI113X_DEVCNTL_INTR_ISA :
			strcat(buf, "[CSC parallel isa irq]");
			break;
		case TI113X_DEVCNTL_INTR_SERIAL :
			strcat(buf, "[CSC serial isa irq]");
			break;
		case TI113X_DEVCNTL_INTR_NONE :
			strcat(buf, "[pci only]");
			break;
		case TI12XX_DEVCNTL_INTR_ALLSERIAL :
			strcat(buf, "[FUNC pci int + CSC serial isa irq]");
			break;
	}
#if 0
	pci_cfgwrite(tag, TI113X_PCI_SYSTEM_CONTROL, syscntl, 4);
	pci_cfgwrite(tag, TI113X_PCI_DEVICE_CONTROL, devcntl, 1);
#endif /* 0 */
	pci_cfgwrite(tag, TI113X_PCI_CARD_CONTROL,  cardcntl, 1);
	if(ti113x){
		cardcntl = pci_cfgread(tag, TI113X_PCI_CARD_CONTROL, 1);
		switch(cardcntl & TI113X_CARDCNTL_MASK){
			case TI113X_FUNC0_VALID :
				if(!func) strcat(buf, "[FUNC pci int]");
				break;
			case TI113X_FUNC1_VALID :
				if(func) strcat(buf, "[FUNC pci int]");
				break;
		}
	}
	printf("%s\n",buf);
}

/*
 * CLPD683X PCI-CardBus Host Adapter specific function code.
 * This function is separated from pcic_pci_attach().
 * Support Device: CLPD6832.
 * Takeshi Shibagaki(shiba@jp.freebsd.org).
 */
static void
clpd683x_pci_init(pcici_t tag, u_long pcic_pci_id)
{
	u_long	brgcntl;
	char	buf[128];

	strcpy(buf, "CLPD683X PCI Config Reg: ");
	brgcntl = pci_cfgread(tag, CB_PCI_BRIDGE_CTRL, 2);
	/*
	 * CLPD683X management interrupt enable bit is bit 11 in bridge
	 * control register(offset 0x3d).
	 * When this bit is turned on, card status change interrupt sets
	 * on ISA IRQ interrupt.
	 * Takeshi Shibagaki(shiba@jp.freebsd.org).
	 */
	brgcntl |= CL6832_BCR_MGMT_INT_ENA;
	pci_cfgwrite(tag, CB_PCI_BRIDGE_CTRL, brgcntl, 2);
	brgcntl = pci_cfgread(tag, CB_PCI_BRIDGE_CTRL, 2);
	switch(brgcntl & CL6832_BCR_MGMT_INT_ENA){
		case CL6832_BCR_MGMT_INT_ENA :
			strcat(buf, "[CSC isa irq]");
			break;
	}
	printf("%s\n",buf);
}

/*
 * RL5C4XX PCI-CardBus Host Adapter specific function code.
 * This function is separated from pcic_pci_attach().
 * Support Device: RL5C465,RL5C475,RL5C476.
 * Test Device: RL5C478.
 * Takeshi Shibagaki(shiba@jp.freebsd.org).
 */
static void
rl5c4xx_init_setup(pcici_t tag, u_long pcic_pci_id)
{
	u_long	mngcntl;
	char	buf[128];
	int 	new = (pcic_pci_id == PCI_DEVICE_ID_RICOH_RL5C478_CARDBUS);

	strcpy(buf, "RL5C4XX PCI Config Reg: ");
	if(!new){
		mngcntl = pci_cfgread(tag, 0x91, 1);
		/*
		 * This code is made with hacking PCI Config Reg of
		 * AL-N2T515J5(RL5C476).
		 * When The bit 3 of PCI Config Reg(offset 0x91) turned on,
		 * CSC interrupt would set to ISA IRQ.
		 * Takeshi Shibagaki(shiba@jp.freebsd.org).
		 */ 
		mngcntl |= 0x08;
		pci_cfgwrite(tag, 0x91, mngcntl, 1);
		mngcntl = pci_cfgread(tag, 0x91, 1);
		switch(mngcntl & 0x08){
			case 0x08 :
				strcat(buf, "[CSC isa irq]");
				break;
		}
	}
	printf("%s\n",buf);
}

/*
 * General PCI based card dispatch routine.
 */
static void
pcic_pci_attach(pcici_t tag, int unit)
{
	int	i;
	u_int8_t	func;
	u_int8_t	bus, slot;
	u_long		pcic_pci_id, command, legacy16;
	struct slot_ctrl	*cinfo;

	bus = (u_int8_t)pci_get_bus(tag);
	slot = (u_int8_t)pci_get_device(tag);
	func = (u_int8_t)pci_get_function(tag);

	if (bootverbose)
		printf("bus = %d, slot = %d, func = %d\n", bus, slot, func);

	pcic_pci_id = pci_conf_read(tag, PCI_ID_REG);

	/* Init. CardBus/PC-card controllers as 16-bit PC-card controllers */

	/* Place any per "slot" initialization here */
	/*
	 * In sys/pci/pcireg.h, PCI_COMMAND_STATUS_REG must be separated
	 * PCI_COMMAND_REG(0x04) and PCI_STATUS_REG(0x06).
	 * Takeshi Shibagaki(shiba@jp.freebsd.org).
	 */
	command = pci_cfgread(tag, PCI_COMMAND_STATUS_REG, 4);
	command |= PCI_COMMAND_IO_ENABLE | PCI_COMMAND_MEM_ENABLE;
	pci_cfgwrite(tag, PCI_COMMAND_STATUS_REG, command, 4);

	switch (pcic_pci_id) {
	/* CardBus Bridges */
	case PCI_DEVICE_ID_PCIC_TI1130_CARDBUS:
	case PCI_DEVICE_ID_PCIC_TI1131_CARDBUS:
	case PCI_DEVICE_ID_PCIC_TI1220_CARDBUS:
	case PCI_DEVICE_ID_PCIC_TI1221_CARDBUS:
	case PCI_DEVICE_ID_PCIC_TI1250_CARDBUS:
		ti1xxx_pci_init(tag, pcic_pci_id, func);
		generic_cardbus_attach(tag, pcic_pci_id, unit);
		break;
	case PCI_DEVICE_ID_PCIC_CLPD6832_CARDBUS:
		clpd683x_pci_init(tag, pcic_pci_id);
		generic_cardbus_attach(tag, pcic_pci_id, unit);
		break;
	case PCI_DEVICE_ID_TOSHIBA_TOPIC95_CARDBUS:
	case PCI_DEVICE_ID_TOSHIBA_TOPIC97_CARDBUS:
		generic_cardbus_attach(tag, pcic_pci_id, unit);
		break;
	case PCI_DEVICE_ID_RICOH_RL5C465_CARDBUS:
	case PCI_DEVICE_ID_RICOH_RL5C475_CARDBUS:
	case PCI_DEVICE_ID_RICOH_RL5C476_CARDBUS:
	case PCI_DEVICE_ID_RICOH_RL5C478_CARDBUS:
		rl5c4xx_init_setup(tag, pcic_pci_id);
		generic_cardbus_attach(tag, pcic_pci_id, unit);
		break;

	/* Legacy PC-card Bridges */
	case PCI_DEVICE_ID_PCIC_OZ6729:
	case PCI_DEVICE_ID_PCIC_OZ6730:
	case PCI_DEVICE_ID_PCIC_CLPD6729:
		legacy16 = pci_conf_read(tag, PCI_MAP_REG_START) & ~PCI_MAP_IO;
		/* XXX -- should be initialized more carefully */
		if (!legacy16) {
			legacy16 = get_pci_pcic_addr();
			pci_conf_write(tag, PCI_MAP_REG_START, 
				legacy16 | PCI_MAP_IO);
		}
#if 0
		pci_legacy_attach(tag, pcic_pci_id, legacy16);
#endif
		set_pci_pcic_info(legacy16, tag, unit);
		break;
	}

#if 0
	if (func == 0) {
		/* Place any per "chip" initialization here */
		switch (pcic_pci_id) {
		default:
			break;
		}
	}
#endif

#if 0
	/* 
	 * TODO:
	 *	use PCI IRQ for CSC
	 */
	/* XXXX */
	if(!(pci_map_int(tag, cbintr, (void *)0, NULL))) {
		printf();
		return;
	}
#endif

	if (bootverbose)
		print_cardbus_registers(tag);

	return;
}

char
set_controller_info(pcici_t tag, struct slot_ctrl *cinfo)
{
	char controller = -1;
	pcidi_t id;
#ifdef CB_TEST
	u_long  data, class, subclass;

	data = pci_conf_read(tag, PCI_CLASS_REG);
	class = data & PCI_CLASS_MASK;
	subclass = data & PCI_SUBCLASS_MASK;

	if (class == PCI_CLASS_BRIDGE && subclass == PCI_SUBCLASS_BRIDGE_CARDBUS) {
		cinfo->cardbus = 1;
	}
#endif /* CB_TEST */

	if (!cinfo) return -1;

	id = pci_conf_read(tag, PCI_ID_REG);

	switch (id) {
	/* 32bit CardBus bridges */
	case PCI_DEVICE_ID_PCIC_CLPD6832_CARDBUS:
		controller = PCIC_PD6832;
		cinfo->name = "Cirrus Logic PD-6832 "
				"[CardBus bridge mode]";
		break;
	case PCI_DEVICE_ID_PCIC_TI1130_CARDBUS:
		controller = PCIC_TI1130;
		cinfo->name = "TI PCI-1130 "
				"[CardBus bridge mode]";
		break;
	case PCI_DEVICE_ID_PCIC_TI1131_CARDBUS:
		controller = PCIC_TI1131;
		cinfo->name = "TI PCI-1131 "
				"[CardBus bridge mode]";
		break;
	case PCI_DEVICE_ID_PCIC_TI1220_CARDBUS:
		controller = PCIC_TI1220;
		cinfo->name = "TI PCI-1220 "
				"[CardBus bridge mode]";
		break;
	case PCI_DEVICE_ID_PCIC_TI1221_CARDBUS:
		controller = PCIC_TI1221;
		cinfo->name = "TI PCI-1221 "
				"[CardBus bridge mode]";
		break;
	case PCI_DEVICE_ID_PCIC_TI1250_CARDBUS:
		controller = PCIC_TI1250;
		cinfo->name = "TI PCI-1250 "
				"[CardBus bridge mode]";
		break;
	case PCI_DEVICE_ID_TOSHIBA_TOPIC95_CARDBUS:
		controller = PCIC_TOSHIBA;
		cinfo->name = "TOSHIBA TOPIC95 "
				"[CardBus bridge mode]";
		break;
	case PCI_DEVICE_ID_TOSHIBA_TOPIC97_CARDBUS:
		controller = PCIC_TOSHIBA;
		cinfo->name = "TOSHIBA TOPIC97 "
				"[CardBus bridge mode]";
		break;
	case PCI_DEVICE_ID_RICOH_RL5C465_CARDBUS:
		controller = PCIC_RL5C465;
		cinfo->name = "Ricoh RL5C465 "
				"[CardBus bridge mode]";
		break;
	case PCI_DEVICE_ID_RICOH_RL5C475_CARDBUS:
	case PCI_DEVICE_ID_RICOH_RL5C476_CARDBUS:
		controller = PCIC_RL5C475;
		cinfo->name = "Ricoh RL5C475/476 "
				"[CardBus bridge mode]";
		break;
	case PCI_DEVICE_ID_RICOH_RL5C478_CARDBUS:
		controller = PCIC_RL5C478;
		cinfo->name = "Ricoh RL5C478 "
				"[CardBus bridge mode]";
		break;
	/* 16bit PC-card bridges */
	case PCI_DEVICE_ID_PCIC_CLPD6729:
		controller = PCIC_PD6729;
		cinfo->name = "Cirrus Logic PD-6729/6730 "
				"[i82365 compatible mode]";
		break;
	case PCI_DEVICE_ID_PCIC_OZ6729:
	case PCI_DEVICE_ID_PCIC_OZ6730:
		controller = PCIC_OZ6730;
		cinfo->name = "O2micro OZ6729/6730 "
				"[i82365 compatible mode]";
		break;
	default:
#ifdef CB_TEST
		if (cinfo->cardbus) {
			cinfo->name = "generic CardBus bridge "
					"[CardBus bridge mode]";
		}
#endif /* CB_TEST */
		break;
	}

	return controller;
}

static void
pci_legacy_attach(pcici_t tag, pcidi_t id, u_long legacy16)
{
#if 0
	struct slot_ctrl *cinfo;

	cinfo = malloc(sizeof *cinfo, M_DEVBUF, M_NOWAIT);
	if (!cinfo) return NULL;

	/* initialize slot_ctrl *cinfo */
	bzero(cinfo, sizeof *cinfo);
	cinfo->cardbus = 0;

	/* XXXX */
	pcic_probe(cinfo, legacy16, PCIC_IRQ);
	/* override controller info */
	set_controller_info(tag, cinfo);

	pcic_attach(cinfo);
#endif

	return;
}

static void
generic_cardbus_attach(pcici_t tag, pcidi_t id, int unit)
{
	u_int8_t	func;
	u_long	cb_base;
	u_int8_t	cb_bus, cb_sub_bus;
	u_int8_t	cb_irq;
	u_int16_t	command, brgcntl;
	static u_long   iobase;	/* XXXX */

#ifdef CB_TEST
	static u_long	cb_base_mem = CB_BASE_MEM;
	static u_int8_t	cb_bus_base = CB_BUS_BASE;
#if 0
	struct slot_ctrl *cinfo;

	cinfo = malloc(sizeof *cinfo, M_DEVBUF, M_NOWAIT);
	if (!cinfo) return;
#endif

	/* PCI Command and Status */
	command = pci_cfgread(tag, CB_PCI_COMMAND, 2);
	command |= CB_CMD_DEFAULT;
	pci_cfgwrite(tag, CB_PCI_COMMAND, command, 2);

	/* PC-Card Socket Status/Control Registers Base Address */
	cb_base = pci_cfgread(tag, CB_PCI_SOCKET_BASE, 4);
	if (cb_base == 0) {
		cb_base = cb_base_mem;
		pci_cfgwrite(tag, CB_PCI_SOCKET_BASE, cb_base, 4);
		cb_base_mem += 0x1000;
	}
	cb_base = (u_long)pmap_mapdev(cb_base, 0x1000);

	/* CardBus Bus Number */
	cb_bus = pci_cfgread(tag, CB_PCI_CB_BUS_NUM, 1);
	cb_sub_bus = pci_cfgread(tag, CB_PCI_CB_SUB_BUS_NUM, 1);
	if (cb_bus == 0) {
		/* XXXX */
		cb_bus = cb_bus_base;
		cb_sub_bus = cb_bus_base + 2;
		pci_cfgwrite(tag, CB_PCI_CB_BUS_NUM, cb_bus, 1);
		pci_cfgwrite(tag, CB_PCI_CB_SUB_BUS_NUM, cb_sub_bus, 1);
		cb_bus_base += 3;
	}

	/* Interrupt Line */
	cb_irq = pci_cfgread(tag, CB_PCI_INT_LINE, 1);
	if (cb_irq > 15) cb_irq = 0;
#if 0
	if (cb_irq == 0) {
		/* XXXX test code */
		cb_irq = 10;
		pci_cfgwrite(tag, CB_PCI_INT_LINE, cb_irq, 1);
	}
#endif

	/* Bridge Control */
	brgcntl = pci_cfgread(tag, CB_PCI_BRIDGE_CTRL, 2);
#if 0
	brgcntl &= ~CB_BCR_CB_RESET;
#endif
	if (!cb_irq) brgcntl |= CB_BCR_INT_EXCA; /* XXX */
	pci_cfgwrite(tag, CB_PCI_BRIDGE_CTRL, brgcntl, 2);
#endif /* CB_TEST */

#ifndef CB_TEST
#ifndef PCI_INTR_ENA
	/*
	 * This code exists to output ISA IRQ indicated in ExCA register(0x03).
	 * The recent PAO needs this 3 lines absolutely. Don't remove!!
	 * Takeshi Shibagaki(shiba@jp.freebsd.org).
	 */
	brgcntl = pci_cfgread(tag, CB_PCI_BRIDGE_CTRL, 2);
	brgcntl |= CB_BCR_INT_EXCA;
	pci_cfgwrite(tag, CB_PCI_BRIDGE_CTRL, brgcntl, 2);
#else /* PCI_INTR_ENA */
	/*
	 * To output PCI INT indicated in Interrupt Pin Register(0x3d).
	 * When each IRQ is routed to per slot, would be enable.
	 * Takeshi Shibagaki(shiba@jp.freebsd.org).
	 */
	brgcntl = pci_cfgread(tag, CB_PCI_BRIDGE_CTRL, 2);
	brgcntl &= ~CB_BCR_INT_EXCA;
	pci_cfgwrite(tag, CB_PCI_BRIDGE_CTRL, brgcntl, 2);
#endif /* PCI_INTR_ENA */
#endif /* CB_TEST */

	func = (u_int8_t)pci_get_function(tag);

	/* 16bit Legacy Mode Base Address */
	if (func == 0) {
#if 1
		u_long  legacy16;

		legacy16 = pci_cfgread(tag, CB_PCI_LEGACY16_IOADDR, 2)
			& ~PCI_MAP_IO;
		iobase = legacy16;	/* XXXX */
		if (!legacy16) {
			u_int   res;

			legacy16 = get_pci_pcic_addr();
			if (!legacy16) {
				printf("not free legacy mode address\n");
				return;
			}
			iobase = legacy16;	/* XXXX */
			legacy16 |= PCI_MAP_IO;
			pci_cfgwrite(tag, CB_PCI_LEGACY16_IOADDR, legacy16, 2);
			res = pci_cfgread(tag, CB_PCI_LEGACY16_IOADDR, 2)
				& ~PCI_MAP_IO;
			if (bootverbose)
				printf("pcic%d, Legacy PC-card "
					"16bit I/O address is set to 0x%x "
					"(res: 0x%x)\n", 
					unit, iobase, res);
			iobase = res;
		} else {
			if (bootverbose)
				printf("pcic%d, Legacy PC-card "
					"16bit I/O address [0x%x]\n", 
					unit, iobase);
		}
	}
#else
	pci_cfgwrite(tag, CB_PCI_LEGACY16_IOADDR, 0, 4);
#endif

#if 0
	/* initialize slot_ctrl *cinfo */
	bzero(cinfo, sizeof *cinfo);
	cinfo->cardbus = 1;
	cinfo->tag = tag;

	/* XXXX */
	pcic_probe(cinfo, cb_base, PCIC_IRQ);
	/* override controller info */
	set_controller_info(tag, cinfo);

	/* attach controller and slot */
	pcic_attach(cinfo);
#endif

#ifdef CB_TEST
        /* XXXX */
	cb_slots[unit].tag = tag;
	cb_slots[unit].cb_base = cb_base;
	cb_slots[unit].cb_bus = cb_bus;
	cb_slots[unit].cb_irq = cb_irq;
#endif /* CB_TEST */

	set_pci_pcic_info(iobase, tag, unit);

	return;
}

static void
print_cardbus_registers(pcici_t tag)
{
	int i,j;
	u_char *p;
	u_long *pl;

	printf ("PCI Config space:\n");
	for (j=0;j<0x100;j+=16) {
		printf("%02x: ", j);
		for (i=0;i<16;i+=4) {
			printf(" %08x", pci_conf_read(tag, i+j));
		}
		printf("\n");
	}
#ifdef CB_TEST
	/* pci_conf_write(tag, 0x44, 1); */
	p = (u_char*)pmap_mapdev(pci_conf_read(tag, 0x10), 0x1000);
	pl = (u_long *)p;
	printf ("Cardbus Socket registers:\n");
	printf("00: ");
	for (i=0;i<4;i+=1) printf(" %08x:",pl[i]);
	printf("\n10: ");
	for (i=4;i<8;i+=1) printf(" %08x:",pl[i]);
	printf ("\nExCa registers:\n");
	for (i=0;i<0x40;i+=16)
		printf("%02x: %16D\n",i, p+0x800+i," ");
#endif /* CB_TEST */
}

#endif /* NPCI > 0 */
