/*
 * Copyright (c) 1997 Ted Faber
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice immediately at the beginning of the file, without modification,
 *    this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Absolutely no warranty of function or purpose is made by the author
 *    Ted Faber.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: pcic_p.h,v 1.3.2.1 1998/02/07 20:38:58 nate Exp $
 */

#ifndef	_PCIC_P_H_
#define	_PCIC_P_H_

#ifdef PCIFBSD22	/* compatibility with FreeBSD 3.0 PCI code */
u_long	pci_cfgread(pcici_t tag, u_long reg, int bytes);
void	pci_cfgwrite(pcici_t tag, u_long reg, u_long data, int bytes);
#endif /* PCIFBSD22 */

/* PCI-PCMCIA Device IDs */
#define	PCI_DEVICE_ID_PCIC_CLPD6729		0x11001013ul
#define	PCI_DEVICE_ID_PCIC_OZ6729		0x67291217ul
#define	PCI_DEVICE_ID_PCIC_OZ6730		0x673A1217ul

/* PCI-CardBus Device IDs */
#define PCI_DEVICE_ID_PCIC_CLPD6832_CARDBUS	0x11101013ul
#define PCI_DEVICE_ID_PCIC_TI1130_CARDBUS	0xac12104cul
#define PCI_DEVICE_ID_PCIC_TI1131_CARDBUS	0xac15104cul
#define PCI_DEVICE_ID_PCIC_TI1220_CARDBUS	0xac17104cul
#define PCI_DEVICE_ID_PCIC_TI1221_CARDBUS	0xac19104cul	/* XXX - test */
#define PCI_DEVICE_ID_PCIC_TI1250_CARDBUS	0xac16104cul
#define PCI_DEVICE_ID_TOSHIBA_TOPIC95_CARDBUS	0x060a1179ul	/* XXX - test */
#define PCI_DEVICE_ID_TOSHIBA_TOPIC97_CARDBUS	0x060f1179ul	/* XXX - test */
#define PCI_DEVICE_ID_RICOH_RL5C465_CARDBUS	0x04651180ul
#define PCI_DEVICE_ID_RICOH_RL5C475_CARDBUS	0x04751180ul
#define PCI_DEVICE_ID_RICOH_RL5C476_CARDBUS	0x04761180ul
#define PCI_DEVICE_ID_RICOH_RL5C478_CARDBUS	0x04781180ul	/* XXX - test */

/* PCI Configuration Registers (common) */
#define	CB_PCI_VENDOR_ID	0x00	/* vendor ID */
#define	CB_PCI_DEVICE_ID	0x02	/* device ID */
#define	CB_PCI_COMMAND		0x04	/* PCI command */
#define	CB_PCI_STATUS		0x06	/* PCI status */
#define	CB_PCI_REVISION_ID	0x08	/* PCI revision ID */
#define	CB_PCI_CLASS		0x09	/* PCI class code */
#define	CB_PCI_CACHE_LINE_SIZE	0x0c	/* Cache line size */
#define	CB_PCI_LATENCY		0x0d	/* PCI latency timer */
#define	CB_PCI_HEADER_TYPE	0x0e	/* PCI header type */
#define	CB_PCI_BIST		0x0f	/* Built-in self test */
#define	CB_PCI_SOCKET_BASE	0x10	/* Socket/ExCA base address reg. */
#define	CB_PCI_CB_STATUS	0x16	/* CardBus Status */
#define	CB_PCI_PCI_BUS_NUM	0x18	/* PCI bus number */
#define	CB_PCI_CB_BUS_NUM	0x19	/* CardBus bus number */
#define	CB_PCI_CB_SUB_BUS_NUM	0x1A	/* Subordinate CardBus bus number */
#define	CB_PCI_CB_LATENCY	0x1A	/* CardBus latency timer */
#define	CB_PCI_MEMBASE0		0x1C	/* Memory base register 0 */
#define	CB_PCI_MEMLIMIT0	0x20	/* Memory limit register 0 */
#define	CB_PCI_MEMBASE1		0x24	/* Memory base register 1 */
#define	CB_PCI_MEMLIMIT1	0x28	/* Memory limit register 1 */
#define	CB_PCI_IOBASE0		0x2C	/* I/O base register 0 */
#define	CB_PCI_IOLIMIT0		0x30	/* I/O limit register 0 */
#define	CB_PCI_IOBASE1		0x34	/* I/O base register 1 */
#define	CB_PCI_IOLIMIT1		0x38	/* I/O limit register 1 */
#define	CB_PCI_INT_LINE		0x3C	/* Interrupt Line */
#define	CB_PCI_INT_PIN		0x3D	/* Interrupt Pin */
#define	CB_PCI_BRIDGE_CTRL	0x3E	/* Bridge Control */
#define	CB_PCI_SUBSYS_VENDOR_ID	0x40	/* Subsystem Vendor ID */
#define	CB_PCI_SUBSYS_ID	0x42	/* Subsystem ID */
#define	CB_PCI_LEGACY16_IOADDR	0x44	/* Legacy 16bit I/O address */

/* For PCI Command register (CB_PCI_COMMAND) */
#define CB_CMD_DEFAULT		0x0007

/* For Bridge Control register (CB_PCI_BRIDGE_CTRL) */
#define CB_BCR_CB_RESET		0x0040
#define CB_BCR_INT_EXCA		0x0080

/* Vendor specific PCI configuration regisrers */

/* Texas Instruments PCI-1130/1131 CardBus Controller */
#define TI113X_PCI_SYSTEM_CONTROL	0x80	/* System Control */
#define TI113X_PCI_RETRY_STATUS		0x90	/* Retry Status */
#define TI113X_PCI_CARD_CONTROL		0x91	/* Card Control */
#define TI113X_PCI_DEVICE_CONTROL	0x92	/* Device Control */
#define TI113X_PCI_BUFFER_CONTROL	0x93	/* Buffer Control */
#define TI113X_PCI_SOCKET_DMA0		0x94	/* Socket DMA Register 0 */
#define TI113X_PCI_SOCKET_DMA1		0x98	/* Socket DMA Register 1 */

/* Card control register (TI113X_SYSTEM_CONTROL == 0x80) */
#define TI113X_SYSCNTL_VCC_PROTECT	0x00200000u
#define TI113X_SYSCNTL_CLKRUN_SEL	0x00000080u
#define	TI113X_SYSCNTL_PWRSAVINGS	0x00000040u
#define TI113X_SYSCNTL_KEEP_CLK		0x00000002u
#define TI113X_SYSCNTL_CLKRUN_ENA	0x00000001u

/* Card control register (TI113X_CARD_CONTROL == 0x91) */
#define TI113X_CARDCNTL_RING_ENA	0x80u
#define TI113X_CARDCNTL_ZOOM_VIDEO	0x40u
#define TI113X_CARDCNTL_PCI_IRQ_ENA	0x20u
#define TI113X_CARDCNTL_PCI_IREQ	0x10u
#define TI113X_CARDCNTL_PCI_CSC		0x08u
#define	TI113X_CARDCNTL_MASK		(TI113X_CARDCNTL_PCI_IRQ_ENA | TI113X_CARDCNTL_PCI_IREQ | TI113X_CARDCNTL_PCI_CSC)
#define	TI113X_FUNC0_VALID		TI113X_CARDCNTL_MASK
#define	TI113X_FUNC1_VALID		(TI113X_CARDCNTL_PCI_IREQ | TI113X_CARDCNTL_PCI_CSC)
/* Reserved bit				0x04u */
#define TI113X_CARDCNTL_SPKR_ENA	0x02u
#define TI113X_CARDCNTL_INT		0x01u

/* Device control register (TI113X_DEVICE_CONTROL == 0x92) */
#define	TI113X_DEVCNTL_5V_SOCKET	0x40u
#define	TI113X_DEVCNTL_3V_SOCKET	0x20u
#define	TI113X_DEVCNTL_INTR_MASK	0x06u
#define	TI113X_DEVCNTL_INTR_NONE	0x00u
#define	TI113X_DEVCNTL_INTR_ISA		0x02u
#define	TI113X_DEVCNTL_INTR_SERIAL	0x04u
/* TI112X specific code */
#define	TI12XX_DEVCNTL_INTR_ALLSERIAL	0x06u

/* Cirrus Logic PD-6832 CardBus Controller */
#define CL6832_PCI_DMA_SLAVE		0x48	/* DMA Slave configuration */
#define CL6832_PCI_SOCKET_NUMBER	0x4C	/* Socket Number */

/* bridge control register (CB_PCI_BRIDGE_CTRL) */
#define CL6832_BCR_IREQ_INT_ENA		0x0080u	/* IREQ-INT Enable */
#define CL6832_BCR_MGMT_INT_ENA		0x0800u	/* Management Intr Enable */

/* Vendor specific ExCA registers (for common part, see i82365.h) */

/* Texas Instruments PCI-1130/1131 CardBus Controller */
#define	TI113X_ExCA_IO_OFFSET0		0x36	/* Offset of I/O window */
#define	TI113X_ExCA_IO_OFFSET1		0x38	/* Offset of I/O window */
#define	TI113X_ExCA_MEM_WINDOW_PAGE	0x3C	/* Memory Window Page */

/* Cirrus Logic PD-6832 CardBus Controller */
#define CL6832_ExCA_MISC1		0x16	/* Misc control 1 */
#define CL6832_ExCA_FIFO		0x17	/* FIFO control */
#define CL6832_ExCA_MISC2		0x1e	/* Misc control 2 */
#define CL6832_ExCA_CHIP_INFO		0x1f	/* Chip information */
#define CL6832_ExCA_ATA			0x1f	/* ATA Control */
#define CL6832_ExCA_EXTEND_INDEX	0x2e	/* Extended Index */
#define CL6832_ExCA_EXTEND_DATA		0x2f	/* Extended Data */

/* Extended Index (CL6832_ExCA_EXTEND_INDEX == 0x2e) */
#define CL6832_ExCA_EXT_EXT1		0x03	/* Extension Control 1 */
#define CL6832_ExCA_EXT_GEN_MAP_U0	0x05	/* Gen Map Upper addr 0 */
#define CL6832_ExCA_EXT_GEN_MAP_U1	0x06	/* Gen Map Upper addr 1 */
#define CL6832_ExCA_EXT_GEN_MAP_U2	0x07	/* Gen Map Upper addr 2 */
#define CL6832_ExCA_EXT_GEN_MAP_U3	0x08	/* Gen Map Upper addr 3 */
#define CL6832_ExCA_EXT_GEN_MAP_U4	0x09	/* Gen Map Upper addr 4 */
#define CL6832_ExCA_EXT_EXT2		0x0b	/* Extension Control 2 */
#define CL6832_ExCA_EXT_GEN_MAP_U5	0x20	/* Gen Map Upper addr 5 */
#define CL6832_ExCA_EXT_GEN_MAP_U6	0x21	/* Gen Map Upper addr 6 */
#define CL6832_ExCA_EXT_PCI_SPACE	0x22	/* PCI Space Control */
#define CL6832_ExCA_EXT_PCCARD_SPACE	0x23	/* PC-Card Space Control */
#define CL6832_ExCA_EXT_WINDOW_TYPE	0x24	/* Window Type Select */
#define CL6832_ExCA_EXT_MISC3		0x25	/* Misc Control 3 */
#define CL6832_ExCA_EXT_SMB_POWER	0x26	/* SMB Sock Pwr Ctrl Addr */
#define CL6832_ExCA_EXT_GEN_MAP_EXTRA0	0x27	/* Gen Map Extra Control 0 */
#define CL6832_ExCA_EXT_GEN_MAP_EXTRA1	0x28	/* Gen Map Extra Control 1 */
#define CL6832_ExCA_EXT_GEN_MAP_EXTRA2	0x29	/* Gen Map Extra Control 2 */
#define CL6832_ExCA_EXT_GEN_MAP_EXTRA3	0x2a	/* Gen Map Extra Control 3 */
#define CL6832_ExCA_EXT_GEN_MAP_EXTRA4	0x2b	/* Gen Map Extra Control 4 */
#define CL6832_ExCA_EXT_GEN_MAP_EXTRA5	0x2c	/* Gen Map Extra Control 5 */
#define CL6832_ExCA_EXT_GEN_MAP_EXTRA6	0x2d	/* Gen Map Extra Control 6 */
#define CL6832_ExCA_EXT_CSC		0x2e	/* Ext. Card Status Change */
#define CL6832_ExCA_EXT_MISC4		0x2f	/* Misc Control 4 */
#define CL6832_ExCA_EXT_MISC5		0x30	/* Misc Control 5 */
#define CL6832_ExCA_EXT_MISC6		0x31	/* Misc Control 5 */

/* infomation about PCI PC-card bridges */

struct	pci_pcic {
	u_int8_t	bus;
	u_int8_t	slot;
	u_long		legacy16;
	u_int		pci_id;
};

#ifdef CB_TEST
struct	cb_slot {
	/* CardBus support */
	pcici_t		tag;
	u_long		cb_base;
	u_int8_t	cb_bus;
	u_int8_t	cb_irq;
};

#ifndef CB_BASE_MEM
#define CB_BASE_MEM	0x68000000
#endif
#ifndef CB_BUS_BASE
#define CB_BUS_BASE	0x20
#endif

/*
 *      CardBus Register index addresses.
 */

#define CB_STATUS_EVENT		0x00	/* Status Event Register */
#define CB_STATUS_MASK		0x04	/* Status Mask Register */
#define CB_PRESENT_STATE	0x08	/* Present State Register */
#define CB_FORCE		0x0c	/* Force Register */
#define CB_CONTROL		0x10	/* Control Register */

/* For Present State and Force register (CB_PRESENT_STATE and CB_FORCE) */
#define CB_CSTSCHG	0x00000001u	/* Card Status Change */
#define CB_CCD1		0x00000002u	/* Card Detects 1 */
#define CB_CCD2		0x00000004u	/* Card Detects 2 */
#define CB_POWER_CYCLE	0x00000008u	/* Power Cycle */
#define CB_16BIT_CARD	0x00000010u	/* 16-bit PC Card */
#define CB_CB_CARD	0x00000020u	/* CardBus PC Card */
#define CB_INTERRUPT	0x00000040u	/* Interrupt pin from card */
#define CB_NOT_A_CARD	0x00000080u	/* Not a Card */
#define CB_DATA_LOST	0x00000100u	/* Data Lost */
#define CB_BAD_VCC_REQ	0x00000200u	/* Bad Vcc Request */
#define CB_5V_CARD	0x00000400u	/* 5V Card */
#define CB_3V_CARD	0x00000800u	/* 3V Card */
#define CB_XV_CARD	0x00001000u	/* XV Card */
#define CB_YV_CARD	0x00002000u	/* YV Card */
#define CB_CV_TEST	0x00004000u	/* test VS and CCD */
#define CB_5V_AVAIL	0x10000000u	/* 5V Vcc Voltage Available */
#define CB_3V_AVAIL	0x20000000u	/* 3V Vcc Voltage Available */
#define CB_XV_AVAIL	0x40000000u	/* XV Vcc Voltage Available */
#define CB_YV_AVAIL	0x80000000u	/* YV Vcc Voltage Available */

/* For Control register (CB_CONTROL) */
#define CB_VPP_CTRL	0x0007u	/* Vpp control bits */
#define CB_VPP_0V	0x0000u	/* 0 volts */
#define CB_VPP_12V	0x0001u	/* 12 volts */
#define CB_VPP_5V	0x0002u	/* 5 volts */
#define CB_VPP_3V	0x0003u	/* 3 volts */
#define CB_VCC_CTRL	0x0070u	/* Vcc control bits */
#define CB_VCC_0V	0x0000u	/* 0 volts */
#define CB_VCC_5V	0x0020u	/* 5 volts */
#define CB_VCC_3V	0x0030u	/* 3 volts */
#define CB_STOP_CLOCK	0x0080u	/* stop CardBus clock */
#endif /* CB_TEST */

#endif	/* _PCIC_P_H_ */
